(* **************************************************************************
 *
 * Dependency analysis.
 *
 * $Source: /home/uniform/awe/repository/src/Extensions/abs_tactics.sml,v $
 * $Revision: 1.2 $ of $Date: 2005/07/12 13:47:55 $
 *)

signature ABS_TACTICS = 
  sig
    (* -- elementary abstraction tactics -- *)
    val abs_thms     : string list -> thm-> thm 
    val abs_ops      : (string* typ) list-> thm-> thm   
    val abs_types    : typ list-> thm-> thm

    val expand_thms  : string list -> thm -> thm

    (* -- composed tactics -- *)

    (* Abstract to another theory *)
    val abs_to_thy   : theory-> thm-> thm
    (* Abstract to another theory, also abstracting given operations.
     * Use this for overloaded operations (typeclasses, e.g. 0 or +). *)
    val abs_to_thy_with : theory-> (string* typ) list-> thm-> thm

    (* the following are legacy tactics (used to be called repl_...) *)
    val abs_thms_thy : theory-> thm-> thm
    val abs_ops_thy  : theory-> thm-> thm
    val abs_types_thy: theory-> thm-> thm
    val full_abs_old : theory-> thm-> thm  (* theory argument is ignored *)

    (* abstracts everything *)
    val full_abs     : thm-> thm 

    (* Translate along a signature morphism:
     * abstract all theorems from target theory, then move across. *)
    val abs_translate : SigMorph.sig_morph-> thm-> thm

    (* Translate an axiomatic base along a signature morphism.
     * First argument is the signature morphism, second is the mapping
     * of the base, and the third is a "roadmap" of the theorems to be 
     * translated in dependency order paired with a list of its
     * preconditions, as returned by the DepAna.saturate. 
     * We do the dependency analysis separately, as it takes time and users
     * may want to try different bases etc. first before actually moving
     * anything.
     * As opposed to the other tactics, this binds theorems to their name.
     * Theorems have the same name as before, but in the target theory, so
     * Src.foo becomes Trg.foo. 
     *)
    val abs_translate_base :  SigMorph.sig_morph ->
                (string * string) list -> (string * string list) list -> unit
     
    (* visible flags *)
    val check_active : bool ref
    val verbose      : bool ref
    val default_sort : sort ref
  end

structure AbsTac (* : ABS_TACTICS *) = 

struct

    val check_active = ref true
    val verbose      = ref true
    val default_sort = ref ["HOL.type"]

    fun error s    = Library.error ("AbsTac: "^s)
    fun warning s  = Library.warning ("AbsTac: "^s)

    fun prtlen l = BasisLibrary.Int.toString (length l)
    fun prt_verb s = if !verbose then writeln s else ()

    fun prt_verb_ops ops = 
	prt_verb("Abstracting ops: "^(commas (map fst ops)))
    fun prt_verb_types_isar thy thm ts = 
	 prt_verb("Abstracting types: "^(commas (map (Sign.string_of_typ (sign_of thy)) ts)))
    fun prt_verb_types thm = 
	 prt_verb_types_isar (theory_of_thm thm)

    fun check res doit =
	case (!check_active, res) of
	    (_, None)=> doit()
	  | (true, Some err) => error err
          | (false, Some err)=> (warning err; doit())


    (* --- Elementary tactics ---------------------------------------------- *)

    (* --- Expanding theorems -------------------------------------------- *)
    fun expand_thms_isar list thy thm =
      let val sorted_list = rev (DepAna.sort_dep (DepAna.get_dg thy) list)
          val sg = sign_of_thm thm
          val prf = Aux.fullprf_of thm
          val nuprf = foldl (fn (curprf, expth)=> 
                        Reconstruct.expand_proof sg [(expth, None)] curprf) 
                        (prf, sorted_list)
          val full_nuprf = Reconstruct.expand_proof sg [("", None)] nuprf
      in Aux.derive_thm thy full_nuprf end

   fun expand_thms list thm = expand_thms_isar list (theory_of_thm thm) thm

    (* --- Abstracting theorems -------------------------------------------- *)

    (* We expand "" theorems in proofs BEFORE doing abstraction *)
    fun abs_thms_isar names thy thm = 
      let val anon_thm = Aux.occurs_thm "" (Aux.strip_thm (Aux.proof_of thm))
	  val nuthm = if anon_thm then (Aux.check_prf_of thy thm) else thm
          val abs_thms = (foldl (fn (th, nm)=> 
		Abstract.abstract_thm thy nm th)) (nuthm, names)
      in abs_thms end

    fun abs_thms names thm = abs_thms_isar names (theory_of_thm thm) thm


    (* --- Abstracting operations ------------------------------------------ *)

    (* operation's name is no good as a varname, e.g. infixes or starts
     * with a non-letter. *)
    fun no_var_name nu_smb = 
	String.isPrefix "op " nu_smb orelse 
	not (Char.isAlpha (String.sub(nu_smb, 0)))

    fun abs_ops_isar ops thy thm = 
      let fun nuop i = String.concat ["nuop", BasisLibrary.Int.toString i]
	  fun smb_prs i acc []         = acc
	    | smb_prs i acc ((smb, sty)::lst) = 
	      let val nu_smb = NameSpace.base smb
	      in if no_var_name nu_smb
		     then smb_prs (i+1) (((smb, sty), (nuop i))::acc) lst
		 else smb_prs i (((smb, sty), nu_smb)::acc) lst end
          val smb_pairs = smb_prs 0 [] ops
          fun abs_op ((nmtyp, varname), thm) = 
 	      Abstract.abstract_op thy nmtyp varname thm
      in foldr abs_op (smb_pairs, thm) end

    fun abs_ops ops thm = abs_ops_isar ops (theory_of_thm thm) thm

    (* This looks wrong -- should we not call abs_ops_isar below ? *)
    fun abs_ops_by_name_isar opnm thy thm = 
	(case Sign.const_type(sign_of_thm thm) opnm of
	     Some optyp=> abs_ops [(opnm, optyp)] thm 
	   | None => error ("No such operation: "^opnm))

    fun abs_ops_by_name opnm thm = abs_ops_by_name_isar opnm (theory_of_thm thm) thm
		 

    (* --- Abstracting types ---------------------------------------------- *)

    fun abs_types_isar typs thy thm = 
	let fun make_typ_vars i [] = []
	      | make_typ_vars i (a::b) = 
		(a, (String.concat ["nutyp", BasisLibrary.Int.toString i])):: 
		make_typ_vars (i+1) b
	    val typ_prs = make_typ_vars 1 typs
	    (* Abstract over a type. 
	     * The sort of the new type variable are the classes collected from
	     * the type-- I am not sure if this is not too restrictive.
	     * --> At least, the get_classes function does not seem to work?!?
	     *)	
	    fun abs_typ ((typ, nm), thm) = 
		Abstract.abstract_type thy typ 
		              (nm, !default_sort) thm
		           (* (nm, Inspect.get_classes_from_typ typ) thm *)
	in foldr abs_typ (typ_prs, thm)
	end

    fun abs_types typs thm = abs_types_isar typs (theory_of_thm thm) thm
	
	
    (* --- Composed tactics ------------------------------------------------ *)
 
    (* Legacy functions: abstract over all thms/opns/types in a theory *)
    fun abs_thms_thy thy thm = 
	let val absthms = Inspect.dep_thms_in [thy] thm
	    val _       = prt_verb ("Abstracting theorems: "^(commas absthms))
	in  abs_thms absthms thm
	end
    fun abs_ops_thy thy thm = 
	let val absops = Inspect.dep_ops_in [thy] thm
	    val _      = prt_verb_ops absops
	in  abs_ops absops thm
	end
    fun abs_types_thy thy thm  = 
	let val abstypes = Inspect.dep_types_in [thy] thm
	    val _        = prt_verb_types thm abstypes
	in  abs_types abstypes thm
	end

    (* Full abstration: does everything *)
    fun full_abs thm = 
	let val thy     = theory_of_thm thm
	    val absthm  = abs_thms_thy thy thm
	    val varthm  = abs_ops_thy thy absthm
	    val typthm  = abs_types_thy thy varthm
	in typthm end

    fun full_abs_old thy = full_abs

    (* Abstract away from one theory to another, by successively abstracting
     * the theorems, operations and types.
     *)

    val abs1 = ref TrueI
    val abs2 = ref TrueI
    val abs3 = ref TrueI

    fun abs_to_thy_with nuthy absops absthm =
	let val depthms = Inspect.dep_thms_not_in nuthy absthm
	    val _       = prt_verb("Abstracting theorems: "^(commas depthms))
	    val _       = abs1 := abs_thms depthms absthm
	    val _       = prt_verb("Abstracted "^(prtlen depthms)^" theorems.")
	    val depops  = absops @ Inspect.dep_ops_not_in nuthy (!abs1)
	    val _       = prt_verb_ops depops
	    val _       = abs2 := abs_ops depops (!abs1)
	    val _       = prt_verb("Abstracted "^(prtlen depops)^" operations.")		
	    val deptypes = Inspect.dep_types_not_in nuthy (!abs2)
	    val _        = prt_verb_types absthm deptypes
	    val _        = abs3 := abs_types deptypes (!abs2)
	    val _        = prt_verb("Abstracted "^(prtlen deptypes)^" types.")
       in  Abstract.abstract_thy nuthy (!abs3)
       end

    fun abs_to_thy thy = abs_to_thy_with thy []

    (* --- Abstract a specific type in the theorem ------------------------ *)

    (* Find a type by name. This can give us a list of results, in particular
     * the many different instances of one type constructor. *)
    fun get_type_by_name nm (types, typ as Type(tnm, args)) = 
	foldl (get_type_by_name nm) 
	      (if nm = tnm then typ ins types else types, args)
      | get_type_by_name nm (types, _) = types

    fun find_types_in_type types (tp as Type(nm, args)) =
	(tp mem types) orelse (exists (find_types_in_type types) args)
      | find_types_in_type types _ = false

    fun find_types_in_ops sg types (opnm, _) =
        (case Sign.const_type sg opnm of
	     Some tp => find_types_in_type types tp 
	   | None => false)

    fun find_ops_in_thms thy ops thnm =
	(case Aux.lookup_thm thy thnm of
	     Some thm => 
		 let val thops = Inspect.get_ops_from_term(prop_of thm, [])
		 in  exists (fn oneop=> oneop mem ops) thops
		 end
           | None => false)

    (* Note: we only abstract the type occuring in the statement of the 
     * theorem, not everywhere in the full proof.
     *)

    fun abs_from_type_isar_with givenops typnm thy thm =
	let val types   = Inspect.get_types_from_term(prop_of thm, [])
            val abstyps = foldl (get_type_by_name typnm) ([], types)
	    val _       = prt_verb_types_isar thy thm abstyps 
            val ops     = Inspect.all_dep_ops thm
	    val sg      = sign_of_thm thm 
	    val moreops = mapfilter 
		(fn opnm=> (case Sign.const_type sg opnm of
				Some t=> Some (opnm, t) 
			      | None=> None)) givenops 
	    val absops  = filter (find_types_in_ops sg abstyps) ops @ moreops
	    val _       = prt_verb_ops absops
	    val thms    = Inspect.all_dep_thms thm
	    val absthms = filter (find_ops_in_thms thy absops) thms
	    val _       = prt_verb ("Abstracting thms: "^(commas absthms))
	in (abs1 := abs_thms_isar absthms thy thm; prt_verb "Abstracted thms";
	    abs2 := abs_ops_isar absops thy (!abs1); prt_verb "Abstracted ops"; 
	    abs3 := abs_types_isar abstyps thy (!abs2); prt_verb "Abstracted types.";
            !abs3)
	end

    fun abs_from_type_with givenops typnm thm = 
        abs_from_type_isar_with givenops typnm (theory_of_thm thm) thm

    val abs_from_type = abs_from_type_with []
    val abs_from_type_isar = abs_from_type_isar_with []


				   
(* ------------------------------------------------------------------------------------ *)

    (* --- Translate along a signature morphism ---------------------------- *)

    fun abs_translate sigma thm = 
	let val trg     = SigMorph.trg sigma
	    val depthms = Inspect.dep_thms_in [SigMorph.src sigma] thm
	    val _       = prt_verb("Abstracting theorems: "^(commas depthms))
	    val absthm  = abs_thms depthms thm
	    val _       = prt_verb("Abstracted "^(prtlen depthms)^" theorems.")
	in  Abstract.abstract_map_thy trg (SigMorph.map_proof sigma) (absthm)
	end

    fun abs_translate_base sigma basemap roadmap =
	let val trg     = SigMorph.trg sigma
	    val _       = Context.set_context (Some trg)
	    val trgnm   = Aux.name_of_thy trg
	    val get_thm_src = Aux.get_thm (SigMorph.src sigma)
	    fun mk_newname thmnm = 
 	           NameSpace.append trgnm (NameSpace.base thmnm) 
(* unqualified:	   trgnm ^ "_" ^ NameSpace.base thmnm *)
	    val thm_name_map = basemap @ (map (fn (thm, _)=> 
					       (thm, mk_newname thm)) roadmap)
            fun map_name nm =
		case assoc_string(thm_name_map, nm) of
		    Some nu => nu | None => error ("No such theorem: "^nm)
	    fun comp_prem(th, premnm) = 
		Aux.comp (thm (map_name premnm)) th
	    fun abs_one (thmnm, premnms) = 
		bind_thm (NameSpace.base (map_name thmnm),
			  let val _ = prt_verb ("Moving "^thmnm^"...")
			      val absthm = abs_translate sigma 
				                         (get_thm_src thmnm)
			  in  foldl comp_prem (absthm, premnms)
			  end)
	in seq abs_one roadmap;
	   prt_verb ("Moved "^(string_of_int (length roadmap))^" theorems.")
	end

end

