(* **************************************************************************
 *
 * Various auxiliary functions.
 *
 * $Source: /home/uniform/awe/repository/src/Extensions/auxiliary.sml,v $
 * $Revision: 1.1 $ of $Date: 2005/06/10 13:45:14 $
 *)

structure Aux =

struct

(* ************************************************************* *)
(*                    Auxiliaries                                *)
(* ************************************************************* *)


(* ***** Operations and logical types of a signature ***** *)

fun get_ops sg = Symtab.keys (#const_tab (Sign.rep_sg sg))
fun get_ltypes sg = Type.logical_types (Sign.tsig_of sg)

(* ******************************************************* *)



fun is_dot #"." = true
  | is_dot   _  = false

fun drop_prefix str = 
    Substring.string (Substring.taker (not o is_dot) (Substring.all str)) 

fun has_dot s = 
    Substring.foldr (fn (c, b)=> b orelse (is_dot c))
                    false (Substring.all s)

fun show_opmap (s, (snm, tnm)) = (s ^ "\nFound: (\"" ^ snm ^ "\" |-> \"" ^ tnm ^ "\")" )
fun show_amb (s, (nm, nmlist)) = (s ^ "\n\"" ^ nm ^ "\" |-> [ " ^ (commas nmlist) ^ " ]")


(* Name of theory : *)
(* fun name_of_thy thy = Sign.name_of (sign_of thy) *)
(* Must be changed because the signs of Isar transitions are nameless! *)
fun name_of_thy thy = PureThy.get_name thy
fun dot_name thy = name_of_thy thy^ "."


(* ********* ThyMorph Auxiliaries ************ *) 

exception THYMORPH_NAME of string;

fun ext_def_split s =
 let val l  = size s;
 in
   if (l <= 4) then None
   else let val es = explode s;
            val (n, e) = splitAt (l - 4, es)
        in
         if (implode e) = "_def" then (Some (implode n))
                                 else None
        end
 end;
     

(* Translate a qualified const-name into either unqualified name with underscore for dot or the desired 
    also _unqualified_ name from the "renamelist" *) 
fun name_transl renamelist name =
 let val l = NameSpace.unpack name;  
 in
   case l of
     [d, n]    => (case (ext_def_split n) of       (* name was qualified, i. e. DaC.f_product_def *)

                    Some nn   => ( case (assoc_string (renamelist, nn)) of  (* extended definition name *)
                                                       Some v => v ^ "_def"
                                                     | None   => d ^ "_" ^ n )
                  | None      => ( case (assoc_string (renamelist, n)) of  (* not extended name *)
                                                       Some v => v 
                                                     | None   => d ^ "_" ^ n ) )

   | [d, n, t] => ( case (assoc_string (renamelist, n)) of  (* name was qualified, i. e. DaC.applyDaC.simps *)
                      Some v => v ^ "_" ^ t
                    | None   => d ^ "_" ^ n ^ "_" ^ t )  (* the second underscore should be a dot
                                                          * but then the problem arises that Isabelle cannot
                                                          * bind the thms with dot in the name, because it
                                                          * seems like a qualified name, i.e.
                                                          * DaC.applyDaC.induct |---> <DaC_applyDaC>.induct
                                                          * where <DaC_applyDaC>.induct is actually correct
                                                          * and _unqualified_, since it is just the name of 
                                                          * the theorem, but for bind it looks like a qualifier.
                                                          *) 

   | [n]       => ( case (ext_def_split n) of                                 (* unqualified name *)
                     Some nn   => ( case (assoc_string (renamelist, nn)) of  (* extended definition name *)
                                                       Some v => v ^ "_def"
                                                     | None   => n )
                   | None      => ( case (assoc_string (renamelist, n)) of    (* not extended name *)
                                                       Some v => v 
                                                     | None   => n ) )

   |  _        => raise THYMORPH_NAME ("exception in name translation: incorrect name " ^ name)
 end; 



(* Check if the term corresponds to a constdef *) 
fun chk_term (t $ _) = chk_term t
  | chk_term (Const ("==", _)) = true
  | chk_term _ = false


(* Returns the operation name of a constdef *)
fun get_opname (Const("Trueprop", _)) = None
  | get_opname (Const("==>", _)) = None
  | get_opname (Const(nm, _)) = Some nm
  | get_opname ((Const ("==", _)) $ r) = get_opname r 
  | get_opname (l $ r) = get_opname l
  | get_opname _ = None


(* Returns a constdef-term with a new opname *)
fun set_opname (Const(nm, t)) name = Const (name, t)
  | set_opname ((Const ("==", t)) $ r) name = ((Const ("==", t)) $ (set_opname r name)) 
  | set_opname (l $ r) name = ((set_opname l name) $ r)

(* Select constdefs from a thms list *)
val select_defs = (filter (chk_term o prop_of o snd)) : ((string * Thm.thm) list -> (string * Thm.thm) list) 

(* Select thms that are not defs (that is axioms or theorems) from a thms list *)
val select_not_defs = (filter (not o chk_term o prop_of o snd)) : ((string * Thm.thm) list -> (string * Thm.thm) list) 
 

(* Check whether a proof term corresponds to a theorem (true) or an axiom (false) *)
fun chk_proof (PAxm _) = false
  | chk_proof (PThm (_, p, _, _)) = chk_proof p
  | chk_proof (p % q) = chk_proof p
  | chk_proof _ = true 


(* Select theorems (not axioms!!!) from a thms list *)
val select_thms = (filter (chk_proof o proof_of o snd)) o select_not_defs  

(* Select axioms (not constdefs!!!) from a thms list *)
val select_axioms = (filter (not o chk_proof o proof_of o snd)) o select_not_defs  
  


(* *** Isar transitions *** *)

fun is_parent t1 t2 = 
 let val sg1 = sign_of t1
 in
  if (Sign.is_draft sg1) then false
   else (let val p = parents_of t2
             val nms = map (Sign.name_of o sign_of) p
         in
          (Sign.name_of sg1) mem nms
         end)
 end

(* b = true  --> returns the ancestors of thy, that are _not_ Isar transitions 
   otherwise --> that are Isar transitions *) 
fun isar_ancestors_of b thy =
  let fun chk_trans_thy thy = 
                             let val name = Sign.name_of (sign_of thy)
                                 val l = explode name
                                 fun chk [] = false
                                   | chk (x::xs) = if x = ":" then true else chk xs
                             in
                               chk l
                             end
      (* Find all ancestors of thy that are Isar-transitions *)
      val trans_ancestors = if b then filter (not o chk_trans_thy) (ancestors_of thy)
                                 else filter chk_trans_thy (ancestors_of thy)
  in
   trans_ancestors
  end


(* *** Accumulate thms (thms_of) from Isar transitions *** *)
fun isar_thms_of thy =
            let val ext_trans_ancestors = thy::(isar_ancestors_of false thy);
                val accu_thms = 
                                let fun f (l, thy) = (l @ (thms_of thy))  
                                in
                                 foldl f ([], ext_trans_ancestors)
                                end
             in
              accu_thms
             end

(* *** get the corresponding definition of opname, iff such exists. Searching in _all_ ancestors *** *)
fun find_def thy opnm =
  let val anc = isar_ancestors_of true thy
      val thy_axs = select_defs (isar_thms_of thy)
      val anc_axs = foldl op@ ([], map axioms_of anc)

      fun find_def_ (opnm, axs) =
       let fun chk [] = None
             | chk ((nm, thm)::axls) = let val term = prop_of thm
                                       in
                                          case (get_opname term) of
                                            Some opnm' => (if (opnm = opnm') then (Some (nm, thm)) else (chk axls))
                                          | None => chk axls
                                       end
       in
        chk axs
       end
  in
   find_def_ (opnm, (thy_axs@anc_axs))
  end


(* Split a list of thms using a condition *)
fun split_list (cond : 'a -> bool) [] = ([], [])
  | split_list cond (x::xs) = let val (l, r) = split_list cond xs
                              in
                               if (cond x) then (x::l, r) else (l, x::r)
                              end
                                           



(* ************************************************************* *)
(*           PROOF RECONSTRUCTION AND THEOREM DERIVATION         *)
(* ************************************************************* *)

(* From the proof checker: look up a theorem.
 * Note that thm, which appears to do the same, but does not need a 
 * theory argument, depends on the elusive local context. 
 *)
fun lookup_thm thy =
  let val tab = foldr Symtab.update
      (flat (map thms_of (thy :: Theory.ancestors_of thy)), 
       Symtab.empty)
  in (fn s => Symtab.lookup (tab, s))
  end

fun get_thm thy s = 
    (case lookup_thm thy s of 
	 None => error ("Unknown theorem " ^ quote s)
       | Some thm => thm)

(* Freezes free variables in proof *)
fun freeze prf = 
      let val (pt2, _)= Proofterm.freeze_thaw_prf (prf) in pt2 end

(* Strip a PThm followed by applications of %;
 * ie. strip_thm PThm(..., prf, ...) % .. % ..  is prf.
 * Note that Proofterm.strip_thm also strips away outer %%, which seems
 * potentially harmful for our uses.
 *)
fun strip_thm prf = 
   (case Proofterm.strip_combt prf of
      (PThm (_, prf', _, _), _) => prf'
          |  _ => prf)

(* Accessor functions on theorem representations *)
fun get_prf thm  = #2 (#der (rep_thm thm))
(* ... and strip top Prf(..) *)
fun proof_of thm = strip_thm (get_prf thm)

fun proof_of' thm = (Thm.name_of_thm thm, proof_of thm)

(* Checks whether a theorem "name" occurs in a proof *)
fun occurs_thm name (PThm(("", _), p, _, _))= 
	if (name= "") then true else (occurs_thm name p)
  | occurs_thm name (PThm((thmnm, _), _, _, _)) = (name= thmnm)
  | occurs_thm name (Abst(_, _, p))= occurs_thm name p
  | occurs_thm name (AbsP(_, _, p))= occurs_thm name p
  | occurs_thm name (p % _)        = occurs_thm name p
  | occurs_thm name (p %% q)       = if (occurs_thm name p) 
					then true else (occurs_thm name q)
  | occurs_thm _    _              = false


(* **************************************************** *
 * Necessary to expand reconstructed proofterm for      *
 * anonymous theorems (""), this is now done in one go  *
 * by functions "reconstruct_prf" and "fullprf_of".     *
 * **************************************************** *)

(* ISABELLE2003: modified argument to     *
 * expand_proof from [""] to [("", None)] *)
fun reconstruct_prf sg prop prf= 
   let val prf' = Reconstruct.reconstruct_proof sg prop prf
   in if (occurs_thm "" prf')
      then (Reconstruct.expand_proof sg [("", None)] prf')
      else prf' 
   end

(* Replaces build_prf 
fun fullprf_of thm = 
      reconstruct_prf (sign_of_thm thm) (prop_of thm) (proof_of thm);
*)


fun reconstruct_prf' sg prop nm prf = 
let val prf' = Reconstruct.reconstruct_proof sg prop prf
    val prf'' = (if (occurs_thm nm prf')
                  then (Reconstruct.expand_proof sg [(nm, None)] prf')
                  else prf')
in
  if (occurs_thm "" prf'') then (Reconstruct.expand_proof sg [("", None)] prf'')
                           else prf''  
end


fun fullprf_of thm = reconstruct_prf' (sign_of_thm thm) (prop_of thm) (Thm.name_of_thm thm) (proof_of thm)

fun fullprf_of' thm = 
      (Thm.name_of_thm thm, fullprf_of thm);


(* *********************************** *)
(* ***         derive_thm          *** *)
(* *********************************** *) 


(* Derive theorem from proofterm in given theory *)
fun derive_thm thy prf = 
     Drule.standard (varifyT (ProofChecker.thm_of_proof thy (freeze prf))) 
        handle e=> let val _ = writeln ("Exception in derive_thm: ")
                    in print_exn e end;



(* Verify that the proof of a theorem can be proofchecked *)
fun check_prf_of thy thm = derive_thm thy (fullprf_of thm)

(* ************************************************************* *)
(*     MAPPING AND FOLDING THEOREMS/AXIOMS IN PROOF TERMS        *)
(* ************************************************************* *)


(* These functions should be in proofterm.sml: *)

fun fold_terms_in_prf f (a, Abst(_, _, prf)) =  fold_terms_in_prf f (a, prf)
  | fold_terms_in_prf f (a, AbsP(_, Some t, prf)) =
    fold_terms_in_prf f (f (t, a), prf)
  | fold_terms_in_prf f (a, AbsP(_, None, prf)) = fold_terms_in_prf f (a, prf)
  | fold_terms_in_prf f (a, PThm(_, prf, t, _)) = f (t, a)
  | fold_terms_in_prf f (a, PAxm(_, t, _))  = f (t, a)
  | fold_terms_in_prf f (a, prf1 %% prf2) = 
    fold_terms_in_prf f (fold_terms_in_prf f (a, prf1), prf2)
  | fold_terms_in_prf f (a, prf % Some t) = f(t, fold_terms_in_prf f (a, prf))
  | fold_terms_in_prf f (a, Hyp t) = f(t, a)
  | fold_terms_in_prf f (a, MinProof ps) = foldl (fold_terms_in_prf f) (a, ps) 
  | fold_terms_in_prf f (a, _)     = a

fun fold_thms_in_prf f (a, Abst(_, _, prf)) = 
    fold_thms_in_prf f (a, prf)
  | fold_thms_in_prf f (a, AbsP(_, _, prf)) = 
    fold_thms_in_prf f (a, prf)
  | fold_thms_in_prf f (a, p as PThm _)  = f (a, p)
  | fold_thms_in_prf f (a, p as PAxm _)  = f (a, p)
  | fold_thms_in_prf f (a, prf1 %% prf2) = 
    fold_thms_in_prf f (fold_thms_in_prf f (a, prf1), prf2)
  | fold_thms_in_prf f (a, prf % _)      = fold_thms_in_prf f (a, prf)
  | fold_thms_in_prf f (a, _)            = a
    
fun map_thms_in_prf f (Abst(s, t, prf)) = Abst(s, t, map_thms_in_prf f prf)
  | map_thms_in_prf f (AbsP(s, t, prf)) = AbsP(s, t, map_thms_in_prf f prf)
  | map_thms_in_prf f (p as PThm _)     = f p
  | map_thms_in_prf f (p as PAxm _)     = f p
  | map_thms_in_prf f (prf1 %% prf2)    = map_thms_in_prf f prf1 %%
                                          map_thms_in_prf f prf2
  | map_thms_in_prf f (prf % t)         = map_thms_in_prf f prf % t
  | map_thms_in_prf f q                 = q


(* ************************************************************* *)
(*                  PROOF PRINTING FUNCTIONS                     *)
(* ************************************************************* *)

(* Print proof *)
fun print_prf sg prf = 
      writeln (Pretty.string_of 
       (ProofSyntax.pretty_proof sg (strip_thm prf)))

(* Print proof of a theorem *)
fun print_prf_thm thm = 
      print_prf (sign_of_thm thm) (get_prf thm)

(* ************************************************************* *)
(*                   FORWARD RESOLUTION                          *)
(* ************************************************************* *)

(* This is the forward resolution that we use to discharge proof obligations,
 * i.e. theorems lifted into the premises of the abstracted theorem: we 
 * first make all variables in the t into meta-quantified variables, 
 * because they will be in that form in the premises, then resolve with COMP 
 * (not RS).
 * We can't use RS since this lifts over the assumptions of the state as 
 * well, which in the case of induction is harmful.
 * On the other hand, we do want to lift over the params, but there seems
 * to be no tactic which does only that. The one below is a good guess, 
 * since we can rely on _all_ variables in the proof obligation to be bound.
 *)

fun comp  t s = (Drule.forall_intr_vars t) COMP s
fun comp' n s = (Drule.forall_intr_vars (thm n)) COMP s

(* Simplify premises (in HOL). See simple-example.sml how to use this. *)
fun simp t  = Seq.hd (ALLGOALS (full_simp_tac (simpset_of (theory "HOL"))) t)
fun simpi n t =  Seq.hd (full_simp_tac (simpset_of (theory "HOL")) n t)

end
