(* **************************************************************************
 *
 * Dependency analysis.
 *
 * $Source: /home/uniform/awe/repository/src/Extensions/dep_ana.sml,v $
 * $Revision: 1.1 $ of $Date: 2005/06/10 13:45:14 $
 *)

signature DEPANA_SIG =
sig  type DG

     val get_dg' : theory list -> (string * Thm.thm) list -> DG
     val get_dg : theory-> DG
 
     val sort_dep : DG -> string list-> string list
     val bfs      : DG -> string list-> string list
     val saturate : DG -> string list-> string list->
                      (string* string list) list
     val trans_depends_on_filter : DG -> string -> string list -> string list

end

structure DepAna : DEPANA_SIG  = 
  struct
      (* Debugging. *)
      val debug = ref true;
      fun dwriteln str = if !debug then writeln str else () 

      local open BasisLibrary  in

     (* We say a theorem t1 _depends on_ t2 iff t2 occurs in the proof
      * object of t1; i.e. the proof of t1 needs t2. 
      * The _preconditions_ of a theorem th1 is the list of all theorems 
      * on which th1 depends.
      * A theorem th1 is _derivable_ from a list of theorems T if 
      * the dependents of th1 are contained in T.
      * The _dependents_  of a theorem th1 are the 
      * theorems which depend on th1. So, to prove any of the dependents
      * of th1 it is necessary (but not sufficient) to prove th1.
      *)

     (* The dependency graph -- for each theorem, the list of theorems
      * which depend on it. 
      * Could be made a better structure, e.g. a symtab.
      *)
     datatype DG = DG of (string* string list) list 


     (* Calculate the dependency graph of a theory. 
      * Only contains theorems from this theory. 
      *)
     fun get_dg' thys thms_list = 
	 let 
	     (* Get dependencies for a given proof. 
	      * Axioms depend on themselves, which would bugger up our 
	      * analysis, so we remove that. *)
	     fun get_deps (name, prf) =  
		 (case Proofterm.strip_combt prf of
				     (PAxm _, _) => []
				   | _ => Inspect.dep_thms_in_prf thys (name, prf))


	     (* Note the rev. Order is important here, as theorems are
	      * abstracted in the reverse order of their appearance
	      * in the proof term. *)
             val res_list = (map (apsnd (rev o get_deps o Aux.proof_of')) thms_list)

	 in  DG res_list
	 end

     fun get_dg thy =
      let val thms_list = Aux.select_not_defs (Aux.isar_thms_of thy)
      in
       get_dg' [thy] thms_list
      end

     (* All theorems on which this depends, i.e. 
      * preconditions g th1 = [th | th1 depends_on th];
      * iow. all theorems occuring in this theorems proof object. *)
     fun preconditions (DG g) th = 
	 case assoc_string(g, th) of 
	       Some deps => deps 
	     | None => []  (* or error ("No such theorem: "^th)*) (* or [] *)

     (* True iff th1 depends on th2. *)
     fun depends_on dg th1 th2 = th2 mem preconditions dg th1

     (* True i th is directly provable from thms, i.e. all thms which
      * occur in the proof object of th are in thms. *)
     fun derivable dg th thms = 
	 preconditions dg th subset_string thms 

     (* Dependents of this theorem, i.e. theorems relying on
      * this (possibly amongst others) to be proven.
      *)
     fun dependents (DG g) thnm = 
	 map fst (filter (fn (_, deps)=> thnm mem_string deps) g)

     (* Dependency as an order. th1 LESS th2 if th1 does not depends on th2. 
      * In a list sorted ascending in this order, a theorem will not depend
      * on the ones before it. *)
     (* Nice try, but does not work: this order is not transitive! 	  
     fun not_dep_ord dg (th1, th2) = 
	 if th1 = th2 then EQUAL 
	 else if not (depends_on dg th1 th2) then LESS else GREATER
      *)

     (* A list [t1...tn] of theorems is ordered by dependency iff 
      * !i,j. i< j ==> not (x_i depends on x_j) 
      * In other words, earlier theorems do not depend on later ones. *)     


(* _transitive closure_ of "depends_on" relation *) 
  fun trans_depends_on dg t1 t2 =
   let fun cls t = 
        let val p = preconditions dg t
        in 
          t::(foldl op@ ([], map cls p))
        end
   in
     t2 mem (tl (cls t1))
   end;

 fun trans_depends_on_filter dg thm thm_list =
   let fun cls t = 
        let val p = preconditions dg t
        in 
          t::(foldl op@ ([], map cls p))
        end
       val thm_cls = tl(cls thm)
   in
     filter (fn t => t mem thm_cls) thm_list
   end;


     
     (* Insert a theorem into a list of theorems ordered by transitive dependency,
      * preserving that order. *)
     fun ins_dep_trans dg x []      = [x]
       | ins_dep_trans dg x (rest as y::ys) = 
	    (if x = y then rest 
	     else (if (trans_depends_on dg y x) then (x::rest)
	                                      else (y::(ins_dep_trans dg x ys))))
     

     (* Sort by dependency.*)
     fun sort_dep dg xs = foldl (fn (l, x)=> ins_dep_trans dg x l) ([], xs)


     (* Insert a theorem into a list of theorems ordered by dependency,
      * preserving that order. *)
     fun ins_dep dg x []      = [x]
       | ins_dep dg x (rest as y::ys) = 
	     if x = y then rest 
	     else if depends_on dg y x then x::rest
	                               else y::ins_dep dg x ys

     (* Breadth-first search of all preconditions starting with vs,
      * return a list ordered by dependency. *)
     fun bfs dg vs =
	 let fun bfs' vs []      = vs
	       | bfs' vs (x::xs) =
		   if x mem vs then bfs' vs xs
		   else bfs' (ins_dep dg x vs) (preconditions dg x@xs)
	 in  bfs' [] vs
	 end


     (* Given a list of theorems, close these under derivability, i.e.
      * add in all theorems which we can derive from these.
      *
      * Roughly, this works as follows: we always have a list of candidates,
      * and a current base. When the list of candidates are empty, we are done.
      * Otherwise, insert the first candidate into base. When doing this,
      * we consider all dependents of this candidate as potential new can-
      * didates; from these potential candidates, all that are derivable from
      * the current base become proper candidates.
      *)
     fun close_der dg base = 
	 let (* Filter out candidates from potential new candidates. *)
	     fun find_cands base [] vs      = vs
	       | find_cands base (c::cs) vs = 
		     (* if c is derivable, add it to list of derivable thms *)
		     if derivable dg c base
			 then find_cands (c::base) cs (c::vs)
		     else find_cands base cs vs
             (* Recursively add all dependents to list of candidates.
	      * First param are the candidates, second is the current base. *)
	     fun close [] vs      = vs
	       | close (x::xs) vs =
		 let (* Potential new candidates *)
		     val cands  = sort_dep dg (dependents dg x)
		     (* Insert x into current base *)
		     val nuv = x ins_string vs 
	             (* Filter out candidates from potential candidates *)
		     val nux = find_cands nuv cands []
		     val _ = dwriteln ("Considered thm "^ x^", "^
				      (Int.toString (length nux))^" new to visit, "^(Int.toString (length nuv)^" visited"))
		 in  close (nux @ xs) nuv
		 end
	 in  close base base
	 end

     fun saturate dg base thms = 
	 let (* These are all theorems on which the thms are depending,
	      * i.e. all we need to prove before we can prove thms,
	      * minus the ones on which the base is depending.
	      *)
	     val alldeps = bfs dg thms \\ bfs dg base
	     val _ = dwriteln ("Depending on "^Int.toString (length alldeps) 
			       ^" thms.");
     	     (* These are all theorems derivable from the base,
	      * i.e. given the base we can derive these.
	      *)
	     val allders = close_der dg base
	     val _ = dwriteln (Int.toString (length alldeps) ^
			       " derivable thms.");
	     (* Now, all theorems which are both in alldep and allders are 
	      * the ones we can derive from the base. If we have to throw
	      * out non-derivable theorems, give a warning.
	      * Note that if a theorem is not derivable, neither will be its
	      * dependents, so the remaining list will be in dependency order.
	      *)
	     val derdeps = if alldeps subset_string allders then alldeps
		           else (warning ("Omitted underivable theorems: "^
					  commas (alldeps \\ allders));
				 alldeps inter_string allders)
	 in  map (fn th=>(th, preconditions dg th)) derdeps
	 end		      

     end (* local open BasisLibrary *)
  end


