(* ***********************************************************************
 *
 * Signature morphisms. 
 *
 * $Source: /home/uniform/awe/repository/src/Extensions/sig_morph.sml,v $
 * $Revision: 1.1 $ of $Date: 2005/06/10 13:45:14 $
 *)

(* ************************************************************************
 
 A Signature morphisms is given by a source signature, a target signature,
 and a map of type constructors and operations such that arity and source
 and target are preserved. 

 If the source signature is not ancestor of the target, then we require 
 at least logical types, logical ops from source, which are
 not in the intersection of source and target signatures, to be mapped, 
 otherwise no explicit mappings are required: the signature morphism is 
 done by inclusion. 
 On the other hand, logical types and ops, that are in the intersection
 of signatures may be mapped by a signature morphism, but only by identity.
 So it does not make sence to map them explicitly.

 The checks implemented here are slightly too optimistic: we allow the
 translated target type of an operation to be the _instance_ of the
 translated source type.  This allows for type class operations,
 e.g. 0 mapped to Zero in nat-> bNat, but users may abuse this,
 e.g. by mapping polymorphic equality to bPlus.

 To fix this, we would need to find out whether an operation (such as 0 or +)
 is declared as instance of a type class in this signature; I can't seem
 to be able to extract such information from Isabelle's signature or thy
 datatype.

 *************************************************************************** *)

signature SIGMORPH =
sig
  type sig_morph

  (* A type constructor can be mapped to another type constructor; 
   * a monomorphic type can also be mapped to a composed type. *)
  datatype type_map = Simple of string
                    | Comp   of Term.typ

  type 'a map_rep 

  type ThyData
  val get_thydata : Theory.theory -> ThyData
  val get_local_sigmorph_data : ThyData -> ThyData -> 
 (string list * (string * string Library.option) list * (string * string Library.option) list * string list * string list)


  val src  : sig_morph -> theory
  val trg  : sig_morph -> theory
  val opMap : sig_morph -> (string * string) list
  val tpMap : sig_morph -> (string * type_map) list
  val local_src_ltypes : sig_morph -> string list
  val local_src_ops_tab : sig_morph -> (string * string Library.option) list
  val local_trg_ops_tab : sig_morph -> (string * string Library.option) list
  val ltypes_inter : sig_morph -> string list
  val ops_inter    : sig_morph -> string list


 (* Use a signature morphism to map types, terms and proof terms. *)
  val map_type  : sig_morph-> Term.typ-> Term.typ
  val map_term  : sig_morph-> term-> term
  val map_proof : sig_morph-> Proofterm.proof-> Proofterm.proof 


 (* Construct a signature morphism, checking invariants. *)
  val make : theory-> theory-> 
             (string* type_map) list-> (string* string) list-> sig_morph

 (* Extend a signature morphism, retaining invariants. *)
  val ext  : sig_morph-> (string* type_map) list-> (string* string) list->
             sig_morph



 (* Try to construct a signature morphism by matching the operations from source
    in target, using the given type map. Returns a tuple (a, b, c) where
    - a is the optional signature morphism. 
    - b is the list of unmatched operations and c is the list of ambiguous 
      mappings with possible images, which can be resolved by operation map (4-th argument).

    Note: a = None iff b or c is not empty
  *) 
  val try_sigmorph : theory -> theory -> (string* type_map) list-> (string* string) list 
                     -> (sig_morph Library.option * string list * ((string * (string list)) list))



 (* Extend the codomain of a signature morphism with a new target theory, that
    has to be a descedant of the old target theory. Retaining invariants  *)
  val cod_ext : sig_morph -> Theory.theory -> sig_morph 
  
 (* Extend the domain of a signature morphism with a new source theory, that
    has to be a descedant of the old source theory. Retaining invariants.
    The second argument should be used for type map extension for new logical types, which
    appear just in the new source.   *)
  val dom_ext : sig_morph -> (string * type_map) list -> (string * string) list -> Theory.theory -> sig_morph  



  (* constdefs splitting *)
  val split_src_cdefs : sig_morph -> 
                  ((string * string * (Thm.thm Library.option)) list) * ((string * string * (Thm.thm Library.option)) list)

  val split_trg_cdefs : sig_morph -> 
                  ((string * string * (Thm.thm Library.option)) list) * ((string * string * (Thm.thm Library.option)) list)     
  val is_inclusion : sig_morph -> bool
  val make_inclusion : Theory.theory -> Theory.theory -> sig_morph

  exception SIGMORPH of string

end



structure SigMorph : SIGMORPH =

struct

   datatype type_map = Simple of string
                     | Comp of Term.typ

   type 'a map_rep = (string* 'a) list

   datatype sig_morph = SigMorph of { src : theory
				    , trg : theory
				    , typeMap : type_map map_rep
				    , opMap   : string map_rep
                                    , local_src_ltypes : string list
                                    , local_src_ops_tab : (string * string Library.option) list
                                    , local_trg_ops_tab : (string * string Library.option) list
                                    , ltypes_inter : string list
                                    , ops_inter : string list
				    }

   datatype SigData  = SigData of { sg     : Sign.sg
				  , types  : string list
				  , ops    : string list
				  }

   datatype ThyData = ThyData of { sigdata : SigData
                                 , thy    : Theory.theory
                                 , anc_names : string list
                                 }

   exception SIGMORPH of string


   fun get_ops sg = Symtab.keys (#const_tab (Sign.rep_sg sg))
   fun get_ltypes sg = Type.logical_types (Sign.tsig_of sg)

   (* Get all types and ops living in this signature. *)
   fun get_sigdata sg =
       SigData {sg = sg
		, types = get_ltypes sg
		, ops = get_ops sg
		}

   fun get_thydata thy =
    let val sg = sign_of thy
        val anc_names = map Aux.name_of_thy (Aux.isar_ancestors_of true thy)
    in
     ThyData {sigdata = get_sigdata sg
              , thy = thy
              , anc_names = anc_names
             }
    end 


   fun find_def (opnm, axs) =
          let fun chk [] i = None
                | chk ((nm, thm)::axls) i = let val term = prop_of thm
                                            in
                                             case (Aux.get_opname term) of
                                               Some opnm' => (if (opnm = opnm') then (Some (nm, i)) else (chk axls (i+1)))
                                             | None => chk axls i
                                            end
          in
           chk axs 0
          end

   fun insert (x, y, z as (Some i)) [] = [(x, y, z)]
     | insert (x, y, z as (Some i)) (ls as ((a, b, c)::l)) =
        case c of
         Some i' => (if (i <= i') then ((x, y, z)::ls)
                     else ((a, b, c)::(insert (x, y, z)) l))
        | None => ((a, b, c)::(insert (x, y, z) l))
  


   (* The order of defs, given by thms_of, holds also for an ops_tab, provided by the function insert *)
   fun make_ops_tab' [] axs = []
     | make_ops_tab' (opnm::nms) axs = case (find_def (opnm, axs)) of
                                         Some (nm, i) => insert (opnm, Some nm, Some i) (make_ops_tab' nms axs) 
                                       | None => (opnm, None, None)::(make_ops_tab' nms axs)

   fun make_ops_tab opnms axs = let val tab' = make_ops_tab' opnms axs
                                in
                                 map (fn (a, b, c) => (a, b)) tab'
                                end

   fun update_ops_tab [] axs = []
     | update_ops_tab ((opnm, opt)::nms) axs = case opt of
                                                Some nm => (opnm, opt)::(update_ops_tab nms axs)
                                               | None => ( case (find_def (opnm, axs)) of
                                                            Some (nm, i) => (opnm, Some nm)::(update_ops_tab nms axs)
                                                            | None => (opnm, None)::(update_ops_tab nms axs) ) 

   fun get_local_sigmorph_data 
      (ThyData{sigdata = (SigData{types=src_ltypes, ops=src_ops, ...}), anc_names = src_anc, thy = src, ...}) 
      (ThyData{sigdata = (SigData{types=trg_ltypes, ops=trg_ops, ...}), anc_names = trg_anc, thy = trg, ...})  =

     let val anc_inter = src_anc inter trg_anc
         val local_src_anc = src_anc \\ anc_inter
         val local_trg_anc = trg_anc \\ anc_inter
         val ltypes_inter = src_ltypes inter trg_ltypes
         val ops_inter = src_ops inter trg_ops
         val local_src_ltypes = src_ltypes \\ ltypes_inter
         val local_src_ops    = src_ops \\ ops_inter
         val local_trg_ops    = trg_ops \\ ops_inter

         val src_anc_thys = map theory local_src_anc
         val trg_anc_thys = map theory local_trg_anc
         val src_axs = Aux.select_defs (Aux.isar_thms_of src)
         val src_anc_axs = foldl op@ ([], map (Aux.select_defs o thms_of) src_anc_thys)
         val trg_axs = Aux.select_defs (Aux.isar_thms_of trg)
         val trg_anc_axs = foldl op@ ([], map (Aux.select_defs o thms_of) trg_anc_thys)

         val local_src_ops_tab = make_ops_tab local_src_ops (src_axs @ src_anc_axs) 
         val local_trg_ops_tab = make_ops_tab local_trg_ops (trg_axs @ trg_anc_axs) 
     in
      (local_src_ltypes, local_src_ops_tab, local_trg_ops_tab, ltypes_inter, ops_inter)  
     end 


   fun src (SigMorph{src,...})= src
   fun trg (SigMorph{trg,...})= trg
   fun opMap (SigMorph{opMap,...})= opMap
   fun tpMap (SigMorph{typeMap,...}) = typeMap
   fun local_src_ltypes (SigMorph{local_src_ltypes,...}) = local_src_ltypes
   fun local_src_ops_tab (SigMorph{local_src_ops_tab,...}) = local_src_ops_tab
   fun local_trg_ops_tab (SigMorph{local_trg_ops_tab,...}) = local_trg_ops_tab  
   fun ltypes_inter (SigMorph{ltypes_inter,...}) = ltypes_inter
   fun ops_inter (SigMorph{ops_inter,...}) = ops_inter


(* internal updates *)
   fun set_src (SigMorph{src, trg, typeMap, opMap, local_src_ltypes, local_src_ops_tab, local_trg_ops_tab,
                         ltypes_inter, ops_inter}) nsrc =
       (SigMorph{src=nsrc, trg=trg, typeMap=typeMap, opMap=opMap, 
                 local_src_ltypes=local_src_ltypes,
                 local_src_ops_tab=local_src_ops_tab,
                 local_trg_ops_tab=local_trg_ops_tab,
                 ltypes_inter=ltypes_inter, ops_inter=ops_inter})
   fun set_trg (SigMorph{src, trg, typeMap, opMap, local_src_ltypes, local_src_ops_tab, local_trg_ops_tab,
                         ltypes_inter, ops_inter}) ntrg =
       (SigMorph{src=src, trg=ntrg, typeMap=typeMap, opMap=opMap, 
                 local_src_ltypes=local_src_ltypes,
                 local_src_ops_tab=local_src_ops_tab,
                 local_trg_ops_tab=local_trg_ops_tab,
                 ltypes_inter=ltypes_inter, ops_inter=ops_inter})
   fun set_ltypes (SigMorph{src, trg, typeMap, opMap, local_src_ltypes, local_src_ops_tab, local_trg_ops_tab,
                            ltypes_inter, ops_inter}) nltypes =
       (SigMorph{src=src, trg=trg, typeMap=typeMap, opMap=opMap, 
                 local_src_ltypes=nltypes,
                 local_src_ops_tab=local_src_ops_tab,
                 local_trg_ops_tab=local_trg_ops_tab,
                 ltypes_inter=ltypes_inter, ops_inter=ops_inter})
   fun set_src_ops_tab (SigMorph{src, trg, typeMap, opMap, local_src_ltypes, local_src_ops_tab, local_trg_ops_tab,
                                 ltypes_inter, ops_inter}) nops_tab =
       (SigMorph{src=src, trg=trg, typeMap=typeMap, opMap=opMap, 
                 local_src_ltypes=local_src_ltypes,
                 local_src_ops_tab=nops_tab,
                 local_trg_ops_tab=local_trg_ops_tab,
                 ltypes_inter=ltypes_inter, ops_inter=ops_inter})
   fun set_trg_ops_tab (SigMorph{src, trg, typeMap, opMap, local_src_ltypes, local_src_ops_tab, local_trg_ops_tab,
                                 ltypes_inter, ops_inter}) nops_tab =
       (SigMorph{src=src, trg=trg, typeMap=typeMap, opMap=opMap, 
                 local_src_ltypes=local_src_ltypes,
                 local_src_ops_tab=local_src_ops_tab,
                 local_trg_ops_tab=nops_tab,
                 ltypes_inter=ltypes_inter, ops_inter=ops_inter})
 




  
(* ************* Mappings provided by signature morphisms ********** *) 

   fun map_tp tmap (Type(nm, ts)) = 
       (case  assoc_string(tmap, nm) of
	    Some (Simple to) => Type (to, map (map_tp tmap) ts)
          | Some (Comp to)   => to
          | None             => Type (nm, map (map_tp tmap) ts))
     | map_tp _    var = var (* type variables stay as they are *)

   fun map_t (tmap, opmap) t =
       let fun mapt (Const(c, typ)) = 
	          (case assoc_string(opmap, c) of 
		       Some to => Const(to, map_tp tmap typ)
		     | None    => Const(c, map_tp tmap typ)) 
		       (* Need to map types, e.g. for polymorphic consts's *)
	     | mapt (Free(nm, tp)) = Free(nm, map_tp tmap tp)
	     | mapt (Var(ix, tp))  = Var(ix, map_tp tmap tp)
	     | mapt (b as Bound _) = b
	     | mapt (Abs(nm, tp, t)) = Abs(nm, map_tp tmap tp, mapt t)
	     | mapt (s$t)          = mapt s $ mapt t
       in  mapt t
       end

 


   fun map_type (SigMorph{typeMap, ...}) = map_tp typeMap

   fun map_term (SigMorph{typeMap, opMap, ...}) = map_t (typeMap, opMap)

   fun map_proof sg = (Proofterm.map_proof_terms (map_term sg) (map_type sg))


(* *******  Construction of signature morphisms ********* *) 


   fun type_ar(sg, tycon) = 
       Symtab.lookup(#tycons (Type.rep_tsig (Sign.tsig_of sg)), tycon)

   (* Check a type mapping (from, to)
    * - from in domain, to in range
    * - if to is just a type constructor, arity the same.  *)
   fun chck_typ (SigData{types=src_types, sg= src_sg, ...},
		 SigData{types=trg_types, sg= trg_sg, ...}) (from, to) = 
       if not (from mem src_types) then
	   Some ("Type "^from^": not in domain signature")
       else (case to of
	       Simple tc =>
		 (if not (tc mem trg_types) then
		      Some ("Type "^tc^": not in range signature")
		  else if type_ar(src_sg, from) <> type_ar(trg_sg, tc) then
		      Some ("Type "^ from^": arity not preserved") 
		       else None)
             | Comp tp => 
		  (if type_ar(src_sg, from) <> Some 0 then
		       Some ("Non-monomorphic type "^from^" mapped to \
			     \composed type.") 
		   else (let fun chk_in_trg (Type (nm, ts)) = (nm mem trg_types) andalso 
                                             (foldl (fn (a, b) => (a andalso b)) (true, (map chk_in_trg ts)))
                               | chk_in_trg var = true 
                             val r = chk_in_trg tp
                         in
                          if r then None
                          else Some ("Type " ^ Sign.string_of_typ trg_sg tp ^ " not in target signature.")
                         end)))

   (* Check an operation mapping (from, to):
    * - from in domain, to in range;
    * - type of from, translated with type map, is the same
    *   as to (dodgy-- free variables should be equal up to 
    *   \alpha-conversion, e.g. 'a list and 'b list?    *)
   fun chck_op (SigData{ops=src_ops, sg= src_sg, ...},
		SigData{ops=trg_ops, sg= trg_sg, ...}) typemap (from, to) =
       if is_none (Sign.const_type src_sg from) then
        (* was: not (from mem src_ops), but that excludes type-class opns'
	 * from being mapped (e.g. 0) *) 
	   Some ("Opn "^from^": not in domain signature")
       else if not (to mem trg_ops) then
	   Some ("Opn "^to^": not in range signature")
       else let (* from and to will always have a type *)
		val Some ft= Sign.const_type src_sg from
		val Some tt= Sign.const_type trg_sg to
	    in  if  not (Type.typ_instance(Sign.tsig_of trg_sg, tt, 
					   map_tp typemap ft)) then
		    (* was: map_tp typemap ft <> tt 
		     * but that precludes class operations, see above,
		     * so we just check that the target type is an
		     * _instance_ of the mapped source type. *) 
		  Some ("Opn "^from^": type not preserved")
	        else None
	    end

   (* Check if sources are unique. *)
   fun chck_uniq err = 
       map (fn s=> err^": ambiguous mapping.") o duplicates o map fst


   (* Construct a signature morphism, and check internal invariants: chck_uniq, chck_typ, chck_op.
    *)
   fun make_i src trg typemap opmap src_ltypes src_ops_tab trg_ops_tab ltypes_inter ops_inter =
       let val data = (get_sigdata (sign_of src), get_sigdata (sign_of trg))                   
       in  case chck_uniq "Type" typemap @ chck_uniq "Opn" opmap of
	     (es as _ :: _) => raise (SIGMORPH (commas es))
           | [] =>
           (case mapfilter (chck_typ data) typemap of
	      (ts as _ :: _) => raise (SIGMORPH (commas ts))
	    | [] => (* chck_op only makes sense if chck_typ successful? *)
	    (case mapfilter (chck_op data typemap) opmap of
	       (os as _ :: _) => raise (SIGMORPH (commas os))
	       | [] => (SigMorph {src= src, trg= trg, typeMap= typemap, opMap= opmap,
                                  local_src_ltypes = src_ltypes, local_src_ops_tab = src_ops_tab,
                                  local_trg_ops_tab = trg_ops_tab, ltypes_inter = ltypes_inter, ops_inter = ops_inter})
                               
           ))
       end


(* *** Type- and opmap extensions *** *)

   (* Extend a signature morphism, retaining the internal invariants. *)
   fun ext (SigMorph{src, trg, typeMap, opMap, local_src_ltypes, local_src_ops_tab, local_trg_ops_tab, 
                     ltypes_inter, ops_inter}) nutypes nuops =
       make_i src trg (foldl overwrite (typeMap, nutypes))
                      (foldl overwrite (opMap, nuops)) local_src_ltypes local_src_ops_tab local_trg_ops_tab ltypes_inter ops_inter




(* ******** Construction of signature morphisms ****************** *)

   (* Auxiliary *)
         fun chck_types_inter mp [] = []
           | chck_types_inter mp (x::xs) = case (assoc_string(mp, x)) of
                                            Some i => (case i of
                                                        Simple n => (if (x = n) then chck_types_inter mp xs
                                                                     else x::(chck_types_inter mp xs))
                                                       | Comp t => x::(chck_types_inter mp xs))
                                          | None => chck_types_inter mp xs

         fun chck_ops_inter mp [] = []
           | chck_ops_inter mp (x::xs) = case (assoc_string(mp, x)) of
                                          Some i => (if (x = i) then chck_ops_inter mp xs
                                                     else x::(chck_ops_inter mp xs))
                                         | None => chck_ops_inter mp xs

         fun chck_all_src_mapped obj mp =
           let val dom = map fst mp
               val diff = obj \\ dom
           in
            if (diff = []) then None else (Some diff)
           end

         fun get_lops [] = []
           | get_lops ((opnm, opt)::tab) = case opt of
                                                Some nm => get_lops tab
                                               | None => opnm::(get_lops tab)
 
         fun ext_ops sg = 
          map (fn opnm => ( let val tp = (case (Sign.const_type sg opnm) of
                                            Some t => t
                                          | None => raise SIGMORPH "Unexpected exception in match_ops: Sign.const_type")
                            in (opnm, tp) end )) 
                  

 (* Construct a signature morphism, checking _all_ invariants, 
    i.e. those from make_i and additional:
    - all local source logical types and logical ops have to be mapped
    - if in the type map (op map) an element from intersection is mapped then it is mapped to itself
 *)
     
   fun make src trg typemap opmap =
     let val src_data = get_thydata src
         val trg_data = get_thydata trg
         val (local_src_ltypes, local_src_ops_tab, local_trg_ops_tab, ltypes_inter, ops_inter) = 
            get_local_sigmorph_data src_data trg_data

         val src_lops = get_lops local_src_ops_tab

         val t_inter_chck = case (chck_types_inter typemap ltypes_inter) of
                             [] => None
                            | l => Some ("\nFollowing logical type(s) have to be mapped by identity: " ^ (commas l))
         val o_inter_chck = case (chck_ops_inter opmap ops_inter) of
                             [] => None
                            | l => Some ("\nFollowing const(s) have to be mapped by identity: " ^ (commas l))
         val src_ltypes_chck = case (chck_all_src_mapped local_src_ltypes typemap) of
                                 None => None
                               | Some t_diff => Some ("\nFollowing logical type(s) have to be mapped: " ^ (commas t_diff))
         val src_lops_chck = case (chck_all_src_mapped src_lops opmap) of
                               None => None  
                             | Some o_diff => Some ("\nFollowing logical const(s) have to be mapped: " ^ (commas o_diff))
 
         val chck_res = mapfilter (fn x => x) [t_inter_chck, o_inter_chck, src_ltypes_chck, src_lops_chck]

     in
      case chck_res of
        [] => make_i src trg typemap opmap local_src_ltypes local_src_ops_tab local_trg_ops_tab ltypes_inter ops_inter
       | err => (let val _ = writeln (commas err)
                 in
                   raise SIGMORPH "Signature morphism construction failed."
                 end)
     end




(* **** Try to match ops by given type map **** *)

fun match_ops sigma ext_src_lops ext_trg_ops opMap =
 let fun match_op (((l1, l2, l3), sigma, tops), (opnm, tp)) =
       let val mtp = map_type sigma tp
           fun search typ1 [] = []
             | search typ1 ((nm, typ2)::ops) = (if (typ1 = typ2) then (nm :: (search typ1 ops)) 
                                                                         else (search typ1 ops))  
           fun remove nm [] = []
             | remove nm ((n, t)::xs) = (if (nm = n) then xs else ((n, t)::(remove nm xs)))
  
           val res = search mtp tops
       in
        case res of
         [n] => ( ((opnm, n)::l1, l2, l3), sigma, (remove n tops) )
        | [] => ( (l1, opnm::l2, l3), sigma, tops )
        | _  => ( case (assoc_string (opMap, opnm)) of
                    None => ( (l1, l2, (opnm, res)::l3), sigma, tops )
                  | Some tnm => (if (tnm mem res) then ( ((opnm, tnm)::l1, l2, l3), sigma, tops )
                                                  else ( (l1, l2, (opnm, res)::l3), sigma, tops ))
                )
       end
 in
  (fn ( (l1, l2, l3), s, t) => (l1, l2, l3) ) ( foldl match_op ( (([],[],[]), sigma, ext_trg_ops), ext_src_lops ))
 end
       

(* *** Try to construct a signature morphism by given type map *** *)

fun try_sigmorph src trg tpMap opMap =                       
  let val (src_data as (ThyData{sigdata=(SigData{sg=sg_src, ...}), ...})) = get_thydata src
      val (trg_data as (ThyData{sigdata=(SigData{sg=sg_trg, ops=trg_ops, ...}), ...})) = get_thydata trg

      val (local_src_ltypes, local_src_ops_tab, local_trg_ops_tab, ltypes_inter, ops_inter) = 
            get_local_sigmorph_data src_data trg_data

      val chk1 = case (chck_all_src_mapped local_src_ltypes tpMap) of
                  None => None
                | Some diff => Some ("\nFollowing logical type(s) from the domain theory are not mapped: "
                                                ^ (commas diff) )
      val chk2 = case (chck_types_inter tpMap ltypes_inter) of
                             [] => None
                            | l => Some ("\nFollowing logical type(s) have to be mapped by identity: " ^ (commas l))
      val chks = mapfilter (fn x => x) [chk1, chk2]

  in
   case chks of
    [] =>
      (let val src_lops = get_lops local_src_ops_tab
           val ext_src_lops = ext_ops sg_src src_lops
           val ext_trg_ops  = ext_ops sg_trg trg_ops
           val tsigma = 
            make_i src trg tpMap [] local_src_ltypes local_src_ops_tab local_trg_ops_tab ltypes_inter ops_inter     
           val (l1, l2, l3) = match_ops tsigma ext_src_lops ext_trg_ops opMap
           val _ = writeln (foldl Aux.show_opmap ("", l1))  
       in
        if ((l2 = []) andalso (l3 = [])) then (Some (ext tsigma [] l1), l2, l3)
        else (None, l2, l3)
       end)
   | err => (let val _ = writeln (commas err)
             in
               raise SIGMORPH "Signature morphism construction failed."
             end)
  end            
       



(* *** Domain and codomain extensions *** *)

(* *********************** Extend the codomain of a signature morphism  ********************* *)
    fun cod_ext (sigma as (SigMorph{src, trg, typeMap, opMap, local_src_ltypes, local_src_ops_tab, local_trg_ops_tab,
                                    ltypes_inter, ops_inter})) ntrg = 
       let val n1 = Aux.name_of_thy trg
           val n2 = Aux.name_of_thy ntrg
           fun get_ops_diff thy thy' = 
            let val sg = sign_of thy
                val sg' = sign_of thy'
            in
             (get_ops sg') \\ (get_ops sg)
            end
       in 
        if (n1 = n2) then 
          (let val ops_diff = get_ops_diff trg ntrg
               val defs = Aux.select_defs (Aux.isar_thms_of ntrg)
               val ops_tab_ext = make_ops_tab ops_diff defs
               val upd_ops_tab = update_ops_tab local_trg_ops_tab defs 
               val sigma' = set_trg sigma ntrg
               val sigma'' = set_trg_ops_tab sigma' (upd_ops_tab @ ops_tab_ext)
           in
            sigma''
           end) 
        else (if (Theory.subthy (trg, ntrg)) then 
               (let val ops_diff = get_ops_diff trg ntrg
                    val defs = Aux.select_defs (Aux.isar_thms_of ntrg)
                    val ops_tab_ext = make_ops_tab ops_diff defs
                    val upd_ops_tab = update_ops_tab local_trg_ops_tab defs 
                    val sigma' = set_trg sigma ntrg
                    val sigma'' = set_trg_ops_tab sigma' (upd_ops_tab @ ops_tab_ext)
                in
                 sigma''
                end)  
              else raise SIGMORPH ("Theory " ^ n1 ^ " is not an ancestor of the theory " ^ n2))
       end

(* ******************** Extend the domain of a signature morphism ****************** *)
   
fun dom_ext (sigma as (SigMorph{src, trg, typeMap, opMap, local_src_ltypes, local_src_ops_tab, local_trg_ops_tab,
                                ltypes_inter, ops_inter})) tpMap_ext opMap_ext nsrc =
 let val n1 = Aux.name_of_thy src
     val n2 = Aux.name_of_thy nsrc
     
     fun chk () =
       let val chk_inter = case (chck_types_inter tpMap_ext ltypes_inter) of
                             [] => (case (chck_ops_inter opMap_ext ops_inter) of
                                     [] => None
                                    | l => Some ("\nFollowing consts have to be mapped by identity: " ^ (commas l)))
                            | l1 => (case (chck_ops_inter opMap_ext ops_inter) of
                                     [] => Some ("\nFollowing logical type(s) have to be mapped by identity: " ^ (commas l1))
                                     | l2 => Some ("\nFollowing logical type(s) have to be mapped by identity: " ^ (commas l1) ^ "\nFollowing consts have to be mapped by identity: " ^ (commas l2)))

           
           val sg_nsrc = sign_of nsrc
           val sg_trg = sign_of trg 
           val lt1 = local_src_ltypes @ ltypes_inter
           val lt2 = get_ltypes sg_nsrc
           val diff = lt2 \\ lt1
           val chk_type_ext = let val data_src = SigData{sg=sg_nsrc, types=diff, ops=[]}
                                  val data_trg = SigData{sg=sg_trg, types=(get_ltypes sg_trg), ops=[]}
                                  val data=(data_src, data_trg)  
                              in
                               case (mapfilter (chck_typ data) tpMap_ext) of  
                                [] => None
                               | l => Some (commas l)
                              end

           val sigma' = (if (diff = []) then set_src sigma nsrc 
                         else (let val ext_t_dom = map fst tpMap_ext
                                   val diff' = (diff \\ ext_t_dom)
                                   val tsigma = set_src sigma nsrc
                                   val checks = mapfilter (fn x => x) [chk_type_ext, chk_inter]
                                   val tsigma' = (case checks of
                                                   [] => set_ltypes tsigma (local_src_ltypes @ ext_t_dom)
                                                  | l => (let val _ = writeln (commas l)
                                                                 in
                                                                   raise SIGMORPH "Domain extension failed."
                                                                 end))
                               in
                                if (diff' = []) then ext tsigma' tpMap_ext [] 
                                else raise SIGMORPH ("New logical types in the domain extension, which are not mapped by type map extension: " ^ (commas diff'))
                               end))

           val (src_ops_tab_ext, upd_ops_tab)
                               = let val src_ops = (map fst local_src_ops_tab) @ ops_inter
                                     val nsrc_ops = get_ops sg_nsrc
                                     val ops_diff = nsrc_ops \\ src_ops
                                     val nsrc_defs = Aux.select_defs (Aux.isar_thms_of nsrc)
                                 in
                                  (make_ops_tab ops_diff nsrc_defs, update_ops_tab local_src_ops_tab nsrc_defs)
                                 end

           val nsrc_lops = get_lops src_ops_tab_ext
           val ext_nsrc_lops = ext_ops sg_nsrc nsrc_lops
           val ext_trg_ops  = ext_ops sg_trg (get_ops sg_trg)

           val (l1, l2, l3) = match_ops sigma' ext_nsrc_lops ext_trg_ops (opMap @ opMap_ext) 
           val _ = writeln (foldl Aux.show_opmap ("", l1)) 
   
        in
          if (l2 = []) then 
               (if (l3 = []) then 
                    (let val sigma'' = set_src_ops_tab sigma' (upd_ops_tab @ src_ops_tab_ext)
                     in
                      ext sigma'' [] l1
                     end)
                else let val _ = writeln ("There are ambiguous mapping(s): " ^ (foldl Aux.show_amb ("", l3)))
                     in raise SIGMORPH ("Domain extension for " ^ n1 ^ " to " ^ n2 ^ " failed.") end)
 
          else (if (l3 = []) then let val _ = writeln ("Following operation(s) are not matched: " ^ (commas l2))
                                  in raise SIGMORPH ("Domain extension for " ^ n1 ^ " to " ^ n2 ^ " failed.") end
                else let val _ = writeln ("Following operation(s) are not matched: " ^ (commas l2) ^ "\n There are ambiguous mapping(s): " ^ (foldl Aux.show_amb ("", l3)))
                     in raise SIGMORPH ("Domain extension for " ^ n1 ^ " to " ^ n2 ^ " failed.") end)
        end

  in 
   if (n1 = n2) then chk () 
   else (if (Theory.subthy (src, nsrc)) then chk ()
         else raise SIGMORPH ("Theory " ^ n1 ^ " is not an ancestor of the theory " ^ n2))      
  end 





(* *** constdefs splitting *** *)

fun sel_defs [] = []
  | sel_defs ((opnm, opt)::tab) = case opt of
                                    None => sel_defs tab
                                  | Some defnm => (opnm, defnm)::(sel_defs tab)


(* classify the src-constdefs in thms that are in the domain of sigma
   and in thms that are not in the domain of sigma *)
fun split_src_cdefs sigma 
         = let val src = src sigma
               val local_src_ops_tab = local_src_ops_tab sigma
               val defs = sel_defs local_src_ops_tab             
               val sgm_dom   = map fst (opMap sigma)
               fun cond (def_opnm, defnm) = (def_opnm mem sgm_dom)              
                                    
               val (l1, l2) = Aux.split_list cond defs
               fun ext_by_thm (def_opnm, defnm) = let val def = Some (Aux.get_thm src defnm)
                                                                 handle e => None
                                          in 
                                           (def_opnm, defnm, def)
                                          end
               val (l1', l2') = (map ext_by_thm l1, map ext_by_thm l2)
           in
              (l1', l2')
           end


(* classify the target-constdefs in thms that are in the range of sigma
   and in thms that are not in the range of sigma *)
fun split_trg_cdefs sigma 
         = let val trg = trg sigma
               val local_trg_ops_tab = local_trg_ops_tab sigma
               val defs = sel_defs local_trg_ops_tab             
               val sgm_rng   = map snd (opMap sigma)
               fun cond (def_opnm, defnm) = (def_opnm mem sgm_rng)              
                                    
               val (l1, l2) = Aux.split_list cond defs
               fun ext_by_thm (def_opnm, defnm) = let val def = Some (Aux.get_thm trg defnm)
                                                                 handle e => None
                                          in 
                                           (def_opnm, defnm, def)
                                          end
               val (l1', l2') = (map ext_by_thm l1, map ext_by_thm l2)
           in
              (l1', l2')
           end

fun is_inclusion sigma =
  let val tmap = tpMap sigma
      val opmap = opMap sigma
      fun chk_tpMap [] = true
        | chk_tpMap ((s, t)::mp) = case t of 
                                    Simple t' => (s = t') andalso (chk_tpMap mp)
                                   | Comp _ => false
      fun chk_opMap [] = true
        | chk_opMap ((s, t)::mp) = (s = t) andalso (chk_opMap mp)
  in
   (chk_tpMap tmap) andalso (chk_opMap opmap)
  end

fun make_inclusion sthy tthy =
 let val sg1 = sign_of sthy
     val sg2 = sign_of tthy
 in
  if (Sign.is_draft sg1) then (if (Sign.is_draft sg2) then make sthy tthy [] []
                               else raise SIGMORPH "Domain is draft")
  else (let val anc_nms = map Aux.name_of_thy (Aux.isar_ancestors_of true tthy)
            val sthy_nm = Aux.name_of_thy sthy
            val tthy_nm = Aux.name_of_thy tthy
        in
         if (sthy_nm mem anc_nms) then make sthy tthy [] []
         else raise SIGMORPH ("Theory " ^ sthy_nm ^ " has to be an ancestor of the theory " ^ tthy_nm)
        end)
  end

end

