
signature T_TOPLEVEL =  
sig

(* *** Managing Isabelle's syntax *** *)
 
 val add_mixfix : (string * mixfix) -> unit
 val get_mixfix : string -> mixfix Library.option

 val add_nonterm : (string * bstring) -> unit
 val get_nonterms : string -> bstring list

 val add_syn_decl : (string * (string * bool * bstring * Term.typ * mixfix)) -> unit
 val get_syn_decls : string -> (string * bool * bstring * Term.typ * mixfix) list 

 val add_trrule : (string * ((xstring * string) Syntax.trrule)) -> unit
 val get_trrules : string -> ((xstring * string) Syntax.trrule) list

 val add_parse_ast_trans : (string * string) -> unit
 val add_parse_trans : (string * string) -> unit 
 val add_print_trans : (string * string) -> unit
 val add_print_ast_trans : (string * string) -> unit
 val add_token_trans : (string * string) -> unit

 val get_parse_ast_trans : string -> string list
 val get_parse_trans : string -> string list 
 val get_print_trans : string -> string list
 val get_print_ast_trans : string -> string list
 val get_token_trans : string -> string list


 val translate_syntax : ThyMorph.thy_morph -> ThyMorph.thy_morph

 val upd_rem_tabs : string -> unit

(* *** Managing morphisms and trafos *** *)
 val add_sigmorph : (string * SigMorph.sig_morph) -> unit
 val get_sigmorph : string -> SigMorph.sig_morph Library.option
 val add_thymorph : (string * ThyMorph.thy_morph) -> unit
 val get_thymorph : string -> ThyMorph.thy_morph Library.option

 val add_trafo : (string * Trafo.trafo) -> unit
 val get_trafo : string -> Trafo.trafo Library.option


 val print_sigmorphs : unit -> string 
 val print_thymorphs : unit -> string  
 val print_trafos    : unit -> string 


(* define_trafo : Theory.theory -> Theory.theory -> TRule.Trafo_rule *)


(* *** Construction of morphisms *** *)
 val try_sigmorph : Theory.theory -> Theory.theory -> SigMorph.type_map SigMorph.map_rep -> (string * string) list
                       -> SigMorph.sig_morph
 val try_thymorph : SigMorph.sig_morph -> ThyMorph.thy_morph 
 

(* *** Theorem translation *** *)
 val translate_thm : ThyMorph.thy_morph -> string -> string -> ThyMorph.thy_morph 


(* *** Composition *** *)
 val thymorph_compose : ThyMorph.thy_morph -> ThyMorph.thy_morph -> ThyMorph.thy_morph 

 val instantiate : bool -> (string * ThyMorph.thy_morph) -> (string * ThyMorph.thy_morph) -> 
                   (string * (string * mixfix Library.option)) list -> ThyMorph.thy_morph


 val thymorph_colimit : TextIO.outstream Library.option -> bool -> (string * string) list -> (string * string * string) -> 
                        (ThyMorph.thy_morph * ThyMorph.thy_morph) -> 
                        (Theory.theory * ThyMorph.thy_morph * ThyMorph.thy_morph)

 val fibred_sum : TextIO.outstream Library.option -> (string * string) list -> (string * string * string) ->
                   ThyMorph.thy_morph list -> (Theory.theory * ThyMorph.thy_morph * ThyMorph.thy_morph)

 exception T_TOPLEVEL
 exception COMPOSE of string
 exception COLIMIT of string
 exception INSTANTIATE of string

(* visible flags *)
   val verbose : bool ref

end





structure T_Toplevel : T_TOPLEVEL =
struct

val verbose = ref false;
fun prt_verb s = if !verbose then writeln s else ()


exception T_TOPLEVEL
exception COMPOSE of string
exception COLIMIT of string
exception INSTANTIATE of string

(* *************************** Syntax ****************************** *)

val transl_tab = ref Symtab.empty : ((string list) Symtab.table) ref
val transl_ops_tab = ref Symtab.empty : ((string list) Symtab.table) ref

fun add_transl t idt thynm = 
  let val tab = !t 
      val tl = case (Symtab.lookup (tab, idt)) of
                Some l => l
              | None => []
      val tab' = Symtab.update ((idt, thynm::tl), tab)
  in
   t := tab'
  end

fun is_transl t idt thynm =
  let val tab = !t
      val tl = case (Symtab.lookup (tab, idt)) of
                Some l => l
               | None => []
  in
   thynm mem tl
  end   

val add_transl1 = add_transl transl_tab
val add_transl2 = add_transl transl_ops_tab
val is_transl1 = is_transl transl_tab
val is_transl2 = is_transl transl_ops_tab

fun upd_rem_tab t thynm =
  let val tab = !t
      val keys = Symtab.keys tab
      fun rem [] = []
        | rem (x::xs) = (if (x = thynm) then rem xs else (x::(rem xs)))
      fun chk t [] = t
        | chk t (x::xs) = (case (Symtab.lookup (tab, x)) of
                            None => error ("Unexpected error in upd_rem_tab")
                          | Some i => (let val i' = rem i
                                           val t' = Symtab.update ((x, i'), t)
                                       in
                                        chk t' xs
                                       end))
      val tab' = chk tab keys
  in
   t := tab'
  end

fun upd_rem_tabs thynm = 
 let val _ = upd_rem_tab transl_tab thynm
 in
 upd_rem_tab transl_ops_tab thynm
 end
                                           

(* ************* Mixfix ******************* *)

val parens = enclose "(" ")";
val brackets = enclose "[" "]";
val spaces = space_implode " ";

fun string_of_mixfix NoSyn = ""
  | string_of_mixfix (Mixfix (sy, ps, p)) =
      parens (spaces [quote sy, brackets (commas (map string_of_int ps)), string_of_int p])
  | string_of_mixfix (Delimfix sy) = parens (quote sy)

  | string_of_mixfix (InfixName (sy, p)) = parens (spaces ["infix", quote sy, string_of_int p])
  | string_of_mixfix (InfixlName (sy, p)) = parens (spaces ["infixl", quote sy, string_of_int p])
  | string_of_mixfix (InfixrName (sy, p)) = parens (spaces ["infixl", quote sy, string_of_int p])
  | string_of_mixfix (Infix p) = parens (spaces ["infix", string_of_int p])
  | string_of_mixfix (Infixl p) = parens (spaces ["infixl", string_of_int p])
  | string_of_mixfix (Infixr p) = parens (spaces ["infixr", string_of_int p])
  | string_of_mixfix (Binder (sy, p1, p2)) =
      parens (spaces ["binder", brackets (string_of_int p1), string_of_int p2]);


 val mixfix_tab = ref Symtab.empty : (mixfix Symtab.table) ref

 fun add_mixfix (opname, mxfx) = 
  let val (ref tab) = mixfix_tab
      val tab' = Symtab.update ((opname, mxfx), tab)
  in
   mixfix_tab := tab'
  end 

 fun get_mixfix opname =
  let val (ref tab) = mixfix_tab
  in
   Symtab.lookup (tab, opname)
  end  

(* *********** Nonterminals ************ *)

 val nonterm_tab = ref Symtab.empty : ((bstring list) Symtab.table) ref

 fun add_nonterm (thyname, nt) = 
  let val (ref tab) = nonterm_tab
      val nts = (case (Symtab.lookup (tab, thyname)) of
                  Some l => l
                 | None => [])
      val tab' = Symtab.update ((thyname, nts @ [(nt)]), tab)
  in
   nonterm_tab := tab'
  end 

 fun get_nonterms thyname =
  let val (ref tab) = nonterm_tab
  in
   case (Symtab.lookup (tab, thyname)) of
     Some nts => nts
   | None => []
  end  



(* *********** Syntax ************ *)

 val syntax_tab = ref Symtab.empty : (((string * bool * bstring * Term.typ * mixfix) list) Symtab.table) ref

 fun add_syn_decl (thyname, decl) = 
  let val (ref tab) = syntax_tab
      val decls = (case (Symtab.lookup (tab, thyname)) of
                    Some l => l
                   | None => [])
      fun qlf (s, b, nm, t, m) = (s, b, thyname ^ "." ^ nm, t, m)
      val tab' = Symtab.update ((thyname, decls @ [decl]), tab)
  in
   syntax_tab := tab'
  end 

 fun get_syn_decls thyname =
  let val (ref tab) = syntax_tab
  in
   case (Symtab.lookup (tab, thyname)) of
     Some decls => decls
   | None => []
  end  


  

(* *********** Translations ************ *)


 val trrule_tab = ref Symtab.empty : ((((xstring * string) Syntax.trrule) list) Symtab.table) ref

 fun add_trrule (thyname, trrule) = 
  let val (ref tab) = trrule_tab
      val trrules = (case (Symtab.lookup (tab, thyname)) of
                      Some l => l
                    | None => [])
      val tab' = Symtab.update ((thyname, trrules @ [trrule]), tab)
  in
   trrule_tab := tab'
  end 

 fun get_trrules thyname =
  let val (ref tab) = trrule_tab
  in
   case (Symtab.lookup (tab, thyname)) of
     Some trrules => trrules
   | None => []
  end  

 
(* *********** Translation functions ************ *)


 val parse_ast_tab   = ref Symtab.empty : ((string list) Symtab.table) ref
 val parse_tab       = ref Symtab.empty : ((string list) Symtab.table) ref
 val print_tab       = ref Symtab.empty : ((string list) Symtab.table) ref
 val print_ast_tab   = ref Symtab.empty : ((string list) Symtab.table) ref
 val token_tab       = ref Symtab.empty : ((string list) Symtab.table) ref

(*
val typed_print_tab = ref Symtab.empty : ((string list) Symtab.table) ref
*)

 val tab_list = ref [parse_ast_tab, parse_tab, print_tab, print_ast_tab, token_tab]
 fun index l 1 = hd l
   | index l n = index (tl l) (n-1)
 fun upd l t 1 = (let val t_ref = hd l
                      val _ = (t_ref := t)
                  in
                   t_ref::(tl l)
                  end)
   | upd l t n = (hd l)::(upd (tl l) t (n-1)) 

 fun add_trfun i (thyname, trfun) = 
  let val (ref tablist) = tab_list 
      val (ref tab) = index tablist i
      val trfuns = (case (Symtab.lookup (tab, thyname)) of
                      Some l => l
                    | None => [])
      val tab' = Symtab.update ((thyname, trfuns @ [trfun]), tab)
  in
   tab_list := (upd tablist tab' i)
  end 

 fun get_trfuns i thyname =
  let val (ref tablist) = tab_list
      val (ref tab) = index tablist i
  in
   case (Symtab.lookup (tab, thyname)) of
     Some trfuns => trfuns
   | None => []
  end  

 val add_parse_ast_trans = add_trfun 1
 val add_parse_trans = add_trfun 2  
 val add_print_trans = add_trfun 3
 val add_print_ast_trans = add_trfun 4
 val add_token_trans = add_trfun 5

 val get_parse_ast_trans = get_trfuns 1
 val get_parse_trans = get_trfuns 2  
 val get_print_trans = get_trfuns 3
 val get_print_ast_trans = get_trfuns 4
 val get_token_trans = get_trfuns 5






(* **************************** *)


 val sigma_list    = ref [] : ((string * SigMorph.sig_morph) list) ref 
 val thymorph_list = ref [] : ((string * ThyMorph.thy_morph) list) ref  
 val trafo_list    = ref [] : ((string * Trafo.trafo) list) ref 


 fun remove a [] = []
   | remove a (x::xs) = if (a = fst x) then xs else (x::(remove a xs))  


 fun add_sigmorph (name, sigma) =
   let val (ref sl) = sigma_list
       val sl' = case (assoc_string (sl, name)) of 
                   Some sg => (remove name sl)
                 | None    => sl
       val sl'' = (name, sigma) :: sl'
   in
    sigma_list := sl''
   end

 fun get_sigmorph name =
   let val (ref sl) = sigma_list
   in
    assoc_string (sl, name)
   end 

 fun print_sigmorphs () =
   let val sl = !sigma_list;
       fun prnt_sigmorph (s, (nm, sigma)) = let val srcnm = Aux.name_of_thy (SigMorph.src sigma)
                                                val trgnm = Aux.name_of_thy (SigMorph.trg sigma)
                                            in
                                              (s ^  (nm ^ " : " ^ srcnm ^ " --> " ^ trgnm ^ "\n"))
                                            end
   in
    foldl prnt_sigmorph ("", sl) 
   end    

 fun add_thymorph (name, tau) =
   let val (ref thl) = thymorph_list
       val thl' = case (assoc_string (thl, name)) of 
                   Some t => (remove name thl)
                 | None    => thl
       val thl'' = (name, tau) :: thl'
   in
    thymorph_list := thl''
   end

 fun get_thymorph name =
   let val (ref thl) = thymorph_list
   in
    assoc_string (thl, name)
   end 

 fun print_thymorphs () =
   let val thl = !thymorph_list
       fun prnt_thymorph (s, (nm, tm)) = let val sigma = ThyMorph.sig_morph tm
                                             val srcnm = Aux.name_of_thy (SigMorph.src sigma)
                                             val trgnm = Aux.name_of_thy (SigMorph.trg sigma)
                                         in
                                          (s ^  (nm ^ " : " ^ srcnm ^ " --> " ^ trgnm ^ "\n"))
                                         end

   in
     foldl prnt_thymorph ("", thl)
   end    

 fun add_trafo (name, tr) =
   let val (ref trl) = trafo_list
       val trl' = case (assoc_string (trl, name)) of 
                   Some t => (remove name trl)
                 | None    => trl
       val trl'' = (name, tr) :: trl'
   in
    trafo_list := trl''
   end

 fun get_trafo name =
   let val (ref trl) = trafo_list
   in
    assoc_string (trl, name)
   end 

  fun print_trafos () =
   let val tl = !trafo_list
       fun prnt_trafo (s, (nm, trafo)) = let val pnm = Aux.name_of_thy (Trafo.param trafo)
                                             val bnm = Aux.name_of_thy (Trafo.body trafo)
                                         in
                                          (s ^  (nm ^ " := (" ^ pnm ^ ", " ^ bnm ^ ")\n"))
                                         end
   in
     foldl prnt_trafo ("", tl)
   end    




(* ******** operations on objects **** *)

fun check_inclusion tmorph =
 let val sigma = ThyMorph.sig_morph tmorph
     val tpMap = SigMorph.tpMap sigma
     val opMap = SigMorph.opMap sigma
     fun chk_tpMap [] = true
       | chk_tpMap ((s, t)::mp) = case t of
                                    SigMorph.Simple t' => (if (s = t') then (chk_tpMap mp) else false)
                                   | _ => false
     fun chk_opMap [] = true
       | chk_opMap ((s, t)::mp) = (if (s = t) then (chk_opMap mp) else false)
 in
  (chk_tpMap tpMap) andalso (chk_opMap opMap)
 end

fun find_preimage thynm =
 let val (ref thl) = thymorph_list
     fun chk (nm, tmorph) = 
      let val sigma = ThyMorph.sig_morph tmorph
          val src = SigMorph.src sigma
          val trg = SigMorph.trg sigma
      in
       (if (Aux.name_of_thy trg = thynm) then (Some (src, tmorph)) else None)
      end
 in
  mapfilter (fn x => x) (map chk thl)
 end


fun find_incl_ancestors thynm = 
 let val pimg = find_preimage thynm
     fun sel (thy, tmorph) = let val thynm' = Aux.name_of_thy thy
                             in
                              if ((not (thynm = thynm')) andalso (check_inclusion tmorph)) then (Some thynm') else None
                             end    
     val anc = mapfilter (fn x => x) (map sel pimg)
 in
  foldl op union ([thynm], map find_incl_ancestors anc)
 end




(* ***** Aux *** *)

fun type_arity(sg, tycon) = 
 let val rep = Type.rep_tsig (Sign.tsig_of sg)
 in
  case (Symtab.lookup((#arities rep), tycon)) of
   Some a => (case (Symtab.lookup((#tycons rep), tycon)) of 
               Some i => Some (a, i)
             | None => (let val _ = writeln ("Unexpected exception in type_arity")
                        in
                         raise T_TOPLEVEL
                        end))
 | None => None
 end


(* ******************************************* *)

fun try_sigmorph src trg tpMap opMap =
  let val (opt_sigma, l2, l3) = SigMorph.try_sigmorph src trg tpMap opMap
  in
   case opt_sigma of
    Some sigma => sigma
  | None =>
    (if (l2 = []) then   (let val to_show_amb = foldl Aux.show_amb ("", l3) 
                              val _ = writeln ("\nThere are following ambiguous mapping(s): " ^ to_show_amb)
                              val _ = writeln "Please use the operation map to remove this ambiguity.\n" 
                          in raise T_TOPLEVEL end)
   else (if (l3 = []) then (let val _ = writeln ("\nCannot map following operations: [ " ^ (commas l2) ^ " ]\n")
                            in raise T_TOPLEVEL end)
                      else (let val _ = writeln ("\nCannot map following operations: [ " ^ (commas l2) ^ " ]\n")
                                val _ = writeln "Furthermore: "
                                val to_show_amb = foldl Aux.show_amb ("", l3) 
                                val _ = writeln ("\nthere are following ambiguous mapping(s): " ^ to_show_amb)
                                val _ = writeln "Please use the operation map to remove this ambiguity.\n" 
                            in raise T_TOPLEVEL end)))
  end      
    








(* ************************************************************************************* *)
(* ****************** Translations along a theory morphism ***************************** *)
(* ************************************************************************************* *)

fun translate_op sigma thynm ((thy, opMap_ext, topnms), (opnm, tp)) =
  let val opMap = SigMorph.opMap sigma
      val tp' = Type.unvarifyT tp
      val tp'' = SigMorph.map_type sigma tp'
      val trg_sg = sign_of (SigMorph.src sigma)
      fun find_name_and_ins (nm, thy) =
          case (assoc_string(opMap, opnm)) of
           Some i => (thy, opMap_ext, topnms)
          | None =>                                   
            (if (nm mem topnms) then find_name_and_ins ((nm ^ "_transl"), thy)
             else (let val mxfx = (case (get_mixfix opnm) of
                                    Some mixfix => mixfix
                                   | None => NoSyn)
                       val _ = writeln ("Adding const: " ^ nm ^ " :: \"" ^ (Sign.string_of_typ trg_sg tp'') ^ "\"   " ^
                                         (string_of_mixfix mxfx))
                   in
                    (Theory.add_consts_i [(NameSpace.base nm, tp'', mxfx)] thy, (opnm, nm)::opMap_ext, nm::topnms)
                   end))

      val nunm = (thynm ^ "." ^ (NameSpace.base opnm))
  in
   find_name_and_ins (nunm, thy) 
  end  


fun translate_ops sigma =
  let val src = SigMorph.src sigma
      val trg = SigMorph.trg sigma
      val opMap = SigMorph.opMap sigma
      val local_src_ops_tab = SigMorph.local_src_ops_tab sigma
      val local_trg_ops_tab = SigMorph.local_trg_ops_tab sigma
      val src_opnms = map fst local_src_ops_tab
      val trg_opnms = map fst local_trg_ops_tab 
      fun ext_ops sg = 
        map (fn opnm => (let val tp = case (Sign.const_type sg opnm) of
                                        Some t => t
                                      | None => (let val _ = writeln "Unexpected exception in translate_ops"
                                                 in raise T_TOPLEVEL end)
                         in (opnm, tp) end ))
      val ext_src_ops = ext_ops (sign_of src) src_opnms
      val trgnm = Aux.name_of_thy trg
      val (ntrg, opMap_ext, _) = foldl (translate_op sigma trgnm) ((trg, [], trg_opnms), ext_src_ops)
      val sigma' = (if ((Aux.name_of_thy src) = (Aux.name_of_thy trg)) 
                    then SigMorph.dom_ext sigma [] (opMap @ opMap_ext) ntrg
                    else sigma)
      val sigma'' = SigMorph.cod_ext sigma' ntrg
  in
   SigMorph.ext sigma'' [] opMap_ext
  end 

 
fun translate_def sigma thmMap thynm ((thy, thmMap_ext, tnms), (opnm, defnm, def)) =
  let val opMap = SigMorph.opMap sigma
      val def_term = prop_of def
      val img_opnm = case (assoc_string(opMap, opnm)) of
                      Some i => i
                    | None => (let val _ = writeln "Unexpected exception in translate_def: opMap"
                               in raise T_TOPLEVEL end)
      fun find_name_and_ins (nm, thy) =
          case (assoc_string(thmMap, defnm)) of
           Some i => (thy, thmMap_ext, tnms)
          | None =>                                   
            (if (nm mem tnms) then find_name_and_ins ((nm ^ "_transl"), thy)
             else ( let val mp_term = SigMorph.map_term sigma def_term
                        val mp_term' = Aux.set_opname mp_term img_opnm
                        val mp_term'' = Logic.unvarify mp_term'
                        val _ = writeln ("Adding def: " ^ nm)
                        val thy' = IsarThy.add_defs_i (false, [Thm.no_attributes(NameSpace.base nm, mp_term'')]) thy
                    in
                     (thy', (defnm, nm)::thmMap_ext, nm::tnms)
                    end))

       val nunm = (thynm ^ "." ^ (NameSpace.base defnm))
  in
   find_name_and_ins (nunm, thy) 
  end  


fun translate_defs sigma thmMap trg_thmnms =
  let val sigma' = translate_ops sigma
      val src = SigMorph.src sigma'
      val trg = SigMorph.trg sigma'
      val opMap = SigMorph.opMap sigma'
      val trgnm = Aux.name_of_thy trg

      val local_src_ops_tab = SigMorph.local_src_ops_tab sigma'
      val local_trg_ops_tab = SigMorph.local_trg_ops_tab sigma'
 
      fun sel_defnms [] = []
        | sel_defnms ((opnm, opt)::tab) = case opt of 
                                           Some defnm => (case (assoc_string (thmMap, defnm)) of
                                                           None => (opnm, defnm)::(sel_defnms tab)
                                                          | _ => sel_defnms tab)
                                          | None => sel_defnms tab
      val src_defnms = sel_defnms local_src_ops_tab
      val trg_defnms = map snd (sel_defnms local_trg_ops_tab)
      val ext_src_defnms = mapfilter (fn (opnm, defnm) => (Some (opnm, defnm, Aux.get_thm src defnm))
                                                          handle e => None) src_defnms 
      val trg_nms = trg_defnms @ trg_thmnms

      val (ntrg, thmMap_ext, _) = foldl (translate_def sigma' thmMap trgnm) ((trg, [], trg_nms), ext_src_defnms)
      val sigma'' = (if ((Aux.name_of_thy src) = trgnm) 
                    then SigMorph.dom_ext sigma' [] opMap ntrg
                    else sigma')
      val sigma''' = SigMorph.cod_ext sigma'' ntrg
  in
   (sigma''', thmMap_ext)
  end 

fun translate_defs' tmorph =
  let val sigma = ThyMorph.sig_morph tmorph
  in 
    (let val thmMap = ThyMorph.thmMap tmorph
         val trg_thmnms = (map fst (ThyMorph.local_trg_axms tmorph)) @ (map fst (ThyMorph.local_trg_thms tmorph))
         val (sigma', thmMap_ext) = translate_defs sigma thmMap trg_thmnms
     in
      case thmMap_ext of
       [] => tmorph
      | _ => ThyMorph.make sigma' (thmMap @ thmMap_ext)
     end)
  end


fun translate_syntax init_tmorph =
  let val tmorph = translate_defs' init_tmorph
      val sigma = ThyMorph.sig_morph tmorph
      val local_src_anc = rev (map Aux.name_of_thy (ThyMorph.src_anc tmorph))
      val trg = SigMorph.trg sigma
      val trgnm = Aux.name_of_thy trg
      val trg_sg = sign_of trg

      val nts = foldl (op @) ([], map get_nonterms local_src_anc)
      val syn_decls = foldl (op @) ([], map get_syn_decls local_src_anc)
      val trrules = foldl (op @) ([], map get_trrules local_src_anc)
      val parse_ast_trans = foldl (op @) ([], map get_parse_ast_trans local_src_anc)
      val parse_trans = foldl (op @) ([], map get_parse_trans local_src_anc)
      val print_trans = foldl (op @) ([], map get_print_trans local_src_anc)
      val print_ast_trans = foldl (op @) ([], map get_print_ast_trans local_src_anc)
      val token_trans = foldl (op @) ([], map get_token_trans local_src_anc)

      fun ins_nt (thy, nt) = (if (is_transl1 nt trgnm) then thy
                              else (let val _ = writeln ("Adding nonterminal: " ^ nt)
                                        val _ = add_transl1 nt trgnm
                                    in 
                                     Theory.add_nonterminals [nt] thy
                                    end))
      val trg2 = foldl ins_nt (trg, nts)

      fun transl_tp (Type(nm, ts)) = 
          let val unm = NameSpace.unpack nm
          in
           if ((hd unm) mem local_src_anc) then Type (NameSpace.pack (trgnm::(tl unm)), map transl_tp ts)
           else Type (nm, map transl_tp ts)    
          end
        | transl_tp var = var 

      fun transl_syn_decl (thy, (s, b, nm, tp, mxfx)) =
       (if (is_transl1 nm (Aux.name_of_thy thy)) then thy
        else (let val ntp = transl_tp (SigMorph.map_type sigma tp)
                  val _ = writeln ("Adding syntax declaration: " ^ nm ^ 
                                   " :: \"" ^ (Sign.string_of_typ trg_sg ntp) ^ 
                                   "\"   " ^ (string_of_mixfix mxfx))
                  val _ = add_transl1 nm (Aux.name_of_thy thy)
              in 
               Theory.add_modesyntax_i (s, b) [(nm, ntp, mxfx)] thy
              end))
      val trg3 = foldl transl_syn_decl (trg2, syn_decls) 

      fun show_trrule (Syntax.ParseRule (l, r)) = ("\"" ^ snd l ^ "\"" ^ " => " ^ "\"" ^ snd r ^ "\"" )
        | show_trrule (Syntax.PrintRule (l, r)) = ("\"" ^ snd l ^ "\"" ^ " <= " ^ "\"" ^ snd r ^ "\"")
        | show_trrule (Syntax.ParsePrintRule (l, r)) = ("\"" ^ snd l ^ "\"" ^ " == " ^ "\"" ^ snd r ^ "\"")
      fun ins_trrule (thy, trrule) = (let val _ = writeln ("Adding translation: " ^ (show_trrule trrule))
                                      in
                                       Theory.add_trrules [trrule] thy
                                      end)
      val trg4 = foldl ins_trrule (trg3, trrules)

      fun ins_parse_ast (thy, t) = (let val _ = writeln ("Adding parse_ast_translation: " ^ t)
                                    in
                                      IsarThy.parse_ast_translation t thy
                                    end)
      val trg5 = foldl ins_parse_ast (trg4, parse_ast_trans)
     
 
      fun ins_parse (thy, t) = (let val _ = writeln ("Adding parse_translation: " ^ t)
                                in
                                  IsarThy.parse_translation t thy
                                end)
      val trg6 = foldl ins_parse (trg5, parse_trans)

      fun ins_print (thy, t) = (let val _ = writeln ("Adding print_translation: " ^ t)
                                in
                                  IsarThy.print_translation t thy
                                end)
      val trg7 = foldl ins_print (trg6, print_trans)
   
      fun ins_print_ast (thy, t) = (let val _ = writeln ("Adding print_ast_translation: " ^ t)
                                    in
                                      IsarThy.print_ast_translation t thy
                                    end)
      val trg8 = foldl ins_print_ast (trg7, print_ast_trans)

      fun ins_token (thy, t) = (let val _ = writeln ("Adding token_translation: " ^ t)
                                in
                                  IsarThy.token_translation t thy
                                end)
      val trg9 = foldl ins_token (trg8, token_trans)
      
      val tmorph' = ThyMorph.cod_ext tmorph [] trg9
  in
   tmorph'
  end 



(* **************** Normalization ***************** *)

fun sel_appr_defs sg [] = []
  | sel_appr_defs sg (thm::thms) =
    let val sg' = sign_of_thm thm
    in
     if Sign.subsig(sg', sg) then (thm::(sel_appr_defs sg thms)) 
     else sel_appr_defs sg thms
    end

fun normalize_i to_fold to_unfold all_names (tmorph_data, thm_data) =
  let val (thm_map_ext, thy) = tmorph_data
      val (iimg, (thmnm, thm)) = thm_data

      val thm_sg = sign_of_thm thm          
      val to_fold' = sel_appr_defs thm_sg to_fold 
      val to_unfold' = sel_appr_defs thm_sg to_unfold

      fun find_name nm = (if (nm mem all_names) then (find_name (nm ^ "_norm")) 
                                                else nm)
      val thm' = Seq.hd(fold_tac to_fold' thm)
      val thm'' = Seq.hd(rewrite_tac to_unfold' thm')

      val (nthy, ext) = (if ( (prop_of thm'') aconv (prop_of thm) ) then (thy, [])
                         else (let val nnm = find_name (thmnm ^ "_norm")
                                   val nnm_base = NameSpace.base nnm
                                   val _ = writeln ("Normalizing theorem (axiom) " ^ thmnm ^ " -> " ^
                                                    nnm) 
                                   val _ = writeln (string_of_thm thm'')  
                                   val (thy', l) = PureThy.add_thms [Thm.no_attributes(nnm_base, thm'')] thy
                               in
                                (thy', [(iimg, nnm)])
                               end))
  in
   (ext @ thm_map_ext, nthy)
  end


fun normalize sigma thmMap' =
if (SigMorph.is_inclusion sigma) then (sigma, thmMap')
else
(let val (ThyMorph.ThyMorphData{local_trg_axms, local_trg_thms, ...}) = ThyMorph.get_thymorphdata sigma
     val trg_thms = local_trg_axms @ local_trg_thms  
     val trg_thmnms' = (map fst trg_thms)

     val (sigma', thmMap_ext) = translate_defs sigma thmMap' trg_thmnms'
     val opMap = SigMorph.opMap sigma'
     val src = SigMorph.src sigma'
     val trg = SigMorph.trg sigma'
     val local_trg_ops_tab = SigMorph.local_trg_ops_tab sigma'
     val (mpd, nmpd) = SigMorph.split_trg_cdefs sigma'
     val t = map (fn (a, b, c) => (Some (Aux.get_thm trg b))
                                   handle e => None) mpd 
     val t' = map (fn (a, b, c) => (Some (Aux.get_thm trg b))
                                    handle e => None) nmpd
     val to_fold_trg = mapfilter (fn x => x) t
     val to_unfold_trg = mapfilter (fn x => x) t'  
          
     val trg_defnms = mapfilter (fn (a, b) => (case b of
                                                None => None
                                              | Some n => Some a)) local_trg_ops_tab
     val all_names = trg_thmnms' @ trg_defnms

     val thmMap = thmMap' @ thmMap_ext
     val rev_map = map (fn (a, b) => (b, a)) thmMap

     fun chk_img [] = []
       | chk_img ((nm, thm)::thms) = case (assoc_string(rev_map, nm)) of
                                       None => ((None, (nm, get_thm trg nm))::(chk_img thms)
                                                handle e => chk_img thms)
                                     | Some pimg => ((Some pimg, (nm, get_thm trg nm))::(chk_img thms)
                                                     handle e => chk_img thms)
     val thms_data = chk_img trg_thms

     val (thmMap_ext', ntrg) = foldl (normalize_i to_fold_trg to_unfold_trg all_names) (([], trg), thms_data)

     val thmMap_ext'' = mapfilter (fn (a, b) => case a of 
                                                  Some iimg => (Some (iimg, b))
                                                | None => None) thmMap_ext'

     val sigma'' = (if ((Aux.name_of_thy src) = (Aux.name_of_thy trg)) then SigMorph.dom_ext sigma' [] opMap ntrg
                                                                       else sigma')
     val sigma''' = SigMorph.cod_ext sigma'' ntrg
 in
  (sigma''', thmMap_ext'' @ thmMap)
 end)


(* *********** try_thymorph ************************ *)


fun try_thymorph sigma =
 let val (opt_tmorph, _, opt) = ThyMorph.try_thymorph sigma 
  in
   case opt_tmorph of
    Some tmorph => tmorph
  | None =>
    (case opt of 
      None => (let val _ = writeln ("Unexpected exception in try_thymorph")
               in raise T_TOPLEVEL end)
     | Some (sigma, thmMap) =>
        (let val (sigma', thmMap_ext) = normalize sigma thmMap 
             val (opt_tmorph', l, _) = ThyMorph.try_thymorph sigma'
         in
          case opt_tmorph' of
           Some tmorph' => ThyMorph.thmMap_ext tmorph' thmMap_ext
         | None =>  (let val _ = writeln ("Cannot match following axioms: [ " ^ (commas l) ^ " ]")
                         val _ = writeln ("Theory morphism construction failed")
                     in raise T_TOPLEVEL end)
         end))
  end       




(* *** thm translation *** *)

(* Precondition: thms containes only sorted by transitive dependency thms ( (string * thm) )
   from domain of tmorph and all ops and defs from domain are already translated *)

  fun translate_thms_i tmorph renames thms = 
     let val sigma = ThyMorph.sig_morph tmorph
         val opMap = SigMorph.opMap sigma
         val src   = SigMorph.src sigma
         val trg   = SigMorph.trg sigma
         val src_qlf = Aux.name_of_thy src
         val trg_qlf = Aux.name_of_thy trg
         val thm_map = ThyMorph.thmMap tmorph
         fun sel (tnm, tm) = case (assoc_string (thm_map, tnm)) of
                              Some img => false
                             | None => true
         val sel_thms = filter sel thms
 
         val _ = prt_verb ("\nsel_thms: " ^ commas (map fst sel_thms))
         
         val (thm_map_extended, nthy) = ThyMorph.map_thms tmorph renames sel_thms
         val thm_map_ext = thm_map_extended \\ thm_map
         val tmorph' = (if (src_qlf = trg_qlf) then ThyMorph.dom_ext tmorph [] opMap nthy
                                               else tmorph)
         val tmorph'' = ThyMorph.cod_ext tmorph' thm_map_ext nthy
     in
      tmorph''
     end 


  

   fun create_renamelist (name, ren) trg_qlf trg_thmnms [] = []
     | create_renamelist (name, ren) trg_qlf trg_thmnms (nm::nms) =
       let fun find_name nm = (if ((trg_qlf ^ "." ^ nm) mem trg_thmnms) then (find_name (nm ^ "_transl")) else nm)
       in
        if (name = nm) then (if (ren = "") then (let val nnm = find_name (NameSpace.base nm)
                                                 in 
                                                  (NameSpace.base nm, nnm)::(create_renamelist (name, ren) trg_qlf ((trg_qlf ^ "." ^ nnm) ::(trg_thmnms)) nms)
                                                 end)
                                           else (let val nnm = find_name ren
                                                 in 
                                                  (NameSpace.base nm, nnm)::(create_renamelist (name, ren) trg_qlf ((trg_qlf ^ "." ^ nnm) ::(trg_thmnms)) nms)
                                                 end))
        else (let val nnm = find_name (NameSpace.base nm)
              in 
              (NameSpace.base nm, nnm)::(create_renamelist (name, ren) trg_qlf ((trg_qlf ^ "." ^ nnm) ::(trg_thmnms)) nms)
              end)
       end



fun translate_thm_i tmorph thmnm rename =
 let val init_map = ThyMorph.thmMap tmorph
     val rev_map  = map (fn (a, b) => (b, a)) init_map
     val sgmrph   = ThyMorph.sig_morph tmorph
     val init_src = SigMorph.src sgmrph
     val init_trg = SigMorph.trg sgmrph
 in
 if (SigMorph.is_inclusion sgmrph) then tmorph
 else
    ( case (assoc_string (init_map, thmnm)) of
       None => 
               ( case (assoc_string (rev_map, (Aux.name_of_thy init_trg) ^ "." ^ rename)) of
                 (Some inv_im) => let val _ = writeln ("Name " ^ (Aux.name_of_thy init_trg) ^ "." ^ rename ^ 
                                                        " is already in use by " ^ inv_im) 
                                   in tmorph end            
                | None =>
                   ( let val sigma = ThyMorph.sig_morph tmorph
                         val src   = SigMorph.src sigma
                         val trg   = SigMorph.trg sigma
                         val trg_qlf = Aux.name_of_thy trg
                         val local_trg_ops_tab = SigMorph.local_trg_ops_tab sigma
                         val local_src_thms = ThyMorph.local_src_thms tmorph
                         val dg = ThyMorph.DG_src tmorph
                         val src_thmnms = map fst local_src_thms
                         val depends = DepAna.trans_depends_on_filter dg thmnm src_thmnms
                         val dep_thmnms = DepAna.sort_dep dg (thmnm::depends)
                         val _ = prt_verb ("Translating " ^ thmnm ^ "; deps: " ^ (commas dep_thmnms))
                         val dep_thms = map (fn nm => (nm, (Aux.get_thm src nm))) dep_thmnms
                      
                      in 
                       case dep_thms of
                        [] => tmorph
                      | _ =>
                         (let val trg_thmnms' = 
                            (map fst (ThyMorph.local_trg_axms tmorph)) @ (map fst (ThyMorph.local_trg_thms tmorph))
                              val trg_defnms = mapfilter (fn (a, b) => (case b of
                                                                         None => None
                                                                       | Some n => Some a)) local_trg_ops_tab
                              val trg_thmnms = trg_thmnms' @ trg_defnms
                              val renamelist = create_renamelist (thmnm, rename) trg_qlf trg_thmnms (map fst dep_thms)
                              val tmorph' = translate_thms_i tmorph renamelist dep_thms
                         in
                          tmorph'
                         end )
                       end)
               )
                               
     | (Some im) => (if (((NameSpace.base im) = rename) orelse (rename = "")) then tmorph
                     else 
                       ( case (assoc_string (rev_map, (Aux.name_of_thy init_trg) ^ "." ^ rename)) of
                          (Some inv_im) => let val _ = writeln ("Name " ^ (Aux.name_of_thy init_trg) ^ "." ^ rename ^ 
                                                        " is already in use by " ^ inv_im) 
                                           in tmorph end            
                          | None => 
                             (let val src_qlf = Aux.name_of_thy init_src
                                  val trg_qlf = Aux.name_of_thy init_trg
                                  val thm = Aux.get_thm init_trg im
                                  val opMap = SigMorph.opMap sgmrph
                                  val local_trg_ops_tab = SigMorph.local_trg_ops_tab sgmrph
                                  val trg_thmnms' = 
                                  (map fst (ThyMorph.local_trg_axms tmorph)) @ (map fst (ThyMorph.local_trg_thms tmorph))
                                  val trg_defnms = mapfilter (fn (a, b) => (case b of
                                                                             None => None
                                                                           | Some n => Some a)) local_trg_ops_tab
                                  val trg_thmnms = trg_thmnms' @ trg_defnms

                                  fun find_name thmnm =                                   
                                   if (thmnm mem trg_thmnms) then find_name (thmnm ^ "_transl")
                                                             else thmnm  
                                  
                                  val rename' = find_name (trg_qlf ^ "." ^ rename)
                                  val (trg, l) = PureThy.add_thms [Thm.no_attributes((NameSpace.base rename'), thm)] init_trg
                                  val tmorph' = (if (src_qlf = trg_qlf) then ThyMorph.dom_ext tmorph [] opMap trg
                                                                        else tmorph)
                                  val tmorph'' = ThyMorph.cod_ext tmorph' [(thmnm, rename')] trg 
                              in
                               tmorph''
                              end) ))
        )
 end
        
fun translate_thm tmorph thmnm rename =
 let val tmorph' = translate_defs' tmorph
 in
  translate_thm_i tmorph' thmnm rename
 end
                      

(* ****************** Composition ************************** *)

 fun sigmorph_compose sg1 sg2 =
  let val dom1 = SigMorph.src sg1
      val cod1 = SigMorph.trg sg1
      val dom2 = SigMorph.src sg2
      val cod2 = SigMorph.trg sg2
      val comm = ( if (Aux.name_of_thy cod1 = Aux.name_of_thy dom2) then cod1 
                   else raise COMPOSE ("Composition not possible: codomain of the first morphism is not equal domain of the second morphism") )

      val tpMap1 = SigMorph.tpMap sg1
      val tpMap2 = SigMorph.tpMap sg2
      val opMap1 = SigMorph.opMap sg1
      val opMap2 = SigMorph.opMap sg2

      val dom_data = SigMorph.get_thydata dom1
      val cod_data = SigMorph.get_thydata cod2
      val (src_ltypes, src_ops_tab, _, _, _) = SigMorph.get_local_sigmorph_data dom_data cod_data
      val src_lops = map fst (filter (fn (s, opt) => (case opt of
                                                       None => true
                                                      | _ => false)) src_ops_tab)
                                              

      fun make_nutpMap [] = []
        | make_nutpMap (tp::tps) = let val img1 = case assoc_string(tpMap1, tp) of
                                                   Some i => i
                                                  | None => (SigMorph.Simple tp)
                                       val img2 = case img1 of
                                                   SigMorph.Simple s => (case (assoc_string (tpMap2, s)) of
                                                                                   Some i => i
                                                                                 | None => img1)
                                                 | SigMorph.Comp ct => (SigMorph.Comp (SigMorph.map_type sg2 ct))
                                   in
                                    (tp, img2)::(make_nutpMap tps)
                                   end

      fun make_nuopMap [] = []
        | make_nuopMap (s::ops) = let val img1 = case assoc_string(opMap1, s) of
                                                  Some i => i
                                                 | None => s
                                      val img2 = case assoc_string(opMap2, img1) of
                                                  Some i => i
                                                 | None => img1
                                  in
                                   (s, img2)::(make_nuopMap ops)
                                  end
     in
      SigMorph.make dom1 cod2 (make_nutpMap src_ltypes) (make_nuopMap src_lops)
     end



fun thymorph_compose tm1 tm2 =
    let val tm2' = translate_defs' tm2 
        val thmMap1 = ThyMorph.thmMap tm1
        val img1 = map snd thmMap1

        val tm2'' = foldl (fn (tm, thmnm) => (translate_thm_i tm thmnm "")) (tm2', img1)
        val thmMap2 = ThyMorph.thmMap tm2''
        val sg1 = ThyMorph.sig_morph tm1
        val sg2 = ThyMorph.sig_morph tm2''
        val csg = sigmorph_compose sg1 sg2

        val (ThyMorph.ThyMorphData{local_src_axms, ...}) = ThyMorph.get_thymorphdata csg
        val src_axnms = map fst local_src_axms

        fun make_nuthmMap [] = []
          | make_nuthmMap (ax::axs) = let val img1 = case assoc_string(thmMap1, ax) of
                                                      Some i => i
                                                     | None => ax
                                          val img2 = case assoc_string(thmMap2, img1) of
                                                      Some i => i
                                                    | None => img1
                                      in
                                       (ax, img2)::(make_nuthmMap axs)
                                      end

        val cmp_tmorph = ThyMorph.make csg (make_nuthmMap src_axnms)
    in
     cmp_tmorph
    end


(* ************ Instantiation *************** *)

fun instantiate syn (nm1, incl) (nm2, tmorph) renames =
  if (ThyMorph.is_inclusion incl) 
  then (let val sigma1 = ThyMorph.sig_morph incl
            val sigma2 = ThyMorph.sig_morph tmorph
            val tpMap = SigMorph.tpMap sigma2
            val opMap = SigMorph.opMap sigma2
            val thmMap = ThyMorph.thmMap tmorph
            val dom = (let val dom1 = SigMorph.src sigma1
                           val dom2 = SigMorph.src sigma2
                       in
                         if ((Aux.name_of_thy dom1) = (Aux.name_of_thy dom2)) then dom1
                         else raise INSTANTIATE ("The domains of " ^ nm1 ^ " and " ^ nm2 ^ " must be equal")
                       end)
            val cod1 = SigMorph.trg sigma1
            val cod2 = SigMorph.trg sigma2            
            val cod2_nm = Aux.name_of_thy cod2

            val cod1_sg = sign_of cod1
            val cod2_sg = sign_of cod2
            val cod1_ltypes = ((Aux.get_ltypes cod1_sg) \\ (SigMorph.ltypes_inter sigma1))
                              \\ (Aux.get_ltypes cod2_sg)
            val cod1_ops_tab = let val tab = (SigMorph.local_trg_ops_tab sigma1) 
                                   val cod2_ops = Aux.get_ops cod2_sg
                               in
                                filter (fn (s, opt) => (not (s mem cod2_ops))) tab
                               end
                                 
            val cod1_axms = ThyMorph.local_trg_axms incl
            
            fun ins_typedecl thynm ((thy, tpMap_ext), td) =
            let val opt_r = assoc_string(renames, td)
                val td' = case opt_r of
                            None => NameSpace.base td
                          | Some (r, _) => r       
                val td'' = cod2_nm ^ "." ^ td'
                val (arity, i) = case (type_arity (cod1_sg, td)) of
                                    None => raise INSTANTIATE ("Unexpected exception in ins_typedecl")
                                  | Some (a, i) => (a, i) 
                val ext_ar_l = map (fn (cls, srtl) => (td'', srtl, [cls])) arity
                val mxfx = (case opt_r of
                             None => (if syn then
                                      (case (get_mixfix td) of
                                         None => NoSyn
                                       | Some m => m)
                                      else NoSyn)
                            | Some (r, mx) => (case mx of
                                                Some m => m
                                               | None => (if syn then
                                                           (case (get_mixfix td) of
                                                             None => NoSyn
                                                           | Some m => m)
                                                          else NoSyn)))

                val _ = writeln ("Adding typedecl: " ^ td' ^ "   " ^ (string_of_mixfix mxfx))
                val thy' = thy |> Theory.add_types [(td', i, mxfx)] |> Theory.add_arities_i ext_ar_l
             in
               (thy', (td, SigMorph.Simple td'')::tpMap_ext)
             end
            
            fun ins_typedecls nm thy tds = foldl (ins_typedecl nm) ((thy, []), tds)
            val (ncod2, tpMap_ext) = ins_typedecls cod2_nm cod2 cod1_ltypes
            val ncod2_sg = sign_of ncod2

            fun ins_const thynm ((thy, opMap_ext), c) =
            let val opt_r = assoc_string(renames, c)
                val c' = case opt_r of
                          None => NameSpace.base c
                         | Some (r, _) => r
                val c'' = cod2_nm ^ "." ^ c' 
                fun map_tp tmap (Type(nm, ts)) = 
                       (case assoc_string(tmap, nm) of
	                  Some (SigMorph.Simple to) => Type (to, map (map_tp tmap) ts)
                        | Some (SigMorph.Comp to)   => to
                        | None             => Type (nm, map (map_tp tmap) ts))
                  | map_tp _    var = var

                val tp = case (Sign.const_type cod1_sg c) of
                            None => raise INSTANTIATE ("Unexpected exception in ins_const")
                           | Some t => (map_tp (tpMap @ tpMap_ext) (Type.unvarifyT t))

                val mxfx = (case opt_r of
                             None => (if syn then 
                                       (case (get_mixfix c) of
                                         None => NoSyn
                                        | Some m => m)
                                      else NoSyn)
                            | Some (r, mx) => (case mx of
                                                Some m => m
                                               | None => (if syn then 
                                                           (case (get_mixfix c) of
                                                             None => NoSyn
                                                           | Some m => m)
                                                          else NoSyn)))

                val _ = writeln ("Adding const: " ^ c'' ^ " :: \"" ^ (Sign.string_of_typ ncod2_sg tp) ^ "\"   " ^
                                   (string_of_mixfix mxfx))

                val thy' = thy |> Theory.add_consts_i [(c', tp, mxfx)]
             in
                (thy', (c, c'')::opMap_ext)
             end
             
            fun ins_consts nm thy cs = foldl (ins_const nm)((thy, []), cs)
            val (ncod3, opMap_ext) = ins_consts cod2_nm ncod2 (map fst cod1_ops_tab)
            val ncod3_sg = sign_of ncod3
            val opMap' = (opMap @ opMap_ext)  
            val nsigma = SigMorph.make cod1 ncod3 (tpMap @ tpMap_ext) opMap'
            
            fun sel_defnms [] = []
              | sel_defnms ((opnm, opt)::tab) = case opt of 
                                                 Some defnm => (case (assoc_string (thmMap, defnm)) of
                                                                 None => (opnm, defnm)::(sel_defnms tab)
                                                                | _ => sel_defnms tab)
                                                | None => sel_defnms tab
            val cod1_defnms = sel_defnms cod1_ops_tab

            fun ins_def thynm ((thy, thmMap_ext), (opnm, defnm)) =
            let val defnm' = case (assoc_string(renames, opnm)) of
                              None => (case (assoc_string(renames, defnm)) of
                                        Some (d, _) => d
                                       | None => NameSpace.base defnm)
                             | Some (r, _) => (case (assoc_string(renames, defnm)) of
                                                Some (d, _) => d
                                               | None => (r ^ "_def"))

                val defnm'' = cod2_nm ^ "." ^ defnm'
                val iopnm = case (assoc_string (opMap', opnm)) of
                               Some i => i
                              | None => raise INSTANTIATE ("Unexpected exception in ins_def: image of " ^ opnm)
                val term = prop_of (Aux.get_thm cod1 defnm)
                val mp_term = SigMorph.map_term nsigma term
                val mp_term' = Aux.set_opname mp_term iopnm
                val mp_term'' = Logic.unvarify mp_term'
                val _ = writeln ("Adding def: " ^ defnm'')
                val thy' = thy |> IsarThy.add_defs_i (false, [Thm.no_attributes(defnm', mp_term'')]) 
             in
               (thy', (defnm, defnm'')::thmMap_ext)
             end
             
            fun ins_defs nm thy ds = foldl (ins_def nm) ((thy, []), ds)
            val (ncod4, thmMap_ext) = ins_defs cod2_nm ncod3 cod1_defnms
            val ncod4_sg = sign_of ncod4
            val nsigma' = SigMorph.cod_ext nsigma ncod4

            fun ins_axm ((thy, thmMap_ext), (axnm, thm)) =
              let val axnm' = NameSpace.base axnm
                  val axnm'' = cod2_nm ^ "." ^ axnm'
                  val term = prop_of thm
                  val term' = SigMorph.map_term nsigma' term
                  val thy' = thy |> IsarThy.add_axioms_i [Thm.no_attributes (axnm', Logic.unvarify term')]
                  val _ = writeln ("Adding axiom: " ^ axnm'' ^ " : " ^(string_of_thm (get_axiom thy' axnm''))) 
              in
               (thy', (axnm, axnm'')::thmMap_ext)
              end
            fun ins_axms thy axs = foldl ins_axm ((thy, []), axs)
            val (ncod5, thmMap_ext') = ins_axms ncod4 cod1_axms
            val nsigma'' = SigMorph.cod_ext nsigma' ncod5 

            val ntmorph = ThyMorph.make nsigma'' (thmMap @ thmMap_ext @ thmMap_ext')
            val ntmorph' = (if syn then translate_syntax ntmorph else ntmorph)
       in
        ntmorph'
       end)  
  else raise INSTANTIATE ("Instantiate: theory morphism " ^ nm1 ^ " must be an inclusion")      






(* ********************************************************** *)
(* *********************** Colimit ************************** *)
(* ********************************************************** *)

(*
      tm2
   T2 -----> T3 
   |          |    
   | tm1      | t2
   v     t1   v 
   T1 -----> thy      

*)
 
(* 
   ind = 1 --> s from local source ltypes (ops) of T1 
   ind = 2 --> s from local source ltypes (ops) of T3
*) 

fun rem_dpls v ([], e) = [e]
               | rem_dpls v (e'::mp, e) = if v(e', e) then (rem_dpls v (mp, e))
                                                      else (e'::(rem_dpls v (mp, e)))

fun pluses [] = ""
  | pluses [x] = x
  | pluses (x::y::xs) = x ^ " + " ^ (pluses (y::xs))



fun build_tp_class ind Map1 Map2 s =
let fun rev_assoc ([], s) = None
      | rev_assoc ((t, t')::map, s) = case t' of
                                       SigMorph.Simple td => if (td = s) then (Some t) else rev_assoc(map, s)
                                      | _ => rev_assoc(map, s)
in
(case (assoc_string (Map1, s)) of
   Some tL => (case (assoc_string (Map2, s)) of
                Some tR => (Some tL, Some (SigMorph.Simple s), Some tR)
              | None => (Some tL, Some (SigMorph.Simple s), Some (SigMorph.Simple s)))
  | None => (case (assoc_string (Map2, s)) of
              Some tR => (Some (SigMorph.Simple s), Some (SigMorph.Simple s), Some tR)
             | None => 
               (case (rev_assoc(Map1, s)) of
                 Some t => (case (assoc_string(Map2, t)) of
                             Some tR => (Some (SigMorph.Simple s), Some (SigMorph.Simple t), Some tR)
                            | None => (Some (SigMorph.Simple s), Some (SigMorph.Simple t), Some (SigMorph.Simple t)))
               | None => (case (rev_assoc(Map2, s)) of
                           Some t => (case (assoc_string(Map1, t)) of
                                        Some tL => (Some tL, Some (SigMorph.Simple t), Some (SigMorph.Simple s))
                                      | None => (Some (SigMorph.Simple t), Some (SigMorph.Simple t), Some (SigMorph.Simple s)))
                         | None => (if ind = 1 then (Some (SigMorph.Simple s), None, None)
                                    else (if ind = 2 then (None, None, Some (SigMorph.Simple s))
                                         else raise COLIMIT ("Unexpected exception in build_op_class: index out of range"))
                                  )))))
end



fun build_op_class src_thy ind Map1 Map2 s =
let val rev1 = map (fn (a, b) => (b, a)) Map1
    val rev2 = map (fn (a, b) => (b, a)) Map2
in
(case (assoc_string(Map1, s)) of
 Some tL => (case (assoc_string (Map2, s)) of
                Some tR => (Some tL, Some s, Some tR)
               | None => (case (Aux.find_def src_thy s) of
                           Some _ => (Some tL, Some s, None)
                          | None => (Some tL, Some s, Some s)))
| None => (case (assoc_string(Map2, s)) of
            Some tR => (case (Aux.find_def src_thy s) of
                                    Some _ => (None, Some s, Some tR)
                                   | None => (Some s, Some s, Some tR))
          | None => 
            (case (assoc_string(rev1, s)) of
              Some t => (case (assoc_string(Map2, t)) of
                          Some tR => (Some s, Some t, Some tR) 
                         | None => (case (Aux.find_def src_thy t) of
                                     Some _ => (Some s, Some t, None)
                                    | None => (Some s, Some t, Some t)))
             | None => (case (assoc_string(rev2, s)) of
                         Some t => (case (assoc_string(Map1, t)) of
                                     Some tL => (Some tL, Some t, Some s) 
                                    | None => (case (Aux.find_def src_thy t) of
                                                 Some _ => (None, Some t, Some s)
                                                | None => (Some t, Some t, Some s)))
                        | None => (if ind = 1 then (Some s, None, None)
                                   else (if ind = 2 then (None, None, Some s)
                                         else raise COLIMIT ("Unexpected exception in build_op_class: index out of range"))
                                  )))))
end

fun build_axs_class src_thy ind Map1 Map2 s =
let val rev1 = map (fn (a, b) => (b, a)) Map1
    val rev2 = map (fn (a, b) => (b, a)) Map2
in
 case (assoc_string(Map1, s)) of
  Some tL => (case (assoc_string(Map2, s)) of
               Some tR => (Some tL, Some s, Some tR)
              | None => (let val opt = (Some (Aux.get_thm src_thy s)
                                        handle e => None)
                         in
                          case opt of
                           Some thm => (let val term = prop_of thm
                                        in
                                          if (Aux.chk_term term) then (Some tL, Some s, None)
                                          else (let val proof = proof_of thm
                                                in
                                                 if (Aux.chk_proof proof) then (Some tL, Some s, None)
                                                 else (Some tL, Some s, Some s)
                                                end)
                                        end)
                         | None => (Some tL, Some s, None)
                        end))
 | None => (case (assoc_string(Map2, s)) of
             Some tR => (let val opt = (Some (Aux.get_thm src_thy s)
                                        handle e => None)
                         in
                          case opt of
                           Some thm => (let val term = prop_of thm
                                        in
                                          if (Aux.chk_term term) then (None, Some s, Some tR)
                                          else (let val proof = proof_of thm
                                                in
                                                 if (Aux.chk_proof proof) then (None, Some s, Some tR)
                                                 else (Some s, Some s, Some tR)
                                                end)
                                        end)
                         | None => (None, Some s, Some tR)
                        end)
            | None => 
               (case (assoc_string(rev1, s)) of
                 Some t => (let val opt = (Some (Aux.get_thm src_thy t)
                                           handle e => None)
                            in
                             case opt of
                              Some thm => (let val term = prop_of thm
                                           in
                                            if (Aux.chk_term term) then (Some s, None, None)
                                            else (let val proof = proof_of thm
                                                in
                                                 if (Aux.chk_proof proof) then (Some s, None, None)
                                                 else (case (assoc_string (Map2, t)) of
                                                        Some tR => (Some s, Some t, Some tR)
                                                       | None => (Some s, Some t, Some t)) 
                                                end)
                                        end)
                              | None => (Some s, None, None)
                             end)
                | None => 
                  (case (assoc_string(rev2, s)) of
                    Some t => (let val opt = (Some (Aux.get_thm src_thy t)
                                           handle e => None)
                               in
                                case opt of
                                 Some thm => (let val term = prop_of thm
                                              in
                                               if (Aux.chk_term term) then (None, None, Some s)
                                               else (let val proof = proof_of thm
                                                     in
                                                      if (Aux.chk_proof proof) then (None, None, Some s)
                                                      else (case (assoc_string (Map1, t)) of
                                                        Some tL => (Some tL, Some t, Some s)
                                                       | None => (Some t, Some t, Some s)) 
                                                end)
                                        end)
                                | None => (None, None, Some s)
                               end)
                   | None => (if (ind = 1) then (Some s, None, None)
                              else (if (ind = 2) then (None, None, Some s)
                                    else raise  COLIMIT ("Unexpected exception in build_op_class: index out of range"))))))
end  


fun get_std_repr s [] = None
  | get_std_repr s (cls::classes) = 
     let val (optL, opt, optR) = cls
         fun sel_mixfix [] = NoSyn
           | sel_mixfix (opt::opslst) =
               case opt of
                Some t => (case (get_mixfix t) of
                             Some mxfx => mxfx
                           | None => (sel_mixfix opslst))
               | None => (sel_mixfix opslst)

     in
      if (s mem (mapfilter (fn x => x) [optL, opt, optR])) then
      ( case optL of 
          Some t => Some (t, 1, sel_mixfix [optL, opt, optR] )
        | None => case opt of
                    Some t => Some (t, 2, sel_mixfix [opt, optR])
                  | None => case optR of 
                              Some t => Some (t, 3, sel_mixfix [optR])
                            | None => raise COLIMIT ("Unexpected exception in get_std_op_repr")
      )
      else get_std_repr s classes
     end


                          
fun get_std_type_repr tds_order inter_types s [] = None
  | get_std_type_repr tds_order inter_types s (cls::classes) = 
     let val (optL, opt, optR) = cls
         fun mf ind [] = []
           | mf ind (x::xs) = case x of
                                Some y => (case y of 
                                            SigMorph.Simple t => (t, ind)::(mf (ind+1) xs)
                                           | _ => mf (ind+1) xs)
                              | None => mf (ind+1) xs
         fun find_comp ind [] = None
           | find_comp ind (x::xs) = case x of
                                      Some y => (case y of
                                                  SigMorph.Simple t => find_comp (ind+1) xs
                                                 | _ => Some (y, ind))
                                      | None => find_comp (ind+1) xs

         val l = mf 1 [optL, opt, optR]

         fun insert td [] = [td]
           | insert td (t::ts) = (let val (td', i) = td
                                   val (t', j) = t
                               in
                                if (tds_order td' t') then (t::(insert td ts))
                                else td::(t::ts)
                               end)
         fun sort c = foldr (uncurry insert) (c, [])   
     in
      if (s mem (map fst l)) then
         ( let val sel_inter_types = filter (fn (t, i) => t mem inter_types) l 
           in
             case sel_inter_types of
              [(t, i)] => Some (SigMorph.Simple t, i)
            | [(t1, i1), (t2, i2)] => if (t1 = t2) (* inclusion detected *)
                                      then Some (SigMorph.Simple t1, i1)
                                      else raise COLIMIT ("More than one logical type from intersection of type signatures: " ^ (commas [t1, t2]) ^ " that are in the range of type maps")
            | [(t1, i1), (t2, i2), (t3, i3)] => if (t1 = t2) then Some (SigMorph.Simple t3, i3)
                                                else if (t2 = t3) then Some (SigMorph.Simple t1, i1)
                                                     else raise COLIMIT ("More than one logical type from intersection of type signatures: " ^ (commas [t1, t2, t3]) ^ " that are in the range of type maps")
                                                
            | [] => (case (find_comp 1 [optL, opt, optR]) of
                      None =>
                       (let val l' = sort l
                        in
                         case l' of
                          [] => raise COLIMIT ("Unexpected exception in get_std_type_repr: empty l'")
                         | ((t, i)::ts) => Some(SigMorph.Simple t, i)
                        end)  
                    | optt => optt)

            | is => raise COLIMIT ("Unexpected exception in get_std_type_repr")
           end)
     else get_std_type_repr tds_order inter_types s classes
    end


(* Precondition : s is an _axiom_ from thy1 or thy2 *)
fun get_std_ax_repr thy1 thy2 s [] = None
  | get_std_ax_repr thy1 thy2 s (cls::classes) =
     let val (c1, c2, c3) = cls
         fun isThm x i =(let val opt = if (i = 1) then (Some (Aux.get_thm thy1 x)
                                                        handle e => None)
                                       else (if (i = 2) then (Some (Aux.get_thm thy2 x)
                                                                    handle e => None)
                                             else raise COLIMIT ("Unexpected exception in ax_repr: index out of range"))   
                         in
                          case opt of
                              Some thm1 => (let val proof = proof_of thm1
                                                val b = Aux.chk_proof proof 
                                            in
                                             b
                                            end)
                             | None => false (* This should be actually not possible *)
                          end)     
    in 
      case c1 of 
       Some c => (if (s = c) then 
                    (case c3 of
                      Some c' => (let val r = isThm c' 2
                                  in
                                   case r of
                                    true  => Some (c', true, 2)
                                  | false => Some (c, false, 1)
                                  end)
                     | None => Some (c, false, 1))
                   else (case c3 of 
                          Some c' => (if (s = c') then 
                                       (let val r = isThm c 1
                                        in
                                         case r of
                                          true  => Some (c, true, 1)
                                         | false => Some (c', false, 2)
                                        end)
                                      else get_std_ax_repr thy1 thy2 s classes)
                        | None => get_std_ax_repr thy1 thy2 s classes))
    | None => (case c3 of 
                Some c' => (if (s = c') then Some (c', false, 2)
                            else get_std_ax_repr thy1 thy2 s classes) 
              | None => raise COLIMIT ("Unexpected exception in ax_repr"))
    end


fun cls_to_list (c1, c2, c3) = [c1, c2, c3]

fun isInTp sym [] = false
  | isInTp sym (cls::classes) = let val cls' = mapfilter (fn x => x) (cls_to_list cls)
                                    fun chk [] = false
                                      | chk (c::cs) = (case c of
                                                        SigMorph.Simple s => (sym = s) orelse (chk cs)
                                                       | _ => chk cs)
                                in
                                 (chk cls') orelse (isInTp sym classes)
                                end

fun isInOp sym [] = false
  | isInOp sym (cls::classes) = (sym mem (mapfilter (fn x => x) (cls_to_list cls))) orelse (isInOp sym classes)


fun sym_to_class build isIn Map1 Map2 (classes, s)  =
  if (isIn s classes) 
   then classes
   else (let val cls = build Map1 Map2 s
         in
           (cls::classes)
         end) 
 




fun show_classes [] = ""
     | show_classes (cls::classes) = let val (optL, opt, optR) = cls
                                         val l = mapfilter (fn x => x) [optL, opt, optR]
                                     in 
                                        ("[ " ^ (commas l) ^ " ]") ^ (show_classes classes)
                                     end

   fun show_type (Type (nm, tpl)) = ( nm ^ "[ " ^ (commas (map show_type tpl)) ^ "]" )
     | show_type (TFree (nm, s)) = ( "( " ^ nm ^ ", " ^ (commas s) ^ " )" )
     | show_type (TVar ((is, i), s)) = ( "( (" ^ is ^ ", " ^ (string_of_int i) ^ "), " ^ (commas s) ^ " )" )        

   fun show_type' t = case t of
                         SigMorph.Simple tp => tp
                       | SigMorph.Comp tp => show_type tp

   fun show_opmap [] = ""
     | show_opmap ((s, i)::tm) = ("( " ^ s ^ " |-> " ^ i ^ " )") ^ (show_opmap tm)

   fun show_tpcls [] = ""
     | show_tpcls (cls::tm) = let val (optL, opt, optR) = cls
                                  val l = mapfilter (fn x => x) [optL, opt, optR]
                              in 
                                        ("[ " ^ (commas (map show_type' l)) ^ " ]") ^ (show_tpcls tm)
                              end





fun thymorph_colimit outstream with_output renames (thynm, Lmnm, Umnm) (tm1, tm2) =
 let val sigma1 = ThyMorph.sig_morph tm1
     val sigma2 = ThyMorph.sig_morph tm2
     val src1   = SigMorph.src sigma1
     val src2   = SigMorph.src sigma2
     val pre_trg1 = SigMorph.trg sigma1
     val pre_trg2 = SigMorph.trg sigma2
     val src1_nm = Aux.name_of_thy src1
     val trg1_nm = Aux.name_of_thy pre_trg1
     val trg2_nm = Aux.name_of_thy pre_trg2
  
     fun find_name nm names what verb =
       let val unpnm = NameSpace.unpack nm
           val nm' =
                 (case unpnm of
                     [] => raise COLIMIT "Unexpected exception in find_name"
                  | [nmsp, nms] => (case (assoc_string(renames, nms)) of
                                       Some r => NameSpace.pack [nmsp, r]
                                     | None => nm)
                  | (nmsp::nms) => (let val nms' = NameSpace.pack nms
                                    in
                                      case (assoc_string(renames, nms')) of
                                        Some r => NameSpace.pack [nmsp, r]
                                      | None => nm
                                    end)) 
                        
           fun fnnm nm nms what verb  = 
              if (nm mem nms) then (let val _ = if verb then 
                                                 writeln ("### Warning: " ^ what ^ " " ^ nm ^ " is already in use ")
                                                else ()
                                        val nnm = nm ^ "'"
                                    in
                                     fnnm nnm nms what verb
                                    end)
              else nm   
       in
        fnnm nm' names what verb
       end

fun map_comp tds_order ltypes_inter classes names (Type (nm, tpl)) = 
   (case (get_std_type_repr tds_order ltypes_inter nm classes) of
      None => Type(nm, map (map_comp tds_order ltypes_inter classes names) tpl)
    | Some (cls, i) =>               
             (case cls of
               SigMorph.Simple s => 
                 (let val s' = (thynm ^ "." ^ (NameSpace.base s))
                      val s'' = find_name s' names "type" true 
                  in
                   Type (s'', map (map_comp tds_order ltypes_inter classes names) tpl)
                  end)
              | SigMorph.Comp t =>
                 (case t of 
                    Type (nm', tpl') =>
                      (case tpl' of
                         [] => Type (nm', map (map_comp tds_order ltypes_inter classes names) tpl) 
                        | _ => (case tpl of 
                                [] => Type (nm', map (map_comp tds_order ltypes_inter classes names) tpl')
                               | _ => raise COLIMIT ("Error: type constructor " ^ nm ^ " cannot be replaced by " ^ (show_type t) )))
                  | _ => raise COLIMIT ("Unexpected exception in map_comp"))))
              
  | map_comp tds_order ltypes_inter classes names var = var


fun get_nu_type_img tds_order ltypes_inter names classes sym = 
  let val cls = (case (get_std_type_repr tds_order ltypes_inter sym classes) of
                  Some (c, i) => (case c of
                                   (SigMorph.Simple s) => 
                                     (if (s mem ltypes_inter) then Some (c, (c, i), None)
                                       else
                                        (let val s' = (thynm ^ "." ^ (NameSpace.base s))
                                             val s'' = find_name s' names "type" true 
                                         in
                                          Some ((SigMorph.Simple s''), (c, i), None)
                                         end))
                                   | SigMorph.Comp t => 
                                     Some (SigMorph.Comp (map_comp tds_order ltypes_inter classes names t), (c, i), None)) 
                | None => None)
  in
   (sym, cls)
  end
  
fun get_nu_op_img names classes sym =
  let val cls = (case (get_std_repr sym classes) of
                  Some (c, i, mxfx) => (let val c' = (thynm ^ "." ^ (NameSpace.base c))
                                            val c'' = find_name c' names "const" true
                                        in
                                          Some (c'', (c, i), Some mxfx)
                                        end)
                | None => None)
  in
   (sym, cls)
  end

fun proj (s, c) = case c of
  Some t => (s, t)
| None => raise COLIMIT ("Unexpected exception in thymorph_colimit: " ^ s ^ " has no class")


fun get_gen_img gen_map sym =
  case (assoc_string (gen_map, sym)) of 
    Some (i, (rep, ind), opt_mixfix) => (sym, i, opt_mixfix)
  | None => raise COLIMIT ("Unexpected exception in get_gen_img: " ^ sym ^ " not in map ")

fun get_gen_rep gen_map sym =
  case (assoc_string (gen_map, sym)) of 
   Some (i, (rep, ind), _) => (rep, ind)
 | None => raise COLIMIT ("Unexpected exception in get_gen_img_ind: " ^ sym ^ " not in map ")

fun write_sigmorph trgsg nm srcnm trgnm typemap opmap =
  let val pre = ("\nsigmorph " ^ nm ^ " : " ^ srcnm ^ " --> " ^ trgnm ^ "\n  maps ")
      fun show_tpmap [] = ""
        | show_tpmap ((s, t)::mp) = 
           let val pre = ("(\"" ^ s ^ "\" |-> \"")
           in
             case t of
              SigMorph.Simple lt => (let val repr = (pre ^ lt ^ "\")")
                                     in
                                      case mp of
                                       [] => repr
                                      | _ => (repr ^ ",\n        ") ^ (show_tpmap mp)
                                     end)
            | SigMorph.Comp tp => (let val repr = (pre ^ (Sign.string_of_typ trgsg tp) ^ "\")")
                                     in
                                      case mp of
                                       [] => repr
                                      | _ => (repr ^ ",\n        ") ^ (show_tpmap mp)
                                     end)          
          end
      fun show_opmap [] = ""
        | show_opmap ((s, t)::mp) = 
           let val repr = ("(\"" ^ s ^ "\" |-> \"" ^ t ^ "\")")
           in
            case mp of
             [] => repr
            | _ => (repr ^ ",\n        ") ^ (show_opmap mp)
           end
  in
   (pre ^ "[" ^ (show_tpmap typemap) ^ "]\n       [" ^ (show_opmap opmap) ^ "]\n  strictly\n")
  end                     

fun write_thymorph nm sigma_nm thmMap =
  let val pre = ("\nthymorph " ^ nm ^ " by_sigmorph " ^ sigma_nm ^ "\n  maps ")
      fun show_thmmap [] = ""
        | show_thmmap ((s, t)::mp) = 
           let val repr = ("(\"" ^ s ^ "\" |-> \"" ^ t ^ "\")")
           in
            case mp of
             [] => repr
            | _ => (repr ^ ",\n        ") ^ (show_thmmap mp)
           end
  in
   (pre ^ "[" ^ (show_thmmap thmMap) ^ "]\n")
  end                     

 in
  if (src1_nm = Aux.name_of_thy src2)
  then ( let val src    = theory src1_nm
             val trg1   = theory trg1_nm
             val trg2   = theory trg2_nm
             val sg1 = sign_of trg1
             val sg2 = sign_of src
             val sg3 = sign_of trg2
             fun get_sg_from_ind ind =
              let val sgs = [sg1, sg2, sg3]
                  fun index 1 l = hd l
                    | index n l = index (n-1) (tl l)
              in
               index ind sgs
              end

          
             val _ = if with_output then
                      (writeln ("\n\n\nColimit : " ^ trg1_nm ^ " <-- " ^ src1_nm ^ " --> " ^ trg2_nm)) else ()
(*
             val outstream = if with_output then (Some (TextIO.openOut (thynm ^ ".thy"))) else None
*)
             val _ = if with_output then (writeln ("Creating new theory " ^ thynm ^ " as colimit")) else ()

             val src_anc = map Aux.name_of_thy (Aux.isar_ancestors_of true src)
             val trg1_anc =map Aux.name_of_thy (Aux.isar_ancestors_of true trg1)
             val trg2_anc =map Aux.name_of_thy (Aux.isar_ancestors_of true trg2)
             val int1 = src_anc inter trg1_anc
             val int2 = src_anc inter trg2_anc
             val thy_anc = int1 union int2

             val intLeft = trg1_anc inter thy_anc
             val intUp   = trg2_anc inter thy_anc
             val diffL = (trg1_nm::trg1_anc) \\ intLeft
             val diffU = (trg2_nm::trg2_anc) \\ intUp
            
             val thy = IsarThy.begin_theory thynm thy_anc []
             val _ = upd_rem_tabs thynm
             val thy_ltypes = Aux.get_ltypes (sign_of thy)
             val thy_ops   = Aux.get_ops (sign_of thy)
            
             val _ = case outstream of
                      Some out => TextIO.output (out, "(* This theory is the generated pushout of the colimit:\n ") 
                     | None => ()
             val _ = case outstream of
                      Some out => TextIO.output (out,  (trg1_nm ^ " <-- " ^ src1_nm ^ " --> " ^ trg2_nm ^ " *)\n\n\n")) 
                     | None => ()
             val _ = case outstream of
                      Some out => TextIO.output (out, "header {* " ^ thynm ^ " as colimit *} \n\n\n") 
                     | None => ()
             val _ = case outstream of
                      Some out => TextIO.output (out, "theory " ^ thynm ^ " = " ^ (pluses thy_anc) ^ " : \n\n\n") 
                     | None => ()
(*
             val _ = case outstream of
                      Some out => TextIO.output (out, "use_thy " ^ trg1_nm ^ "\n") 
                     | None => ()
             val _ = case outstream of
                      Some out => TextIO.output (out, "use_thy " ^ trg2_nm ^ "\n\n\n") 
                     | None => ()
*) 
             val tpMap1 = SigMorph.tpMap sigma1
             val tpMap2 = SigMorph.tpMap sigma2
             val opMap1 = SigMorph.opMap sigma1
             val opMap2 = SigMorph.opMap sigma2
             val thm_map1 = ThyMorph.thmMap tm1
             val thm_map2 = ThyMorph.thmMap tm2
             
             val ltypes_inter = thy_ltypes (* foldl (op union) ([], map (Aux.get_ltypes o sign_of o theory) thy_anc) *) 
             val ltypesL = (Aux.get_ltypes sg1) \\ ltypes_inter
             val ltypesU = (Aux.get_ltypes sg3) \\ ltypes_inter

             val ops_inter = thy_ops (* foldl (op union) ([], map (Aux.get_ops o sign_of o theory) thy_anc) *) 
             val opsL = (Aux.get_ops sg1) \\ ops_inter
             val opsU = (Aux.get_ops sg3) \\ ops_inter

             val type_classes' = 
              (foldl (sym_to_class (build_tp_class 1) isInTp tpMap1 tpMap2) ([], ltypesL))
             val type_classes = 
              (foldl (sym_to_class (build_tp_class 2) isInTp tpMap1 tpMap2) (type_classes', ltypesU))
             val op_classes'  = 
             (foldl (sym_to_class (build_op_class src 1) isInOp opMap1 opMap2) ([], opsL))
             val op_classes =
             (foldl (sym_to_class (build_op_class src 2) isInOp opMap1 opMap2) (op_classes', opsU))

             fun make_gen_type_map tds_order classes names [] = []
               | make_gen_type_map tds_order classes names (s::syms) = 
                 let val mp = get_nu_type_img tds_order ltypes_inter names classes s
                     val (sym, (img, (rep, ind), opt)) = proj mp
                 in
                 case img of
                  SigMorph.Simple s => (sym, (img, (rep, ind), opt))::(make_gen_type_map tds_order classes names syms)
                 | _ => (sym, (img, (rep, ind), opt))::(make_gen_type_map tds_order classes names syms)
                 end

            fun make_gen_op_map classes names [] = []
              | make_gen_op_map classes names (s::syms) = 
                 let val mp = get_nu_op_img names classes s
                     val (sym, (img, (rep, ind), mxfx_opt)) = proj mp
                 in
                  (sym, (img, (rep, ind), mxfx_opt))::(make_gen_op_map classes (img::names) syms)
                 end

            val tds_tab = (let fun insert (t, s) [] = [(t, [s])]
                                 | insert (t, s) ((t', sl)::ts) =
                                    (if (t = t') then (t, s::sl)::ts
                                     else (t', sl)::(insert (t, s) ts))    
                               fun mk_tab tab [] = tab
                                 | mk_tab tab ((s, t)::mp) = 
                                   (case t of
                                      SigMorph.Simple td => mk_tab (insert (td, s) tab) mp
                                    | _ => mk_tab tab mp)
                                val tab1 = mk_tab [] tpMap1
                                val tab2 = mk_tab tab1 tpMap2
                           in
                            tab2
                           end)
            fun tds_map td = (case (assoc_string (tds_tab, td)) of
                                 Some sl => length sl
                                | None => 0) 
            fun tds_order t1 t2 = (tds_map t1) < (tds_map t2) 
                                                


            val gen_tpMapL = make_gen_type_map tds_order type_classes thy_ltypes ltypesL  
            val gen_tpMapU = make_gen_type_map tds_order type_classes thy_ltypes ltypesU
            val gen_tpMap = gen_tpMapL @ gen_tpMapU

            val gen_opMapL = make_gen_op_map op_classes thy_ops opsL 
            val gen_opMapU = make_gen_op_map op_classes thy_ops opsU 
            val gen_opMap = gen_opMapL @ gen_opMapU        

            fun map_type_to_class (Type (nm, tpl)) = 
             let val cls = case (assoc_string(gen_tpMap, nm)) of
                            Some (c, (rep, i), _) => c
                          | None   => SigMorph.Simple nm
             in
              case cls of
               SigMorph.Simple s => Type (s, map map_type_to_class tpl)
              | SigMorph.Comp t =>
                 (case t of 
                    Type (nm', tpl') =>
                      (case tpl' of
                         [] => Type (nm', map map_type_to_class tpl) 
                        | _ => (case tpl of 
                                [] => Type (nm', map map_type_to_class tpl')
                               | _ => raise COLIMIT ("Error: type constructor " ^ nm ^ " cannot be replaced by " ^ (show_type t) )))
                  | _ => raise COLIMIT ("Unexpected exception in map_type_to_class"))
              end
             | map_type_to_class var = var            

(* ***** types ***** *)
           val (nutpMapL, nutpMapU, to_ins_tds)  = 
            let val mapL = map (fn (a, b, c) => (a, b)) (map (get_gen_img gen_tpMap) ltypesL)
                val mapU = map (fn (a, b, c) => (a, b)) (map (get_gen_img gen_tpMap) ltypesU)
                val to_ins_tds = 
                 let fun get_arity (cls, srtl) = (srtl, [cls]) (* foldl op@ ([], srtl) *)
                     fun ext_ar (s, img) = 
                     (case img of
                       SigMorph.Simple t => 
                          (if (t mem thy_ltypes) then None 
                           else (let val (rep, ind) = get_gen_rep gen_tpMap s
                                     val rep_s = case rep of
                                                  SigMorph.Simple r => r
                                                | _ => raise COLIMIT ("Unexpected exception in ext_ar: Comp rep")
                                     val sg = get_sg_from_ind ind
                                     val ar = (case (type_arity(sg, rep_s)) of
                                                 Some (a, i) => (rev (map get_arity a), i)
                                               | None => raise COLIMIT ("Unexpected exception in ext_ar: type_ar: " ^ s ^ 
                                                                        ", " ^ (string_of_int ind)))
                                 in
                                   Some ((NameSpace.base t), ar)
                                 end))
                       | _ => None)

                       val imgL = mapfilter ext_ar mapL
                       val imgU = mapfilter ext_ar mapU
                       fun rel ((s, a), (t, a')) = (s = t)
                       val f_img = foldl (rem_dpls rel) (imgL, imgU) 
               in
                 f_img
               end  
            in
             (mapL, mapU, to_ins_tds)
            end

          fun ins_typedecl (thy, td) = 
            let val (t, a) = td
                val (ar_l, i) = a
                val ext_ar_l = map (fn (srtl, srt) => ((thynm ^ "." ^ t), srtl, srt)) ar_l
                val thy' = thy |> Theory.add_types [(t, i, NoSyn)] |> Theory.add_arities_i ext_ar_l
 
                fun type_ar(sg, tycon) = Symtab.lookup(#tycons (Type.rep_tsig (Sign.tsig_of sg)), tycon)
                fun repr 0 = []
                  | repr i = ("'"^(chr (97 + i - 1)))::(repr (i-1)) 

                val t_repr = let val ar = type_ar (sign_of thy', (thynm ^ "." ^ t))
                             in
                              case ar of
			       Some 0 => ""
                              | Some 1 => commas (repr 1)
                              | Some n => ("(" ^ (commas (rev (repr n))) ^ ")")
                              | None => raise COLIMIT ("Unexpected exception in ins_typedecl")
                             end
                              
                val _ = case outstream of 
                         Some out => TextIO.output (out, "\ntypedecl " ^ t_repr ^ " " ^ t ) 
                        | None => ()
                val _ = if with_output then (writeln (t_repr ^ " " ^ t )) else ()
               in 
                 thy'
               end  

             val thy' = let val _ = if with_output then 
                                     (if (to_ins_tds = []) then () else writeln ("\nAdding typedecls:"))
                                    else ()
                            val _ = (if (to_ins_tds = []) then () else
                                      (case outstream of
                                         Some out => TextIO.output (out, "\n\nsection {* Types *}\n")
                                        | None => ()))
                         in
                          foldl ins_typedecl (thy, to_ins_tds)
                         end 


(* ****** ops ********* *)
             val (nuopMapL, nuopMapU, to_ins_ops)=
                 let val mapL = map (get_gen_img gen_opMap) opsL
                     val mapU = map (get_gen_img gen_opMap) opsU                  
                     fun ext_map [] = []
                       | ext_map ((s, t, opt_mxfx)::mp) = 
                          let val (rep, ind) = get_gen_rep gen_opMap s 
                              val sg = get_sg_from_ind ind
                              val tp = case (Sign.const_type sg rep) of
                                        Some x => x
                                      | None => raise COLIMIT ("Unexpected exception in ext_map, op:" ^ rep  ^ ", ind: " ^ (string_of_int ind))
                              val mtp = map_type_to_class tp
                              val mxfx = case opt_mxfx of
                                          Some m => m
                                         | None => raise COLIMIT ("Unexpected exception in ext_map : mixfix ")
                          in
                           (t, Type.unvarifyT mtp, mxfx)::(ext_map mp)
                          end
                                                    
                      val mapL_ext = ext_map mapL
                      val mapU_ext = ext_map mapU
                      fun rel' ((s', t', u'), (s, t, u)) = (s' = s)
                      val to_ins_ops = foldl (rem_dpls rel') (mapL_ext, mapU_ext) 
                 in
                    (map (fn (a, b, c) => (a, b)) mapL, 
                     map (fn (a, b, c) => (a, b)) mapU, 
                     map (fn (a, b, c) => (NameSpace.base a, b, c)) to_ins_ops)
                 end

             fun ins_op sg (thy, (nm, t, mxfx)) = 
               let val _ = case outstream of
                            Some out => TextIO.output (out, (nm ^ " :: \"" ^ (Sign.string_of_typ sg t) ^ "\"   " ^
                                                             (string_of_mixfix mxfx) ^ "\n" )) 
                           | None => ()
                   val _ = if with_output then writeln (nm ^ " :: \"" ^ (Sign.string_of_typ sg t) ^ "\"") else ()
               in
                thy |> Theory.add_consts_i [(nm, t, mxfx)]
               end
             
             val thy'' = let val _ = case outstream of 
                                      Some out => (if (to_ins_ops = []) then () 
                                                   else (TextIO.output (out, "\n\nsection {* Operations *}\n");
                                                         TextIO.output(out, "\nconsts\n")))
                                     | None => () 
                             val _ = if with_output then 
                                      (if (to_ins_ops = []) then () else writeln ("\nAdding consts:"))
                                     else ()
                         in
                           foldl (ins_op (sign_of thy')) (thy', to_ins_ops)
                         end  

(* *************************** *)
(* ******* sigmorphs ********* *)

(*
val _ = writeln ("tpMaps: " ^ (commas (map (fn (a, b) => (a ^ " |-> " ^ (show_type' b))) (nutpMapL @ nutpMapU))))
*)
             val sigmaL = SigMorph.make trg1 thy'' nutpMapL nuopMapL
             val sigmaU = SigMorph.make trg2 thy'' nutpMapU nuopMapU
             val sigmaL_nm = (Lmnm^"_sigma")
             val sigmaU_nm = (Umnm^"_sigma")
             val _ = case outstream of
                      Some out => (TextIO.output (out, "\n\nsection {* Signature morphisms *}\n");
                                   TextIO.output(out, write_sigmorph (sign_of thy'') sigmaL_nm trg1_nm thynm
                                                                      nutpMapL nuopMapL))
                     | None => ()
  
             val _ = case outstream of 
                      Some out => TextIO.output(out, write_sigmorph (sign_of thy'') sigmaU_nm trg2_nm thynm
                                                                     nutpMapU nuopMapU)
                     | None => ()

(* *************************** *)
(* ******* defs ********+++*** *)
             val (thm_mapL1, thm_mapU1, thy''', nsigmaL, nsigmaU) =
              let fun check_op thy sigma nm_ext (checked, (s, t)) =
                     case (Aux.find_def thy s) of
                      Some (nm, thm) => 
                       (let fun chk [] nm (s, t) = None
                              | chk ((t', (nm', nunm), _)::cl) nm (s, t) =
                                 if (t = t') then (Some (t, (nm, nunm), None))
                                 else (chk cl nm (s, t))
                        in
                         case (chk checked nm (s, t)) of
                           Some x => (x::checked)
                         | None => 
                             (let val term = prop_of thm
                                  val term' = Aux.set_opname term t
                                  val term'' = SigMorph.map_term sigma term'
                                  val term''' = Logic.unvarify term''
                                  val nunm = find_name (thynm ^ "." ^ (NameSpace.base t)) [] "axiom" true
                              in
                               ((t, (nm, nunm ^ "_def"), Some (NameSpace.base (nunm ^ "_def"), term'''))::checked)
                              end)
                         end)
                      | None => checked

                    fun transf init opmp nm_ext thy sigma = foldl (check_op thy sigma nm_ext) (init, opmp)                              
                    val transfL = transf [] nuopMapL trg1_nm trg1 sigmaL
                    val transfU = transf transfL nuopMapU trg2_nm trg2 sigmaU

                    val thm_mapL1 = map (fn (a, b, c) => b) transfL
                    val thm_mapU1 = (map (fn (a, b, c) => b) transfU) \\ thm_mapL1
                    
                    val to_ins_defs = mapfilter (fn (a, b, c) => c) transfU
                     
                    fun ins_def (thy, (nm, term)) =
                      let val _ = writeln nm
                          val thy' = thy |> IsarThy.add_defs_i (false, [Thm.no_attributes (nm, term)])
                          val thm = Aux.get_thm thy' (thynm ^ "." ^ nm)
                          val term = Logic.unvarify (prop_of thm)
                          val repr = Pretty.string_of (Sign.pretty_term' (Theory.syn_of thy') (sign_of thy') term)
                          val _ = case outstream of 
                                   Some out => TextIO.output (out, ("\n" ^ nm ^ " : \"" ^ repr ^ "\"\n")) 
                                  | None => () 
                      in
                       thy'
                      end         

                     val nthy = let val _ = if with_output then 
                                             (if (to_ins_defs = []) then () else writeln ("\nAdding defs:"))
                                            else ()
                                    val _ = if (to_ins_defs = []) then () else 
                                             (case outstream of
                                               Some out => 
                                                (TextIO.output (out, "\n\nsection {* Definitions *}\n");
                                                 TextIO.output (out, "\ndefs\n")) 
                                              | None => ())
                                in
                                  foldl ins_def (thy'', to_ins_defs)
                                end

                     val sigmaL' = SigMorph.cod_ext sigmaL nthy
                     val sigmaU' = SigMorph.cod_ext sigmaU nthy
                     
                  in
                   (thm_mapL1, thm_mapU1, nthy, sigmaL', sigmaU')
                  end

(* ************** *)    
(* *** axioms *** *)
(* ************** *)
             val axsL =  foldl (op union) ([], map ((map fst) o Aux.select_axioms o thms_of o theory) diffL) 
             val axsU =  foldl (op union) ([], map ((map fst) o Aux.select_axioms o thms_of o theory) diffU)
             val ax_classes'  = 
             (foldl (sym_to_class (build_axs_class src 1) isInOp thm_map1 thm_map2) ([], axsL))
             val ax_classes =
             (foldl (sym_to_class (build_axs_class src 2) isInOp thm_map1 thm_map2) (ax_classes', axsU)) 

             val repr_tab = map (fn s => (let val r = get_std_ax_repr trg1 trg2 s ax_classes
                                          in
                                           case r of
                                            Some c => (s, c)
                                            | None => raise COLIMIT ("Unexpected exception: axiom " ^ s ^ " has no class")
                                          end)) (axsL @ axsU)

             val (thm_repr, ax_repr) = Aux.split_list (fn (s, (nm, b, i)) => b) repr_tab
             fun split_thm_repr [] = ([], [])
               | split_thm_repr ((s, (nm, b, i))::rs) =
                 let val (lthms, uthms) = split_thm_repr rs
                 in
                   case i of
                    1 => (((s, nm)::lthms), uthms)
                   | 2 => (lthms, ((s, nm)::uthms))
                   | _ => raise COLIMIT ("Unexpected index in split_thm_repr")
                 end
             val (lthms, uthms) = split_thm_repr thm_repr
             fun chk_split (l, u) = case l of
                                     [] => 1
                                    | _ => (case u of
                                             [] => 2
                                            | _ => raise COLIMIT ("Cyclic theorem dependencies"))
             val (thy'''', tmorphL, tmorphU) =
               let val i = chk_split (lthms, uthms)
                   val thy_thms_nms = map fst (Aux.isar_thms_of thy''')
                   fun ext_axm (s, (nm, b, i)) =
                     let val (sthy, sigma, i) = case i of
                                                 1 => (trg1, nsigmaL, 1)
                                               | 2 => (trg2, nsigmaU, 2)
                                               | _ => raise COLIMIT ("Unexpected index in ext_axm")
                         val thm = Aux.get_thm sthy nm
                         val term = prop_of thm
                         val nunm = find_name (thynm ^ "." ^ (NameSpace.base nm)) thy_thms_nms "axiom" true   
                         val nterm = SigMorph.map_term sigma term
                      in
                        if (i = 1) then ((s, (nunm, (NameSpace.base nunm, nterm))), Some (nm, nunm), None)
                        else ((s, (nunm, (NameSpace.base nunm, nterm))), None, Some (nm, nunm))
                      end 
                   val (ext_ax_repr, axm_extL, axm_extU) = let val r = map ext_axm ax_repr
                                                               val f = map (fn (a, b, c) => a) r
                                                               val s = mapfilter (fn (a, b, c) => b) r
                                                               val t = mapfilter (fn (a, b, c) => c) r
                                                           in
                                                             (f, s, t)
                                                           end

                   fun ins_axm (thy, (_, (_, (nm, term)))) =
                       let val thy'  = thy |> IsarThy.add_axioms_i [Thm.no_attributes (nm, Logic.unvarify term)]
                           val thm = Aux.get_thm thy' (thynm ^ "." ^ nm)
                           val _ = if with_output then writeln (nm ^ " : " ^ (string_of_thm thm))
                                   else ()
    
                           val term = Logic.unvarify (prop_of thm)
                           val repr = Pretty.string_of (Sign.pretty_term' (Theory.syn_of thy') (sign_of thy') term)
                           val _ = case outstream of
                                    Some out => TextIO.output (out, ("\n" ^ nm ^ " : \"" ^ repr ^ "\"\n"))
                                   | None => ()
                       in
                         thy'
                       end

                   val nthy = let val _ = if with_output then 
                                           (if (ext_ax_repr = []) then () else writeln ("\nAdding axioms:"))
                                          else ()
                                  val _ = case outstream of 
                                           Some out => (if (ext_ax_repr = []) then () 
                                                        else 
                                               (TextIO.output (out, "\n\nsection {* Axioms *}\n");
                                                TextIO.output (out, "\naxioms\n")))
                                          | None => ()
                              in
                               foldl ins_axm (thy''', ext_ax_repr)
                              end
                   val nsigmaL' = SigMorph.cod_ext nsigmaL nthy
                   val nsigmaU' = SigMorph.cod_ext nsigmaU nthy
                   
                   val (tL, tU, nthy')  =
                    let fun write_trans out mor_nm (u, nm)
                          = let val ren = case assoc_string(renames, nm) of
                                             Some r => r
                                           | None => ""
                            in
                             TextIO.output (out, 
                             ("\ntranslate_thm \"" ^ nm ^ "\" as \"" ^ ren ^ "\" along " ^ mor_nm))
                            end
                        val _ = (case outstream of
                                   Some out => TextIO.output (out, "\n\nsection {* Theory morphisms *}\n")
                                 | None => () )
                    in
                           case i of                    
                            1 => 
                              (let val thmMapU = axm_extU @ thm_mapU1
                                   val tU' = ThyMorph.make nsigmaU' thmMapU
                                   val _ = case outstream of 
                                           Some out => 
                                             (TextIO.output(out, ("\n\n\n" ^ (write_thymorph Umnm sigmaU_nm thmMapU))))
                                           | None => ()

                                   val (tU'', thm_extL) = foldl 
                                  (fn ((tm, tmp), (u, thmnm)) => 
                                    (case assoc_string(renames, thmnm) of
                                      Some r => (translate_thm_i tm thmnm r, (u, thynm ^ "." ^ r)::tmp)  
                                    | None => (translate_thm_i tm thmnm "", (u, thynm ^ "." ^(NameSpace.base thmnm))::tmp))
                                  )
                                    ((tU', []), (lthms @ uthms))

                                   val _ = case outstream of
                                            Some out => map (write_trans out Umnm) (lthms @ uthms)
                                           | None => [()]

                                   val trg = SigMorph.trg (ThyMorph.sig_morph tU'')
                                   val nsigmaL'' = SigMorph.cod_ext nsigmaL' trg
                                   val thmMapL = (axm_extL @ thm_extL @ thm_mapL1)
                                   val tL' = ThyMorph.make nsigmaL'' thmMapL 
                                   val _ = case outstream of 
                                           Some out => 
                                            (TextIO.output(out, ("\n\n\n" ^ (write_thymorph Lmnm sigmaL_nm thmMapL))))
                                           | None => ()
                               in
                                (tL', tU'', trg)
                               end)
                          | 2 => 
                              (let val thmMapL = axm_extL @ thm_mapL1
                                   val tL' = ThyMorph.make nsigmaL' thmMapL
                                   val _ = case outstream of 
                                           Some out => 
                                            (TextIO.output(out, ("\n\n\n" ^ (write_thymorph Lmnm sigmaL_nm thmMapL))))
                                           | None => ()

                                   val (tL'', thm_extU) = foldl 
                                  (fn ((tm, tmp), (u, thmnm)) => 
                                    (case assoc_string(renames, thmnm) of
                                      Some r => (translate_thm_i tm thmnm r, (u, thynm ^ "." ^ r)::tmp)  
                                    | None => (translate_thm_i tm thmnm "", (u, thynm ^ "." ^(NameSpace.base thmnm))::tmp))
                                  )
                                    ((tL', []), (lthms @ uthms))
                                   
                                   val _ = case outstream of
                                            Some out => map (write_trans out Lmnm) (lthms @ uthms)
                                           | None => [()]

                                   val trg = SigMorph.trg (ThyMorph.sig_morph tL'')
                                   val nsigmaU'' = SigMorph.cod_ext nsigmaU' trg
                                   val thmMapU = axm_extU @ thm_extU @ thm_mapU1
                                   val tU' = ThyMorph.make nsigmaU'' thmMapU 
                                   val _ = case outstream of 
                                           Some out => 
                                            (TextIO.output(out, ("\n\n\n" ^ (write_thymorph Umnm sigmaU_nm thmMapU))))
                                           | None => ()
                               in
                                (tL'', tU', trg)
                               end)
                          | _ => raise COLIMIT ("Unexpected index in init_thymorph")
                      end                                                                              
             in
              (nthy', tL, tU)
             end

             val _ = case outstream of
                      Some out => TextIO.output (out, "\n\n\nend") 
                     | None => ()

             val ethy = IsarThy.end_theory thy''''

             val _ = case outstream of
                      Some out =>(TextIO.closeOut out) 
                    | None => ()
         in
          (ethy, tmorphL, tmorphU)
         end)
  else (raise COLIMIT ("No colimit: domains are not equal - " ^ src1_nm ^ ", " ^ (Aux.name_of_thy src2)))
 end


fun fibred_sum opt_out ren (thynm, Lnm, Unm) tmlist =
 case tmlist of
  [] => raise COLIMIT "Nothing to do"
 | [t] => raise COLIMIT "Nothing to do"
 | (t::ts) => (let fun split [x] = ([], x)
                     | split (x::xs) = let val (l, e) = split xs
                                       in
                                        (x::l, e)
                                       end 
                   fun gen_nm thynm n = (thynm ^ "_" ^ (string_of_int (n+1)))
                   fun colimit ((n, nm, t1), t2) = 
                          let val gnm = gen_nm nm n
                              val gmLnm = gen_nm "mL" n
                              val gmUnm = gen_nm "mU" n
                              val out = case opt_out of
                                         None => None
                                        | _ => (Some (TextIO.openOut (gnm ^ ".thy")))
                              val (thy, tL, tU) = thymorph_colimit out true ren (gnm, gmLnm, gmUnm) (t1, t2)
                              val tU' = thymorph_compose t2 tU
                          in
                           (n+1, nm, tU')
                          end
                    val (l, lst) = split ts
                    val (_, _, t') = foldl colimit ((0, thynm, t), l)
               in
                thymorph_colimit opt_out true ren (thynm, Lnm, Unm) (t', lst)
               end)

  



end