(* ***********************************************************************
 *
 * Theory morphisms. 
 *
 * $Source: /home/uniform/awe/repository/src/Extensions/thy_morph.sml,v $
 * $Revision: 1.1 $ of $Date: 2005/06/10 13:45:14 $
 *)



(* ************************************************************************
 
 A theory morphism is given by a signature morphism extended with a map
 from axioms of the source theory to axioms (theorems) of the target 
 theory.
 We require that all axioms from source, which are not from some theory
 from the intersection of (reflexive) ancestor theories of source and target,
 have to be mapped by the theorem map. 
 Similary to SigMorph, axioms or theorems from intersection may be mapped, 
 but only by identity. 

 *************************************************************************** *)


signature THYMORPH =
sig
 type thy_morph
 type thm_map 
 
 datatype ThyMorphData = ThyMorphData of
                        { local_src_axms : (string * Thm.thm) list,
                          local_src_thms : (string * Thm.thm) list,
                          local_trg_axms : (string * Thm.thm) list,
                          local_trg_thms : (string * Thm.thm) list,
                          src_anc        : Theory.theory list,
                          DG_src         : DepAna.DG
                        }

 val get_thymorphdata : SigMorph.sig_morph -> ThyMorphData

(* Projections *)
 val sig_morph       : thy_morph -> SigMorph.sig_morph 
 val thmMap          : thy_morph -> thm_map
 val local_src_axms  : thy_morph -> (string * Thm.thm) list
 val local_src_thms  : thy_morph -> (string * Thm.thm) list
 val local_trg_axms  : thy_morph -> (string * Thm.thm) list
 val local_trg_thms  : thy_morph -> (string * Thm.thm) list
 val src_anc         : thy_morph -> Theory.theory list
 val DG_src          : thy_morph -> DepAna.DG

(* Extensions *)
 val sigma_ext         : thy_morph -> (string* SigMorph.type_map) list-> (string* string) list -> thy_morph
 val thmMap_ext        : thy_morph -> thm_map -> thy_morph

 val cod_ext   : thy_morph -> thm_map -> Theory.theory -> thy_morph
 val dom_ext   : thy_morph -> (string * SigMorph.type_map) list -> (string * string) list -> 
                 Theory.theory -> thy_morph
 

(* Constructions *)

 val make            : SigMorph.sig_morph -> thm_map -> thy_morph

 (* Try to construct a theory morphism by matching the axioms from source
    in target, using the given signature morphism. Returns a tuple (a, b, c) where
    - a is the optional theory morphism. 
    - b is the list of unmatched axioms from source 
    - c is the new signature morphism together with a thm map, where possible definitions from source are matched
    Note: a = None iff b is not empty
          if a = Some t then c is already included into t
  *) 
 val try_thymorph : SigMorph.sig_morph 
                    -> (thy_morph Library.option * string list * ((SigMorph.sig_morph * thm_map) Library.option))

 val is_inclusion : thy_morph -> bool
 val make_inclusion : Theory.theory -> Theory.theory -> thy_morph


(* Use a theory morphism to map theorems. *)

  val map_thms : thy_morph -> (string * string) list -> (string * Thm.thm) list 
                 -> ((string * string) list * Theory.theory)



  exception THYMORPH of string;

 (* visible flags *)
   val verbose : bool ref


end 


structure ThyMorph : THYMORPH =
struct 

val verbose = ref false;
fun prt_verb s = if !verbose then writeln s else ()

type thm_map = (string * string) list

datatype thy_morph = ThyMorph of { sigma     : SigMorph.sig_morph,
                                   thm_mp    : thm_map,
                                   local_src_axms : (string * Thm.thm) list,
                                   local_src_thms : (string * Thm.thm) list,
                                   local_trg_axms : (string * Thm.thm) list,
                                   local_trg_thms : (string * Thm.thm) list,
                                   src_anc : Theory.theory list,
                                   DG_src  : DepAna.DG
                                 }

exception THYMORPH of string



fun sig_morph (ThyMorph{sigma, ...}) = sigma
fun thmMap (ThyMorph{thm_mp, ...}) = thm_mp                                 
fun local_src_axms (ThyMorph{local_src_axms, ...}) = local_src_axms
fun local_src_thms (ThyMorph{local_src_thms, ...}) = local_src_thms
fun local_trg_axms (ThyMorph{local_trg_axms, ...}) = local_trg_axms
fun local_trg_thms (ThyMorph{local_trg_thms, ...}) = local_trg_thms   
fun src_anc        (ThyMorph{src_anc, ...}) = src_anc 
fun DG_src         (ThyMorph{DG_src, ...})  = DG_src  

datatype ThyMorphData = ThyMorphData of
                        { local_src_axms : (string * Thm.thm) list,
                          local_src_thms : (string * Thm.thm) list,
                          local_trg_axms : (string * Thm.thm) list,
                          local_trg_thms : (string * Thm.thm) list,
                          src_anc : Theory.theory list,
                          DG_src  : DepAna.DG
                        }


(* map the term of a thm *)
fun map_thms_term sigma thm = (SigMorph.map_term sigma (prop_of thm)) 
 


fun insert upd ([], (s, t)) = [(s, t)]
  | insert upd ((s', t')::thmMap, (s, t)) =
     if (s = s') then 
      (if (t = t') then insert upd (thmMap, (s, t))
       else
        (let val _ = (if upd then writeln ("Updating " ^ s ^ " to " ^ t )
                      else (writeln ("### Warning: ambigious theorem mapping " ^ s ^ " |-> " ^ t' ^ ", " ^ s ^ " |-> " ^ t ^ "\n### The latter will stay"))) 
         in
          insert upd (thmMap, (s, t))
         end))
     else (s', t')::(insert upd (thmMap, (s, t)))






(* Construct a theory morphism *)
fun make_i sigma thmMap' (ThyMorphData{local_src_axms, local_src_thms, local_trg_axms, local_trg_thms, src_anc, DG_src}) =
 let val trg = SigMorph.trg sigma
     val thmMap = foldl (insert false) ([], thmMap') 
     val chk_all_axms = let val t_dom = map fst thmMap
                            val src_axm_nms = map fst local_src_axms
                            val diff = src_axm_nms \\ t_dom
                        in
                         if (diff = []) then None
                         else raise THYMORPH ("Following axioms from the domain are not mapped: " ^ (commas diff))
                        end 
     fun chk_aconv (axnm, thm) =
       let val img = (case (assoc_string (thmMap, axnm)) of
                         Some i => i
                       | None => raise THYMORPH ("Unexpected exception in ThyMorph.make_i"))
       in
        if (axnm = img) then ((axnm, img), true)
        else 
          (let val i_thm = case (assoc_string (local_trg_axms, img)) of
                            Some t => t
                          | None => (case (assoc_string (local_trg_thms, img)) of
                                         Some t => t
                                       | None => Aux.get_thm trg img)
               val ax_term = map_thms_term sigma thm
           in
            ((axnm, img), (ax_term aconv (prop_of i_thm)))
           end)
         end   
          
     val chk_axms = let val c = map chk_aconv local_src_axms
                    in
                     filter (not o snd) c
                    end
  in 
   if (chk_axms = []) then ThyMorph{sigma=sigma, thm_mp=thmMap, 
                                   local_src_axms=local_src_axms, local_src_thms=local_src_thms,
                                   local_trg_axms=local_trg_axms, local_trg_thms=local_trg_thms,
                                   src_anc=src_anc, DG_src=DG_src}

   else (let fun show (s, t) = ("\naxiom " ^ s ^ " is not proved by " ^ t )
             val _ = writeln("Illegal theorem mappings: " ^ (commas (map (show o fst) chk_axms)))
         in
          raise THYMORPH ("Theory morphism construction failed")
         end)
  end


fun check_map _ _ _ [] = []
  | check_map (data as (ThyMorphData{local_src_axms, local_src_thms, local_trg_axms, local_trg_thms, ...})) 
              sigma thmMap ((srcnm, trgnm)::mp) =
    if (srcnm = trgnm) then check_map data sigma thmMap mp
    else 
    (let val img_chk = case (assoc_string(thmMap, srcnm)) of
                        Some i => (i = trgnm)
                      | None => false
     in
      if img_chk then (check_map data sigma thmMap mp)
      else
         (let val src = SigMorph.src sigma
              val trg = SigMorph.trg sigma 
              val dom_thm = (case (assoc_string (local_src_axms, srcnm)) of
                             Some d => d
                           | None => (case (assoc_string (local_src_thms, srcnm)) of
                                        Some d => d
                                      | None => Aux.get_thm src srcnm))

              val rng_thm = (case (assoc_string (local_trg_axms, trgnm)) of
                              Some d => d
                            | None => (case (assoc_string (local_trg_thms, trgnm)) of
                                        Some d => d
                                      | None => Aux.get_thm trg trgnm))                 

              val dom_term = map_thms_term sigma dom_thm  
           in
            if ((dom_term aconv (prop_of rng_thm)))
            then check_map data sigma thmMap mp 
            else (let val prf1 = proof_of dom_thm
                      val prf2 = proof_of rng_thm
                  in
                   if ((Aux.chk_proof prf1) andalso (Aux.chk_proof prf2)) then
                     (if (Term.could_unify (dom_term, (prop_of rng_thm))) then check_map data sigma thmMap mp
                      else ((srcnm, trgnm)::(check_map data sigma thmMap mp)))
                   else ((srcnm, trgnm)::(check_map data sigma thmMap mp))
                  end)
           end)
      end)
    
          

fun thmMap_ext (ThyMorph{sigma, thm_mp, local_src_axms, local_src_thms, local_trg_axms, local_trg_thms, src_anc, DG_src}) 
                ext =
 let val data = ThyMorphData{local_src_axms=local_src_axms, local_src_thms=local_src_thms,
                             local_trg_axms=local_trg_axms, local_trg_thms=local_trg_thms, src_anc=src_anc, DG_src=DG_src}
 in    
  case (check_map data sigma thm_mp ext) of
    [] => (ThyMorph{sigma=sigma, thm_mp=(foldl (insert true) (thm_mp, ext)), 
                    local_src_axms=local_src_axms, local_src_thms=local_src_thms, 
                    local_trg_axms=local_trg_axms, local_trg_thms=local_trg_thms, src_anc=src_anc, DG_src=DG_src})  
  | l  => (let fun show (s, t) = ("\n" ^ s ^ " |-> " ^ t )
                     val _ = writeln("Illegal theorem mappings: " ^ (commas (map show l)))
                 in
                  raise THYMORPH ("Theory morphism extension failed")
                 end)
  end


fun sigma_ext (ThyMorph{sigma, thm_mp, local_src_axms, local_src_thms, local_trg_axms, local_trg_thms, src_anc, DG_src}) 
               typemap_ext opmap_ext =
 let val sigma' = SigMorph.ext sigma typemap_ext opmap_ext
 in    
  (ThyMorph{sigma=sigma', thm_mp=thm_mp, 
            local_src_axms=local_src_axms, local_src_thms=local_src_thms, 
            local_trg_axms=local_trg_axms, local_trg_thms=local_trg_thms, src_anc=src_anc, DG_src=DG_src})
 end



fun get_thymorphdata sigma =
  let val src = SigMorph.src sigma
      val trg = SigMorph.trg sigma

      val src_anc' = Aux.isar_ancestors_of true src
      val trg_anc' = Aux.isar_ancestors_of true trg
      
      val ranc_src_names = map Aux.name_of_thy (src::src_anc')
      val ranc_trg_names = map Aux.name_of_thy (trg::trg_anc')
      val ranc_inter = ranc_src_names inter ranc_trg_names
      val local_src_anc = ranc_src_names \\ ranc_inter
      val local_trg_anc = ranc_trg_names \\ ranc_inter
      val src_anc = case local_src_anc of
                     [] => []
                    | l => src::(map theory (tl l))

      fun get_anc_data sorted ((ppaxms, ppthms), thynm) =
       let fun get_data b thy = 
            let val thms = (if b then Aux.isar_thms_of thy else thms_of thy)
                val (paxms, pthms) = let val not_defs = Aux.select_not_defs thms
                                     in
                                      Aux.split_list (fn (_, thm) => (not (Aux.chk_proof (proof_of thm)))) not_defs
                                     end
            in
              if sorted then
               (let val dg = DepAna.get_dg' [thy] pthms
                    val pthm_nms = map fst pthms
                    val sort_pthm_nms = DepAna.sort_dep dg pthm_nms
                    fun getthm [] = []
                      | getthm (nm::nms) =  (nm, Aux.get_thm thy nm)::(getthm nms)
                                            handle e => getthm nms
                    val sort_pthms = getthm sort_pthm_nms
                in
                 (paxms, sort_pthms)
                end)
              else (paxms, pthms)
            end
        in   
         if (Aux.name_of_thy src = thynm) then
          (let val (axms, thms) = get_data true src
           in
            (ppaxms @ axms, thms @ ppthms)
           end)
         else (if (Aux.name_of_thy trg = thynm) then
                (let val (axms, thms) = get_data true trg
                 in
                  (ppaxms @ axms, thms @ ppthms)
                 end)
               else (let val (axms, thms) = get_data false (theory thynm)
                     in
                      (ppaxms @ axms, thms @ ppthms)
                     end))
        end 
     
      val (local_src_axms, local_src_thms) = foldl (get_anc_data false) (([], []), local_src_anc)
      val (local_trg_axms, local_trg_thms) = foldl (get_anc_data false) (([], []), local_trg_anc)
      val DG_src = DepAna.get_dg' src_anc local_src_thms
  in
    ThyMorphData{local_src_axms=local_src_axms,
                 local_src_thms=local_src_thms,
                 local_trg_axms=local_trg_axms,
                 local_trg_thms=local_trg_thms,
                 src_anc=src_anc, DG_src=DG_src }
  end



fun match_cdefs sigma thm_map =
 let val src = SigMorph.src sigma
     val trg = SigMorph.trg sigma
     val (src_mpd', src_nmpd') = SigMorph.split_src_cdefs sigma
     val (trg_mpd', trg_nmpd') = SigMorph.split_trg_cdefs sigma


     fun aux' thy  = (fn (a, b, c) => (case c of 
                                        None => None
                                      | Some c' => Some (a, b, Aux.get_thm thy b)))
     val src_nmpd = mapfilter (aux' src) src_nmpd' 
     val trg_nmpd = mapfilter (aux' trg) trg_nmpd'
     val src_mpd = mapfilter (aux' src) src_mpd' 
     val trg_mpd = mapfilter (aux' trg) trg_mpd'

     fun match sigma thmMap [] = (sigma, thmMap)
       | match sigma thmMap ((opnm, defnm, thm)::rs) =
         case (assoc_string (thmMap, defnm)) of
          Some i => match sigma thmMap rs
        | None =>
     (let val opMap = SigMorph.opMap sigma
      in  
       case (assoc_string(opMap, opnm)) of
        None =>
         (let val rw_term = map_thms_term sigma thm 
              fun match_trg (defnm, term) [] = None
                | match_trg (defnm, term) ((topnm, tdefnm, tthm)::trs) =
                  let val _ = prt_verb ("Checking: " ^ defnm ^ " - " ^ tdefnm)
                  in
                    if (defnm = tdefnm) then Some (topnm, tdefnm, tthm)
                    else (let val term' = Aux.set_opname term topnm 
                          in
                           if (term' aconv (prop_of tthm)) then Some (topnm, tdefnm, tthm)
                           else (match_trg (defnm, term) trs)
                          end)
                   end

              val match_res = match_trg (defnm, rw_term) (trg_nmpd @ trg_mpd)
           in
             case match_res of 
              None => (match sigma thmMap rs)
            | Some (topnm, tdefnm, tthm) =>
               (let val sigma' = SigMorph.ext sigma [] [(opnm, topnm)]
                    val thmMap' = (defnm, tdefnm)::thmMap
                    val _ = writeln ("Found: (\"" ^ defnm ^ "\" |-> \"" ^ tdefnm ^ "\")" )
                in
                 match sigma' thmMap' rs
                end)
           end)
       | Some topnm =>
                      (case (Aux.find_def trg topnm) of
                         Some (tdefnm, _) => (let val _ = writeln ("Found: (\"" ^ defnm ^ "\" |-> \"" ^ tdefnm ^ "\")" )
                                              in
                                               match sigma ((defnm, tdefnm)::thmMap) rs
                                              end)
                        | None => match sigma thmMap rs)
                        
      end)
 in
  match sigma thm_map (src_nmpd @ src_mpd)
 end
  

 
fun chk_thm sigma thmslist (name, thm) =
     let val thm_term = map_thms_term sigma thm
         fun chk [] = (name, None)
           | chk ((tnm, tm)::ts) = 
             if (tnm = name) then (name, Some name)
             else 
             (if (thm_term aconv (prop_of tm)) then 
                 (let val _ = prt_verb (Aux.show_opmap ("", (name, tnm)))
                   in
                     (name, Some tnm) 
                   end)
              else chk ts) 
        in
         chk thmslist
        end


fun match_thms (tmorph as (ThyMorph{sigma, thm_mp, local_src_axms, local_src_thms, 
                                    local_trg_axms, local_trg_thms, ...})) =
  let fun sel_notmapped (nm, thm) = case (assoc_string(thm_mp, nm)) of
                                      None => true
                                    | _ => false
      val to_chk_thms = filter sel_notmapped local_src_thms
      val match_res = map (chk_thm sigma (local_trg_axms @ local_trg_thms)) to_chk_thms  
  in
   mapfilter (fn (a, b) => (case b of
                            Some b' => Some (a, b') 
                          | None => None)) match_res
  end


fun make sigma thmMap =
  let val (sigma', thmMap') = match_cdefs sigma thmMap
      val data = get_thymorphdata sigma'
      val tmorph = make_i sigma' thmMap' data
      val map_ext = match_thms tmorph
      val _ = if (map_ext = []) then () else writeln (foldl Aux.show_opmap ("\nMatching theorems.", map_ext)) 
  in
   thmMap_ext tmorph map_ext
  end    


fun try_thymorph sigma =
  let val (sigma', thmMap') = match_cdefs sigma []
      val (data as (ThyMorphData{local_src_axms, local_src_thms,
                                 local_trg_axms, local_trg_thms, src_anc, DG_src})) = get_thymorphdata sigma'
      
      val _ = writeln("\nMatching axioms.")
      val match_res = map (chk_thm sigma' (local_trg_axms @ local_trg_thms)) local_src_axms 

      val (ext, not_mapped)    = let fun split [] = ([], [])
                                       | split ((nm, opt)::l) =
                                          let val (mp, nmp) = split l
                                          in
                                            case opt of
                                             Some tnm => ((nm, tnm)::mp, nmp)
                                            | None => (mp, nm::nmp)
                                          end
                                 in
                                  split match_res
                                 end
  
     in
      if (not_mapped = []) then 
          (let val _ = writeln (foldl Aux.show_opmap ("", ext))  
               val tmorph = make_i sigma' (thmMap' @ ext) data
               val map_ext = match_thms tmorph
               val _ = if (map_ext = []) then () else writeln (foldl Aux.show_opmap ("\nMatching theorems.", map_ext)) 
           in
            (Some (thmMap_ext tmorph map_ext), [], None) 
           end) 
      else (None, not_mapped, Some (sigma', thmMap' @ ext)) 
     end


fun cod_ext (tmorph as (ThyMorph{sigma, thm_mp, local_src_axms, local_src_thms,
                                 local_trg_axms, local_trg_thms, src_anc, DG_src})) thmmap_ext thy =
 let val sigma' = SigMorph.cod_ext sigma thy
     val (sigma'', thmMap) = match_cdefs sigma' thm_mp

     val local_trg_nms = map fst (local_trg_axms @ local_trg_thms)
     val nuthms_nms = map fst (Aux.select_not_defs (Aux.isar_thms_of thy))
     val inter_nms = nuthms_nms inter local_trg_nms
     val thy_nms = nuthms_nms \\ inter_nms

(*
     val dg = DepAna.get_dg thy
     val sort_thy_nms = DepAna.sort_dep dg thy_nms
*)
  
     fun getthm [] = []
       | getthm (nm::nms) =  (nm, Aux.get_thm thy nm)::(getthm nms)
                              handle e => getthm nms

     val sort_thy_thms = getthm thy_nms (* sort_thy_nms *)
     val thy_axms = Aux.select_axioms sort_thy_thms
     val thy_thms = Aux.select_thms sort_thy_thms
      
     val tmorph' = ThyMorph{sigma=sigma'', thm_mp=thmMap, local_src_axms=local_src_axms,
                            local_src_thms=local_src_thms, local_trg_axms=(local_trg_axms @ thy_axms),
                            local_trg_thms=(local_trg_thms @ thy_thms), src_anc = src_anc, DG_src=DG_src}
     val tmorph'' = thmMap_ext tmorph' thmmap_ext

     val map_ext = match_thms tmorph''
     val tmorph''' = thmMap_ext tmorph'' map_ext
 in
  tmorph'''
 end


fun dom_ext (tmorph as (ThyMorph{sigma, thm_mp, local_src_axms, local_src_thms,
                                 local_trg_axms, local_trg_thms, src_anc, DG_src})) tpMap opMap thy =
 let val sigma' = SigMorph.dom_ext sigma tpMap opMap thy
     val (sigma'', thm_mp') = match_cdefs sigma' thm_mp
     
     val src = SigMorph.src sigma''
     val src_nms = map fst (Aux.select_not_defs (Aux.isar_thms_of src))
     val nuthms = Aux.select_not_defs (Aux.isar_thms_of thy)
     val nuthms_nms = map fst nuthms
     val inter_nms = nuthms_nms inter src_nms
     val thy_nms = nuthms_nms \\ inter_nms

     val dg = DepAna.get_dg thy
     val sort_thy_nms = DepAna.sort_dep dg thy_nms
     fun getthm [] = []
       | getthm (nm::nms) =  (nm, Aux.get_thm thy nm)::(getthm nms)
                              handle e => getthm nms

     val sort_thy_thms = getthm sort_thy_nms
     val thy_axms = Aux.select_axioms sort_thy_thms
     val thy_thms = Aux.select_thms sort_thy_thms

     val match_res = map (chk_thm sigma'' (local_trg_axms @ local_trg_thms)) thy_axms 
     val (ext, not_mapped)    = let fun split [] = ([], [])
                                       | split ((nm, opt)::l) =
                                          let val (mp, nmp) = split l
                                          in
                                            case opt of
                                             Some tnm => ((nm, tnm)::mp, nmp)
                                            | None => (mp, nm::nmp)
                                          end
                                 in
                                  split match_res
                                 end
     in
      if (not_mapped = []) then 
          (let val tmorph' = 
                        (let val src_anc' = (thy::src_anc)
                             val DG_src' = case thy_thms of
                                            [] => DG_src
                                           |_ => DepAna.get_dg' src_anc' (local_src_thms @ thy_thms)
                         in      
                             ThyMorph{sigma=sigma'', thm_mp=(thm_mp' @ ext), local_src_axms=(local_src_axms @ thy_axms),
                                      local_src_thms=(local_src_thms @ thy_thms), local_trg_axms=local_trg_axms,
                                      local_trg_thms=local_trg_thms, src_anc=src_anc', DG_src=DG_src'}
                         end)
               val map_ext = match_thms tmorph'
           in
            thmMap_ext tmorph' map_ext
           end)
      else (let val _ = writeln ("Cannot match following additional axioms from the new domain:\n " ^ (commas not_mapped))
            in
             raise THYMORPH ("Domain extension failed")
            end)
     end


fun is_inclusion tmorph =
  let val sigma = sig_morph tmorph
      val thm_map = thmMap tmorph
      fun chk_thmMap [] = true
        | chk_thmMap ((s, t)::mp) = (s = t) andalso (chk_thmMap mp)
  in
   (SigMorph.is_inclusion sigma) andalso (chk_thmMap thm_map) 
  end


fun make_inclusion sthy tthy =
 let val sigma = SigMorph.make_inclusion sthy tthy
 in
  make sigma []
 end    


(* ************ Theorem translation ************ *)
(* ********************************************* *)
(* ********************************************* *)



(* *************** map_thms ********************* *)

   fun imap_thm sigma thm_map renlist (thmnm, thm) = 
       let val trg     = SigMorph.trg sigma;
           val src     = SigMorph.src sigma;
           
           
    
	   (* Translate the proof term. Nasty mutual recursion here,
	    * since PThm nodes in the proof term contain proof terms
	    * into which we have to descend recursively (I think). 
	    * Apart from that, we just replace the theorem names in PThm _ 
	    * and PAxm _ nodes accoring to the thm_map, and map types and
	    * terms according to the signature morphism.
	    *)
	   fun map_thm_by_nm thmn typeinst oldprf =
               (case assoc_string(thm_map, thmn) of
		   Some maptrg => 
		       let val _         = prt_verb ("!!! Found theorem (axiom) " ^ thmn ^ " in the theorem map as " ^ 
                                                    maptrg)
                           val (Some th) = Aux.lookup_thm trg maptrg
			   val {der=(_, prf), prop, ...} = rep_thm th
			   val mtypes  = case typeinst of 
			                   Some ts =>
				 	        Some (map 
						      (SigMorph.map_type sigma)
						      ts)
					 | None => None
			   val mprop  = SigMorph.map_term sigma prop
                          
		       in case oldprf of 
			     PThm _ => PThm ((maptrg, []), 
					     map_prf prf, mprop, mtypes)
			   | PAxm _ => PAxm (maptrg, mprop, mtypes)
		       end 
		 | None => oldprf)	    
	
	    and map_thm (pthm as (PThm((thmn, _), _, _, typeinst))) =
	          map_thm_by_nm thmn typeinst pthm
	      | map_thm (paxm as (PAxm (thmn, _, typeinst))) =
	          map_thm_by_nm thmn typeinst paxm
                

            and map_prf prf = Aux.map_thms_in_prf map_thm prf


            val newname = (Aux.name_transl renlist thmnm) 
            (* Derive the translated theorem *)
	    val nuthm   = let val proofterm = (SigMorph.map_proof sigma (map_prf (Aux.fullprf_of thm)))
                              val _ =  prt_verb ("\n+++ Deriving: " ^ newname) 
                           in 
                             Aux.derive_thm trg proofterm
                          end
       in 
	    (newname, nuthm)
       end
	
		
   fun imap_thms sigma init_map renlist thms = 
       let val trg = SigMorph.trg sigma
           val src = SigMorph.src sigma
           val trg_qlf = Aux.name_of_thy trg
           val trg_thms = map fst (Aux.isar_thms_of trg)
           
       in  foldl (fn ((thm_map, pthy, tthmnms), (thmnm, thm)) =>
		     let val (newname, nuthm) = imap_thm sigma thm_map renlist (thmnm, thm)
                         fun find_name_and_ins (thmnm, thy) =                                   
                             if (thmnm mem tthmnms) 
                               then (find_name_and_ins ((thmnm ^ "_transl"), thy))
                               else          
                                  (let val base_nm = NameSpace.base thmnm
                                       val (nthy, l) = PureThy.add_thms [Thm.no_attributes(base_nm, nuthm)] thy
                                       val _ = writeln ("\nAdded theorem " ^ base_nm ^ ":")
                                       val _ = writeln (string_of_thm nuthm)
                                   in
                                     (thmnm, nthy)
                                   end)                                  
                         val (newname', nthy) = find_name_and_ins (trg_qlf ^ "." ^ newname, pthy)
		     in   
                          ((thmnm, newname')::thm_map, nthy, newname' :: tthmnms) 
		     end)
	          ((init_map, trg, trg_thms), thms)
       end


   fun map_thms (ThyMorph {sigma, thm_mp,...}) renlist thms = 
      let val (m, t, _) = imap_thms sigma thm_mp renlist thms
      in
       (m, t)
      end


end
