
signature TRAFO =
 sig
  type trafo

  (* Make a Traforule, checking the invariant, that the parameter theory is an ancestor
     of the body theory *)
  val make : Theory.theory -> Theory.theory -> trafo  
  
  (* Selector functions *)
  val param : trafo -> Theory.theory
  val body  : trafo -> Theory.theory
  val thy_morph : trafo -> ThyMorph.thy_morph
  


  exception TRAFO
  exception APPLY_TRAFO of string
 end


structure Trafo : TRAFO =
struct


datatype trafo = TRule of ThyMorph.thy_morph 
exception TRAFO
exception APPLY_TRAFO of string;



(* Create a trafo, retaining invariants *)
fun make t1 t2 = 
  let val n1 = Aux.name_of_thy t1
      val n2 = Aux.name_of_thy t2
      val fst_inv = (if (n1 = n2) then
                      (if (Sign.is_draft (sign_of t2)) then None
                       else (if (Sign.is_draft (sign_of t1)) 
                             then Some("*** Theory " ^ n1 ^ " is not an ancestor of " ^ n2 )
                             else (if (Theory.subthy (t1, t2)) then None
                                   else Some("*** Theory " ^ n1 ^ " is not an ancestor of " ^ n2 ))))
                     else (let val anc = map Aux.name_of_thy (Aux.isar_ancestors_of true t2)
                           in
                            if (n1 mem anc) then None
                            else Some("*** Theory " ^ n1 ^ " is not an ancestor of " ^ n2 )
                          end)) 

      val t1_axioms = map fst (Aux.select_axioms(Aux.select_not_defs (Aux.isar_thms_of t1)))
      val t2_axioms = map fst (Aux.select_axioms(Aux.select_not_defs (Aux.isar_thms_of t2)))
      val new_axioms = t2_axioms \\ t1_axioms
      val snd_inv = case new_axioms of
                       [] => None
                    |  l  => Some ("*** Extra rules " ^ (commas l) ^ " in the body theory " ^ n2 ); 
      val res = (mapfilter (fn x => x) [fst_inv, snd_inv]); 
  in
   case res of
      [] =>  (let val sigma = SigMorph.make t1 t2 [] []
                  val tmorph = ThyMorph.make sigma []
              in
               TRule tmorph
              end)
    | l  =>  let val _ = writeln (cat_lines l) in raise TRAFO end           
  end


fun param (TRule tmorph) = SigMorph.src(ThyMorph.sig_morph tmorph)
fun body (TRule tmorph) = SigMorph.trg(ThyMorph.sig_morph tmorph)
fun thy_morph (TRule tmorph) = tmorph


(*


(* ********** Help functions ************* *)

(* Returns the thm from body theory, that corresponds to the name in argument, if such thm exists *)  
fun find_def_by_name (TRule {param, body, ...}) name =
 let val cdefs = ((Aux.sel_cdefs (axioms_of param))) @ ((Aux.sel_cdefs (axioms_of body)));
     fun chk (axname, t) = case (Aux.get_cdef_name (prop_of t)) of 
                                  Some nm => (nm = name)              
                                | None    => raise TRAFORULE ("exception in find_def_by_name");
 in
    let val res = (filter chk cdefs) in
     if (res = nil) then None
     else if (length res = 1) then (Some (hd res))
                              else raise TRAFORULE ("exception in find_def_by_name: ambigious definition of " ^ name)
    end
 end; 


(* classify the source-constdefs in thms that are in the domain of sigma
   and in thms that are not in the domain of sigma (composite definitions). 
   Returns the tuple of lists of pairs (name, thm). 
*)
fun split_src_cdefs' sigma 
         = let val src = SigMorph.src sigma;         
               val src_cdefs_pairs = (Aux.sel_cdefs o axioms_of) src;
               val sgm_dom         = SigMorph.get_opMap_domain sigma;
               fun cond cdef_pair  = case (Aux.get_cdef_name (prop_of (snd cdef_pair))) of
                                        Some nm => (nm mem sgm_dom)              
                                      | None    => false (* raise TRAFORULE "exception in split_src_cdefs"; *)
           in
              (Aux.split_name_thm_list cond src_cdefs_pairs)
           end;

(* classify the target-constdefs in thms that are in the target of sigma
   and in thms that are not in the target of sigma (composite definitions). 
   Returns the tuple of lists of pairs (name, thm). 
*)
fun split_trg_cdefs' sigma 
         = let (* checks whether the name of thy has a ":" in the name, i. e. this thy is an Isar-transition *) 
               fun chk_trans_thy thy = 
                             let val name = Sign.name_of (sign_of thy)
                               (*  val _ = writeln ("!!! thy name: " ^ name) *)
                                 val l = explode name
                                 fun chk [] = None
                                   | chk (x::xs) = if x = ":" then (Some thy) else chk xs
                             in
                               chk l
                             end

                val trg = SigMorph.trg sigma;
                (* Find all ancestors of trg that are Isar-transitions *)
                val trans_ancestors = mapfilter (fn x => x) (map chk_trans_thy (ancestors_of trg));

                val ext_trans_ancestors = trg::(trans_ancestors);

                val accu_ax = 
                                let fun f (l, thy) = l @ (axioms_of thy) 
                                in
                                 foldl f ([], ext_trans_ancestors)
                                end

              (*  val _ = writeln ("!!! axioms of trg : " ^ (commas (map fst accu_ax))) *)    
      
               val trg_cdefs_pairs = Aux.sel_cdefs accu_ax;
               val sgm_trg         = SigMorph.get_opMap_range sigma;
               fun cond cdef_pair  = case (Aux.get_cdef_name (prop_of (snd cdef_pair))) of
                                        Some nm => (nm mem sgm_trg)              
                                      | None    => raise TRAFORULE "exception in split_trg_cdefs'"
           in
              (Aux.split_name_thm_list cond trg_cdefs_pairs)
           end;


(* *********** Handle composite definitions from parameter part ********* *)

(* Get names of all ops from param, that have been defined exactly in param *) 
fun get_param_op_names (TRule {param, body, ...}) =
       let val ops1 = SigMorph.get_ops (sign_of param);
           val ops2 = foldl op@ ([], map (SigMorph.get_ops o sign_of) (parents_of param)); 
       in
          rev (ops1 \\ ops2) 
       end;


(* Select composite definition opnames in param, i.e. consts that have a definition in param 
   Assuming, that all consts without definition in param (i. e. they belong to the signature)
   are in the domain of sigma, it is sufficient to consider only composite defs. 
*) 
fun get_composite_opnames (trafo as (TRule {param, body, ...})) =
       let val axs   = axioms_of param;
           fun sort_dep ax_list =
             let val dep_graph    = DepAna.get_dg param;
                 val sort_names   = DepAna.bfs dep_graph (map fst ax_list);
             in
               map (fn n => (n, get_axiom param n)) sort_names
             end;
    
           fun f (axname, thm)  = case (Aux.get_cdef_name (prop_of thm)) of
                                       None => None
                                     | Some opname => Some (opname, axname)
       in
        mapfilter f (sort_dep axs)
       end;


(* Extend consts names from param with their unvarified type term *)
fun ext_with_types_param trule =
 let val param_opnames = map fst (get_composite_opnames trule); 
     val TRule{param, ...} = trule;
     val psg = sign_of param;
     fun f sg nm = 
       let val tp = Sign.const_type sg nm 
       in
          case tp of
           Some t => (nm, Type.unvarifyT t, NoSyn)  (* without MixFix notation !!! *)
         | None   => raise TRAFORULE ("exception in ext_with_types: "^ nm ^" has no type!")
       end
 in
  (map (f psg) param_opnames) 
 end;




(* This function takes a signature morphism and the operation name of a constdef from sigma source as arguments.
 * It returns the operation name of the corresponding definition from sigma target if such exists, i. e. the
 * term of this operation is alfa equal to the term of the constdef from src. *)    
fun find_corr_opname sigma cdef_opname =
 let val (bs, cs) = split_src_cdefs' sigma;
     val (bt, ct) = split_trg_cdefs' sigma;
     val to_unfold_src = map snd cs;
     val to_unfold_trg = map snd ct;
     fun assoc_def ([], opname) = raise TRAFORULE ("Exception in find_corr_opname: " ^ opname ^ " has no definition") 
       | assoc_def ((n, def)::l, opname) = case (Aux.get_cdef_name (prop_of def)) of
                                            Some name => (if (name = opname) then def
                                                             else assoc_def (l, opname))
                                           | None => raise TRAFORULE (" Exception in find_corr_opname");
     val assoc_thm = assoc_def (cs, cdef_opname);
     val _   = prt_verb ("!!! Found composite definition for opname: " ^ cdef_opname); 
     val rw_thm  = Seq.hd (rewrite_tac to_unfold_src assoc_thm);
     val rw_term = prop_of rw_thm;
     val mrw_term= SigMorph.map_term sigma rw_term;
     fun find t [] = None
       | find t ((axname, th)::fs)=let val rw_th = Seq.hd (rewrite_tac to_unfold_trg th);
                                       val rw_t  = prop_of rw_th;
                                   in
                                      if ( t aconv rw_t ) then 
                                         (let val topname = case (Aux.get_cdef_name (prop_of th)) of
                                                             (Some name) => name
                                                            | None => raise TRAFORULE ("Def " ^ axname ^ " has no opname ")
                                              val _ = prt_verb ("+++ Found corresponding operation : " ^ topname); 
                                           in
                                            Some (topname, axname)
                                           end)
                                       else find t fs
                                   end
   in 
      find mrw_term ct 
  end; 


 
fun transl_param_consts trule tmorph trg_qlf =
 let val ext_list = ext_with_types_param trule;
     val sigma = ThyMorph.get_sigma tmorph;
     val namespace = trg_qlf;  

     fun t tr = let val (qlf_opname, tp, mxfx) = tr;
                    val corr_op = find_corr_opname sigma qlf_opname;
                    val baseopname = NameSpace.base qlf_opname; 
                    val trgqlf_opname = namespace ^ "." ^ baseopname;  
                in
                 case corr_op of
                   None => (Some (baseopname, SigMorph.map_type sigma tp, mxfx), None, (qlf_opname, trgqlf_opname))
                 | Some (copname, caxname) => (None, Some (qlf_opname, caxname), (qlf_opname, copname))
                end
 in
  map t ext_list
 end;


fun transl_param_cdefs trf_rule thymorph trg_qlf opnameslist corr_ax_list =
  let val sigma  = ThyMorph.get_sigma thymorph;
      val namespace = trg_qlf;
     
      
      fun def_ext sgmph (qlf_opname, trgqlf_opname)  = 
       let val optdef = find_def_by_name trf_rule qlf_opname
       in
       case optdef of  (* check whether there is a definition in the body theory associated with this name *) 
         Some (axname, def) =>
         (case (assoc_string (corr_ax_list, qlf_opname)) of  
                         (* in trans_param_consts corresponding axiom name was founded? *)

           None =>    (* i. e. we have to translate def to trg and the name stay unchanged except qualifier *) 
             (let  val t  = prop_of def;
                   val tt = Aux.set_cdef_name t trgqlf_opname;  (* Set the definition name in term *)
                   val ttt = SigMorph.map_term sgmph tt;
                   val baseaxname = NameSpace.base axname;
                   val trgqlf_axname = namespace ^ "." ^ baseaxname;
              in
               Some (Some (baseaxname, Logic.unvarify ttt), (qlf_opname, trgqlf_opname), (axname, trgqlf_axname)) 
                                                                                      
              end)

           | Some coaxname => Some (None, (qlf_opname, trgqlf_opname), (axname, coaxname))
          ) 
        | None => None
       end;
  in
   mapfilter (def_ext sigma) opnameslist 
  end;

     

(* ************* Handle the consts and constdefs from body ************* *)

 

(* Get names of all ops from body, that are not in param *) 
fun get_body_op_names (TRule {param, body, ...}) =
       let val ops1 = SigMorph.get_ops (sign_of param);
           val ops2 = SigMorph.get_ops (sign_of body); 
       in
          rev (ops2 \\ ops1) 
       end;



(* Extend consts names from body with their unvarified type term *)
fun ext_with_types trule =
 let val body_opnames  = get_body_op_names trule;
     val TRule{body, ...} = trule;
     val bsg = sign_of body;
     fun f sg nm = 
       let val tp = Sign.const_type sg nm 
       in
          case tp of
           Some t => (nm, Type.unvarifyT t, NoSyn)  (* without MixFix notation !!! *)
         | None   => raise TRAFORULE ("exception in ext_with_types: "^ nm ^" has no type!")
       end
 in
  (map (f bsg) body_opnames)
 end;


(* Translate consts names with their type term and Mixfix term along the signature morphism 
   The resulting list consists of tuples (t1, t2) where t1 is a triple that is ready for 
   adding to a theory using Theory.add_consts_i function, and t2 is a tuple that consists of
   operation name (these are already qualified) and its _qualified_ translation. 
   t2 is needed as argument for transl_body_cdefs and for sigmorph extension !!!
*)   
fun transl_body_consts trule tmorph renamelist trg_qlf =
 let val ext_list = ext_with_types trule;
     val sigma = ThyMorph.get_sigma tmorph;
     val namespace = trg_qlf; 

     fun t tr = let val (name, tp, mxfx) = tr;
                    val tname = Aux.name_transl renamelist name; 
                    val qlf_tname = namespace ^ "." ^ tname;  
                in
                 ((tname, SigMorph.map_type sigma tp, mxfx), (name, qlf_tname))
                end
 in
  map t ext_list
 end;
 

 


(* This function takes a traforule, a theory morphism and a list of tuples of names: (n, qn), where
   n is a qualified name of an operation from body and qn its _qualified_ translation. Such list is a part of 
   the result of the transl_body_consts function. 
   If there exists a corresponding definition of the operation associated with a name in the body theory, so
   this definition term will be translated along the signature morphism.
   The resulting list contains the list of tuples (t1, t2) where t1 is tuple with translated definition 
   that has to be added to the target theory, and t2 is a tuple with axiom-name and translated axiom-name 
   of the definition. Such tuple has to be added to the thm_map of the theory morphism.
*)   

  
fun transl_body_cdefs trf_rule thymorph renlist nameslist =
  let val sigma  = ThyMorph.get_sigma thymorph;
    
      fun def_ext sgmph (name, qlf_tname)  = 
       let val optdef = find_def_by_name trf_rule name
       in
       case optdef of  (* check whether there is a definition in the body theory associated with this name *) 
         Some (axname, def) =>   
         (let val taxname = let val [q, nm] = NameSpace.unpack qlf_tname
                            in
                            case (Aux.ext_def_split axname) of
                               Some n => q ^ "." ^ (Aux.name_transl renlist n) ^ "_def" 
                             | None   => q ^ "." ^ (Aux.name_transl renlist axname)
                            end;
              val t  = prop_of def;
              val tt = Aux.set_cdef_name t qlf_tname;  (* Set the definition name in term *)
              val ttt = SigMorph.map_term sgmph tt;
          in
            Some ((NameSpace.base taxname, Logic.unvarify ttt), (axname, taxname)) 
          end)
        | None => None
       end;
  in
   mapfilter (def_ext sigma) nameslist 
  end;






(* ************************ apply_trafo ************************* *)

fun ins_arrows (s1, s2) = ("(" ^ s1 ^ " |-> " ^ s2 ^ ")")

fun apply_trafo trule thymorph0 renamelist trg_qlf thy =
    let val param        = get_param trule;
        val body         = get_body trule;
        
        val thm_map = ThyMorph.get_thm_map thymorph0
        
 
        (* *** Translate composite constdefs from param to target *** *)
        val trans_pconsts= transl_param_consts trule thymorph0 trg_qlf;
        val to_ins_list1 = mapfilter (fn x => x) (map (fn (a, b, c) => a) trans_pconsts);
        val axnames      = mapfilter (fn x => x) (map (fn (a, b, c) => b) trans_pconsts);

        val opnames      = map (fn (a, b, c) => c) trans_pconsts;       
   
        val trans_pdefs  = transl_param_cdefs trule thymorph0 trg_qlf opnames axnames;
        val to_ins_list2 = mapfilter (fn x => x) (map (fn (a, b, c) => a) trans_pdefs);
        val opMap_ext    = map (fn (a, b, c) => b) trans_pdefs; 

        val thmMap_ext   = map (fn (a, b, c) => c) trans_pdefs;

        val thy1         = Theory.add_consts_i to_ins_list1 thy;            (* Added consts with their types and mixfix *)
      (*  val thy2         = (#1 oo (PureThy.add_defs_i false o map Thm.no_attributes)) to_ins_list2 thy1; *)
        val thy2         = Theory.add_defs_i false to_ins_list2 thy1;


(* *** Adjust the trg thms with to_fold defs *** *)
(* thmMap_ext contains all composite defs pairs, while thm_map contains only basic (axiom, axiom(theorem)) pairs, where
   the target axiom(theorem) has to be checked for folds of composite defs from the range of thmMap_ext *)

        val to_fold_names= map snd thmMap_ext; 
        val to_fold_list = map (fn n => (n, get_axiom thy2 n)) to_fold_names;
        val to_fold = map snd to_fold_list
        val trg_thm_list = map (fn (s, t) => (s, t, get_thm thy2 t)) thm_map;
        fun fold_thm (sthmn, tthmn, thm) = case (assoc_string (to_fold_list, tthmn)) of
                                              Some t => None
                                            | None => (let val nthm = Seq.hd (fold_tac to_fold thm)
                                                       in
                                                         if ((prop_of nthm) aconv (prop_of thm))
                                                         then None 
                                                         else
                                                          let val ntthmn = tthmn ^ "_folded"
                                                              val _ = prt_verb ("+++ Added folded theorem (axiom) " 
                                                                                ^ ntthmn)
                                                          in
                                                           bind_thm (NameSpace.base ntthmn, nthm); 
                                                           Some (sthmn, ntthmn)
                                                          end
                                                        end)
        val fold_thmMap_ext = mapfilter (fn x => x) (map fold_thm trg_thm_list);

(*   ***   *)

        val thymorph1    = ThyMorph.ext_sigma_codomain thymorph0 thy2; 
        val thymorph2    = ThyMorph.sigma_ext thymorph1 [] opMap_ext;
        val _            = prt_verb ("+++ Extended sigma with new operations: \n" 
                                     ^ (commas (map ins_arrows opMap_ext)) ); 
 
        val thymorph3    = ThyMorph.ext_thm_map thymorph2 (fold_thmMap_ext @ thmMap_ext); 
        val _            = prt_verb ("+++ Extended thm_map with new axioms: \n" 
                                     ^ (commas (map ins_arrows (fold_thmMap_ext @ thmMap_ext))) );         


        (* *** Translate defs from body to target *** *)
        val thymorph4    = ThyMorph.ext_sigma_domain thymorph3 body;     (* Extends the domain of sigma to body theory *)
        val _            = prt_verb "+++ Extended thymorph";

        val trans_bconsts= transl_body_consts trule thymorph4 renamelist trg_qlf;      (* Translate the consts of body *)
        val to_ins_list3 = map fst trans_bconsts;                                      (* to_insert list *)
        val opMap_ext2   = map snd trans_bconsts;                                 (* consts names and their translations *)
        val thy3         = Theory.add_consts_i to_ins_list3 thy2;
        val _            = writeln ("\n+++ Added consts: \n" 
                                     ^ (commas (map (fn (a, b, c) => a) to_ins_list3)) );  

        val trans_bdefs  = transl_body_cdefs trule thymorph4 renamelist opMap_ext2
        val to_ins_list4 = map fst trans_bdefs;
        val _   = writeln ("to_ins_list4: " ^ (commas (map fst to_ins_list4)))
        val thmmap_ext2  = map snd trans_bdefs;
       
      (*  val thy4         = (#1 oo (PureThy.add_defs_i false o map Thm.no_attributes)) to_ins_list4 thy3; *) 
        val thy4         = Theory.add_defs_i false to_ins_list4 thy3

        val thymorph7    = ThyMorph.ext_sigma_codomain thymorph4 thy4;
        val thymorph8    = ThyMorph.ext_thm_map thymorph7 thmmap_ext2; 
        val _            = prt_verb ("+++ Extended thm_map with new axioms: \n" 
                                     ^ (commas (map ins_arrows thmmap_ext2)) );


        (* *** Translate theorems from body to target *** *)

        val thm_names    = map fst (Aux.sel_thms (thms_of body));  
                                                              (* Get all names of proper thms (i.e. not defs) 
                                                               * of the body theory *)
        val dep_graph    = DepAna.get_dg body;
        val sort_names   = DepAna.bfs dep_graph thm_names;
        val sort_thmlst  = Aux.sel_thms(map (fn name => (name, get_thm body name)) sort_names);
        val _            = prt_verb "\n+++ Adding theorems:";       
        val mthm_res     = ThyMorph.map_thms thymorph8 renamelist trg_qlf sort_thmlst;  
     in       
        thy4
     end; 


fun apply_trafo2 param body type_map op_map thm_map renamelist trg_qlf thy =
 let val trafo = mk_Traforule param body
     val sigma0 = SigMorph.make param thy type_map op_map
     val thymorph0     = ThyMorph.mk_thymorph sigma0 thm_map;         (* Create the initial thymorph *)
     val _             = prt_verb "+++ Created initial thymorph";
 in
  apply_trafo trafo thymorph0 renamelist trg_qlf thy
 end;

*)

end